//******************************************************************************
// Copyright (c) 2011 - 2014 Dave Hein
// See end of file for terms of use.
//******************************************************************************
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* INLINE SPIN
CON
  _clkmode = xinput
  _xinfreq = $4e495053 ' SPIN
*/
// SPIN OBJECT c : "clib"
void set_linefeed();
void c_start();
// SPIN OBJECT ser : "cserial"
int ser_kbhit();
//i2c : "Basic_I2C_Driver"
int ReadLong();
void WriteLong();
int WriteWait();
// SPIN OBJECT spinix : "exit"
void exit();
// SPIN OBJECT

// Constants
#ifdef CSPIN_FLAG
#define FILE int
#else
#define clkfreq 1
#endif
#define TYPE_NOUN 1
#define TYPE_VERB 2
#define TYPE_PRONOUN 3
#define TYPE_ADJECT 4
#define TYPE_OTHER 5
#define TYPE_NAME 6
#define TYPE_ARTICLE 7
#define TYPE_QUERY 8
#define TYPE_RELATED 32
#define TYPE_EQUATE 33
#define TYPE_RELATE2 34
#define TYPE_MASK 0x3f
#define DELETE_MASK 0x40
#define MEMORY_MASK 31
#define BUF_SIZE 10000

void  start(int argc, char **argv);
int   rand(void);
int   kbhit(void);
int   GetTicks(void);
int   toupper(int val);
char *SkipChar(char *str, int val);
char *FindChar(char *str, int val);
int   tokenize(char *str, char **tokens);
char *FindWord(char *str);
void  ResetVocab(void);
void  ResetBasicVocab(void);
void  DumpHexWords(void);
void  DumpAsciiWords(char *name, int num);
void  AddWord(int attrib, char *str);
int   UnknownWord(char **tokens, int num, int i);
void  DeleteWord(char **tokens, int num);
void  SomethingWrong(char *str);
void  AddEquate(char *ptr1, char *ptr2);
char *GetNextItem(char *ptr);
char *GetNextEquate(char *ptr);
int   Match(char *ptr1, char *ptr2);
int   RemoveArticles(char **tokens, int num);
char *FindWords(char *word1, char *word2);
int   isvowel(int val);
void  AddToMemory(char *name, char *ptr1);
void  PrintMemory(void);
int   CheckMemory(char *ptr0, char *ptr1);
int   GetPlural(char *name);
char *GetVerb(int plural, char *ptr1);
void  PrintStatement(char *name, char *ptr1, int remember);
void  PrintQuestion(char *ptr1);
void  Process(char **tokens, int num, int question);
void  WaitForInput(void);
void  SaveVocab(void);
int   CheckCommand(char **tokens, int num, char *buffer);
void  config_linefeed(int argc, char **argv);
int   main(int argc, char **argv);

// Global variables
FILE *speak = 0;
int seed = 1;
int debugmode = 0;
int changeflag = 0;
int MemoryIndex = 0;
char *types[] = {"noun", "verb", "pronoun", "adjective",
        "other", "name", "article", "query"};

// Basic vocabulary
static int basicsize = 249;
char basicvocab[] = {
    0x03, 0x49, 0x00, 0x00, 0x19, 0x02, 0x61, 0x6d, 0x00, 0x00, 0x6b, 0x07, 0x61, 0x00, 0x00, 0x00,
    0x01, 0x68, 0x75, 0x6d, 0x61, 0x6e, 0x00, 0x00, 0x00, 0x20, 0x00, 0x05, 0x00, 0x10, 0x00, 0x00,
    0x03, 0x79, 0x6f, 0x75, 0x00, 0x00, 0x38, 0x02, 0x61, 0x72, 0x65, 0x00, 0x00, 0x70, 0x06, 0x4c,
    0x65, 0x72, 0x6e, 0x65, 0x72, 0x00, 0x00, 0x50, 0x20, 0x00, 0x27, 0x00, 0x2e, 0x00, 0x61, 0x02,
    0x69, 0x73, 0x00, 0x00, 0x75, 0x01, 0x70, 0x72, 0x6f, 0x67, 0x72, 0x61, 0x6d, 0x00, 0x00, 0x00,
    0x20, 0x00, 0x3f, 0x00, 0x45, 0x00, 0x66, 0x02, 0x65, 0x71, 0x75, 0x61, 0x6c, 0x73, 0x00, 0x01,
    0x17, 0x21, 0x00, 0x2e, 0x00, 0x00, 0x21, 0x00, 0x20, 0x00, 0x00, 0x21, 0x00, 0x27, 0x00, 0x7a,
    0x21, 0x00, 0x05, 0x01, 0x04, 0x21, 0x00, 0x05, 0x00, 0x7f, 0x21, 0x00, 0x3f, 0x00, 0x00, 0x21,
    0x00, 0x27, 0x00, 0x00, 0x21, 0x00, 0x3f, 0x00, 0x00, 0x02, 0x65, 0x71, 0x75, 0x61, 0x6c, 0x00,
    0x01, 0x12, 0x21, 0x00, 0x57, 0x00, 0x00, 0x21, 0x01, 0x09, 0x00, 0x00, 0x08, 0x77, 0x68, 0x6f,
    0x00, 0x01, 0x2b, 0x08, 0x77, 0x68, 0x61, 0x74, 0x00, 0x01, 0x30, 0x21, 0x01, 0x23, 0x00, 0x00,
    0x21, 0x01, 0x1c, 0x00, 0x00, 0x01, 0x63, 0x6f, 0x6d, 0x70, 0x75, 0x74, 0x65, 0x72, 0x00, 0x01,
    0x49, 0x02, 0x72, 0x75, 0x6e, 0x73, 0x00, 0x01, 0x74, 0x20, 0x01, 0x41, 0x00, 0x45, 0x00, 0x00,
    0x02, 0x64, 0x6f, 0x65, 0x73, 0x00, 0x01, 0x6a, 0x02, 0x72, 0x75, 0x6e, 0x00, 0x01, 0x6f, 0x02,
    0x64, 0x6f, 0x00, 0x01, 0x65, 0x21, 0x01, 0x50, 0x00, 0x00, 0x21, 0x01, 0x5f, 0x00, 0x00, 0x21,
    0x01, 0x41, 0x00, 0x00, 0x21, 0x01, 0x58, 0x00, 0x00};

// ****************************************************
// Replace the following lines with the dumped output
// ****************************************************
    int version = 2;
    int size = 6923;
char vocab[] = {
    0x03, 0x49, 0x00, 0x00, 0x14, 0x02, 0x61, 0x6d, 0x00, 0x00, 0x53, 0x01, 0x68, 0x75, 0x6d, 0x61,
    0x6e, 0x00, 0x00, 0x6d, 0x20, 0x00, 0x05, 0x00, 0x0b, 0x00, 0x2e, 0x01, 0x70, 0x72, 0x6f, 0x67,
    0x72, 0x61, 0x6d, 0x00, 0x01, 0x59, 0x01, 0x50, 0x72, 0x6f, 0x70, 0x00, 0x02, 0x0f, 0x20, 0x00,
    0x1b, 0x00, 0x26, 0x00, 0x45, 0x02, 0x63, 0x61, 0x6e, 0x00, 0x00, 0x00, 0x06, 0x73, 0x6c, 0x65,
    0x65, 0x70, 0x00, 0x03, 0x1d, 0x20, 0x00, 0x35, 0x00, 0x3c, 0x00, 0x00, 0x02, 0x61, 0x72, 0x65,
    0x00, 0x03, 0x26, 0x21, 0x00, 0x4c, 0x00, 0x5e, 0x02, 0x69, 0x73, 0x00, 0x03, 0x30, 0x21, 0x00,
    0x58, 0x00, 0x00, 0x01, 0x6d, 0x61, 0x6d, 0x6d, 0x61, 0x6c, 0x00, 0x03, 0x44, 0x20, 0x00, 0x58,
    0x00, 0x63, 0x00, 0x7e, 0x04, 0x6c, 0x69, 0x76, 0x69, 0x6e, 0x67, 0x00, 0x03, 0x62, 0x20, 0x00,
    0x58, 0x00, 0x74, 0x01, 0x05, 0x20, 0x00, 0x35, 0x00, 0x3c, 0x01, 0x15, 0x04, 0x74, 0x68, 0x69,
    0x6e, 0x6b, 0x00, 0x04, 0x07, 0x20, 0x00, 0x35, 0x01, 0x0c, 0x01, 0x25, 0x01, 0x62, 0x65, 0x69,
    0x6e, 0x67, 0x00, 0x31, 0x53, 0x22, 0x00, 0x58, 0x00, 0x74, 0x01, 0x1c, 0x01, 0x38, 0x01, 0x68,
    0x75, 0x6d, 0x61, 0x6e, 0x73, 0x00, 0x35, 0x2f, 0x21, 0x01, 0x2e, 0x01, 0x46, 0x02, 0x6c, 0x65,
    0x61, 0x72, 0x6e, 0x00, 0x2f, 0x4e, 0x20, 0x00, 0x35, 0x01, 0x3d, 0x00, 0x00, 0x06, 0x73, 0x6f,
    0x66, 0x74, 0x77, 0x61, 0x72, 0x65, 0x00, 0x04, 0x13, 0x20, 0x00, 0x58, 0x01, 0x4d, 0x01, 0x6f,
    0x01, 0x61, 0x70, 0x70, 0x6c, 0x69, 0x63, 0x61, 0x74, 0x69, 0x6f, 0x6e, 0x00, 0x04, 0x1a, 0x20,
    0x00, 0x58, 0x01, 0x60, 0x02, 0x02, 0x01, 0x70, 0x72, 0x6f, 0x67, 0x72, 0x61, 0x6d, 0x73, 0x00,
    0x14, 0x62, 0x21, 0x01, 0x76, 0x00, 0x00, 0x01, 0x70, 0x72, 0x6f, 0x70, 0x00, 0x04, 0x21, 0x21,
    0x02, 0x07, 0x02, 0x21, 0x01, 0x70, 0x72, 0x6f, 0x70, 0x65, 0x6c, 0x6c, 0x65, 0x72, 0x00, 0x05,
    0x7a, 0x21, 0x02, 0x14, 0x02, 0x33, 0x01, 0x50, 0x72, 0x6f, 0x70, 0x65, 0x6c, 0x6c, 0x65, 0x72,
    0x00, 0x06, 0x6a, 0x21, 0x02, 0x26, 0x02, 0x40, 0x02, 0x72, 0x75, 0x6e, 0x73, 0x00, 0x07, 0x73,
    0x20, 0x02, 0x38, 0x01, 0x4d, 0x02, 0x53, 0x06, 0x68, 0x61, 0x72, 0x64, 0x77, 0x61, 0x72, 0x65,
    0x00, 0x08, 0x01, 0x20, 0x00, 0x58, 0x02, 0x47, 0x02, 0x6d, 0x02, 0x68, 0x61, 0x73, 0x00, 0x08,
    0x33, 0x01, 0x63, 0x6f, 0x75, 0x6e, 0x74, 0x65, 0x72, 0x73, 0x00, 0x08, 0x45, 0x20, 0x02, 0x5a,
    0x02, 0x61, 0x02, 0x7c, 0x02, 0x75, 0x73, 0x65, 0x73, 0x00, 0x08, 0x63, 0x20, 0x02, 0x74, 0x02,
    0x61, 0x03, 0x0b, 0x01, 0x63, 0x68, 0x69, 0x70, 0x00, 0x08, 0x73, 0x20, 0x00, 0x58, 0x03, 0x03,
    0x00, 0x00, 0x04, 0x64, 0x6f, 0x72, 0x6d, 0x61, 0x6e, 0x74, 0x00, 0x35, 0x74, 0x22, 0x00, 0x58,
    0x01, 0x1c, 0x03, 0x12, 0x00, 0x00, 0x21, 0x00, 0x05, 0x03, 0x2b, 0x21, 0x00, 0x58, 0x00, 0x00,
    0x21, 0x00, 0x05, 0x03, 0x35, 0x21, 0x00, 0x4c, 0x00, 0x00, 0x01, 0x61, 0x6e, 0x69, 0x6d, 0x61,
    0x6c, 0x00, 0x09, 0x39, 0x20, 0x00, 0x58, 0x03, 0x3a, 0x03, 0x4b, 0x20, 0x00, 0x58, 0x00, 0x74,
    0x00, 0x00, 0x04, 0x6d, 0x65, 0x74, 0x61, 0x62, 0x6f, 0x6c, 0x69, 0x7a, 0x69, 0x6e, 0x67, 0x00,
    0x2b, 0x5e, 0x20, 0x00, 0x58, 0x03, 0x52, 0x03, 0x76, 0x04, 0x62, 0x72, 0x65, 0x61, 0x74, 0x68,
    0x69, 0x6e, 0x67, 0x00, 0x30, 0x2e, 0x20, 0x00, 0x58, 0x03, 0x69, 0x00, 0x00, 0x04, 0x72, 0x65,
    0x61, 0x73, 0x6f, 0x6e, 0x00, 0x2b, 0x52, 0x20, 0x00, 0x58, 0x03, 0x7d, 0x04, 0x0e, 0x21, 0x03,
    0x7d, 0x00, 0x00, 0x20, 0x00, 0x58, 0x00, 0x1b, 0x00, 0x00, 0x20, 0x00, 0x58, 0x00, 0x1b, 0x00,
    0x00, 0x21, 0x02, 0x14, 0x04, 0x26, 0x21, 0x02, 0x26, 0x04, 0x2b, 0x21, 0x00, 0x26, 0x04, 0x37,
    0x01, 0x68, 0x75, 0x62, 0x00, 0x09, 0x40, 0x20, 0x02, 0x5a, 0x04, 0x30, 0x04, 0x48, 0x06, 0x6d,
    0x65, 0x6d, 0x6f, 0x72, 0x79, 0x00, 0x09, 0x52, 0x20, 0x02, 0x5a, 0x04, 0x3e, 0x04, 0x62, 0x01,
    0x6d, 0x69, 0x63, 0x72, 0x6f, 0x63, 0x6f, 0x6e, 0x74, 0x72, 0x6f, 0x6c, 0x6c, 0x65, 0x72, 0x00,
    0x09, 0x79, 0x20, 0x00, 0x58, 0x04, 0x4f, 0x04, 0x71, 0x06, 0x50, 0x41, 0x53, 0x4d, 0x00, 0x0a,
    0x0c, 0x20, 0x02, 0x38, 0x04, 0x69, 0x05, 0x07, 0x01, 0x69, 0x6e, 0x74, 0x65, 0x72, 0x70, 0x72,
    0x65, 0x74, 0x65, 0x72, 0x00, 0x0a, 0x57, 0x20, 0x02, 0x5a, 0x04, 0x78, 0x05, 0x15, 0x06, 0x49,
    0x2f, 0x4f, 0x00, 0x0b, 0x1a, 0x20, 0x02, 0x5a, 0x05, 0x0e, 0x05, 0x1c, 0x20, 0x00, 0x35, 0x00,
    0x3c, 0x05, 0x2b, 0x04, 0x66, 0x61, 0x73, 0x74, 0x00, 0x00, 0x00, 0x22, 0x00, 0x58, 0x05, 0x23,
    0x03, 0x03, 0x05, 0x45, 0x04, 0x65, 0x69, 0x67, 0x68, 0x74, 0x00, 0x2e, 0x12, 0x01, 0x63, 0x6f,
    0x67, 0x73, 0x00, 0x0c, 0x53, 0x22, 0x02, 0x5a, 0x05, 0x34, 0x05, 0x3d, 0x05, 0x71, 0x02, 0x69,
    0x6e, 0x74, 0x65, 0x72, 0x70, 0x72, 0x65, 0x74, 0x73, 0x00, 0x00, 0x00, 0x06, 0x53, 0x70, 0x69,
    0x6e, 0x00, 0x0b, 0x21, 0x01, 0x62, 0x79, 0x74, 0x65, 0x63, 0x6f, 0x64, 0x65, 0x73, 0x00, 0x10,
    0x53, 0x22, 0x05, 0x4e, 0x05, 0x5c, 0x05, 0x64, 0x00, 0x00, 0x21, 0x02, 0x07, 0x05, 0x7f, 0x21,
    0x02, 0x26, 0x06, 0x04, 0x21, 0x00, 0x26, 0x06, 0x09, 0x20, 0x02, 0x38, 0x05, 0x23, 0x06, 0x10,
    0x20, 0x02, 0x38, 0x05, 0x5c, 0x06, 0x17, 0x20, 0x02, 0x5a, 0x05, 0x3d, 0x06, 0x25, 0x06, 0x52,
    0x41, 0x4d, 0x00, 0x0c, 0x58, 0x20, 0x02, 0x5a, 0x06, 0x1e, 0x06, 0x35, 0x01, 0x6c, 0x6f, 0x63,
    0x6b, 0x73, 0x00, 0x0c, 0x67, 0x20, 0x02, 0x5a, 0x06, 0x2c, 0x06, 0x44, 0x01, 0x70, 0x69, 0x6e,
    0x73, 0x00, 0x0c, 0x7a, 0x20, 0x02, 0x5a, 0x06, 0x3c, 0x06, 0x52, 0x06, 0x52, 0x4f, 0x4d, 0x00,
    0x0d, 0x1d, 0x20, 0x02, 0x5a, 0x06, 0x4b, 0x06, 0x61, 0x04, 0x67, 0x6f, 0x6f, 0x64, 0x00, 0x2c,
    0x0b, 0x22, 0x00, 0x58, 0x06, 0x59, 0x03, 0x03, 0x00, 0x00, 0x21, 0x02, 0x07, 0x06, 0x6f, 0x21,
    0x02, 0x14, 0x06, 0x74, 0x21, 0x00, 0x26, 0x07, 0x05, 0x04, 0x70, 0x6f, 0x77, 0x65, 0x72, 0x66,
    0x75, 0x6c, 0x00, 0x0d, 0x2e, 0x20, 0x00, 0x58, 0x06, 0x79, 0x07, 0x16, 0x01, 0x50, 0x38, 0x58,
    0x33, 0x32, 0x41, 0x00, 0x0d, 0x3f, 0x20, 0x00, 0x58, 0x07, 0x0c, 0x07, 0x1d, 0x20, 0x00, 0x58,
    0x05, 0x23, 0x07, 0x3a, 0x02, 0x67, 0x65, 0x6e, 0x65, 0x72, 0x61, 0x74, 0x65, 0x73, 0x00, 0x0d,
    0x50, 0x04, 0x76, 0x69, 0x64, 0x65, 0x6f, 0x00, 0x0d, 0x60, 0x20, 0x07, 0x24, 0x07, 0x31, 0x07,
    0x4a, 0x04, 0x61, 0x75, 0x64, 0x69, 0x6f, 0x00, 0x0e, 0x03, 0x20, 0x07, 0x24, 0x07, 0x41, 0x07,
    0x5e, 0x04, 0x76, 0x65, 0x72, 0x73, 0x61, 0x74, 0x69, 0x6c, 0x65, 0x00, 0x0e, 0x16, 0x20, 0x00,
    0x58, 0x07, 0x51, 0x07, 0x65, 0x20, 0x05, 0x4e, 0x05, 0x5c, 0x00, 0x00, 0x02, 0x72, 0x75, 0x6e,
    0x00, 0x0e, 0x1d, 0x21, 0x07, 0x6c, 0x00, 0x00, 0x04, 0x73, 0x74, 0x75, 0x66, 0x66, 0x00, 0x0e,
    0x2f, 0x20, 0x00, 0x58, 0x07, 0x78, 0x08, 0x22, 0x04, 0x65, 0x6c, 0x65, 0x63, 0x74, 0x72, 0x69,
    0x63, 0x61, 0x6c, 0x00, 0x00, 0x00, 0x01, 0x63, 0x69, 0x72, 0x63, 0x75, 0x69, 0x74, 0x73, 0x00,
    0x00, 0x00, 0x22, 0x00, 0x58, 0x08, 0x08, 0x08, 0x16, 0x00, 0x00, 0x02, 0x68, 0x61, 0x76, 0x65,
    0x00, 0x0e, 0x36, 0x21, 0x08, 0x2b, 0x00, 0x00, 0x01, 0x72, 0x65, 0x67, 0x69, 0x73, 0x74, 0x65,
    0x72, 0x73, 0x00, 0x0e, 0x3b, 0x20, 0x00, 0x4c, 0x08, 0x38, 0x08, 0x57, 0x01, 0x63, 0x6f, 0x75,
    0x6e, 0x74, 0x65, 0x72, 0x00, 0x2b, 0x65, 0x21, 0x08, 0x4c, 0x00, 0x00, 0x02, 0x75, 0x73, 0x65,
    0x00, 0x0e, 0x42, 0x21, 0x08, 0x5c, 0x00, 0x00, 0x06, 0x73, 0x69, 0x6c, 0x69, 0x63, 0x6f, 0x6e,
    0x00, 0x0e, 0x52, 0x20, 0x00, 0x58, 0x08, 0x68, 0x09, 0x05, 0x01, 0x63, 0x69, 0x72, 0x63, 0x75,
    0x69, 0x74, 0x00, 0x0f, 0x07, 0x20, 0x00, 0x58, 0x08, 0x7a, 0x09, 0x0c, 0x22, 0x00, 0x58, 0x08,
    0x08, 0x08, 0x7a, 0x09, 0x23, 0x04, 0x69, 0x6e, 0x74, 0x65, 0x67, 0x72, 0x61, 0x74, 0x65, 0x64,
    0x00, 0x31, 0x6d, 0x22, 0x00, 0x58, 0x09, 0x15, 0x08, 0x7a, 0x09, 0x32, 0x01, 0x49, 0x43, 0x00,
    0x31, 0x74, 0x20, 0x00, 0x58, 0x09, 0x2c, 0x00, 0x00, 0x20, 0x00, 0x58, 0x00, 0x74, 0x00, 0x00,
    0x20, 0x02, 0x5a, 0x06, 0x1e, 0x00, 0x00, 0x06, 0x73, 0x74, 0x6f, 0x72, 0x61, 0x67, 0x65, 0x00,
    0x14, 0x56, 0x20, 0x00, 0x58, 0x09, 0x47, 0x09, 0x63, 0x02, 0x73, 0x74, 0x6f, 0x72, 0x65, 0x73,
    0x00, 0x14, 0x5d, 0x20, 0x09, 0x59, 0x01, 0x76, 0x09, 0x72, 0x06, 0x64, 0x61, 0x74, 0x61, 0x00,
    0x12, 0x13, 0x20, 0x09, 0x59, 0x09, 0x6a, 0x00, 0x00, 0x20, 0x00, 0x58, 0x03, 0x03, 0x00, 0x00,
    0x06, 0x61, 0x73, 0x73, 0x65, 0x6d, 0x62, 0x6c, 0x79, 0x00, 0x14, 0x6e, 0x20, 0x00, 0x58, 0x0a,
    0x00, 0x0a, 0x1f, 0x01, 0x6c, 0x61, 0x6e, 0x67, 0x75, 0x61, 0x67, 0x65, 0x00, 0x10, 0x3d, 0x20,
    0x00, 0x58, 0x0a, 0x13, 0x0a, 0x33, 0x04, 0x61, 0x73, 0x73, 0x65, 0x6d, 0x62, 0x6c, 0x65, 0x64,
    0x00, 0x14, 0x75, 0x20, 0x00, 0x58, 0x0a, 0x26, 0x0a, 0x3a, 0x20, 0x00, 0x58, 0x05, 0x23, 0x0a,
    0x41, 0x22, 0x00, 0x58, 0x0a, 0x00, 0x0a, 0x13, 0x0a, 0x52, 0x06, 0x70, 0x61, 0x73, 0x6d, 0x00,
    0x31, 0x5c, 0x21, 0x0a, 0x4a, 0x00, 0x00, 0x20, 0x00, 0x58, 0x00, 0x1b, 0x0a, 0x7a, 0x02, 0x65,
    0x78, 0x65, 0x63, 0x75, 0x74, 0x65, 0x73, 0x00, 0x00, 0x00, 0x01, 0x69, 0x6e, 0x73, 0x74, 0x72,
    0x75, 0x63, 0x74, 0x69, 0x6f, 0x6e, 0x73, 0x00, 0x14, 0x7c, 0x20, 0x0a, 0x5e, 0x0a, 0x6a, 0x0b,
    0x0c, 0x01, 0x6f, 0x70, 0x63, 0x6f, 0x64, 0x65, 0x73, 0x00, 0x15, 0x17, 0x20, 0x0a, 0x5e, 0x0b,
    0x01, 0x0b, 0x13, 0x20, 0x00, 0x58, 0x01, 0x4d, 0x00, 0x00, 0x20, 0x02, 0x74, 0x06, 0x3c, 0x00,
    0x00, 0x20, 0x00, 0x58, 0x0a, 0x13, 0x0b, 0x33, 0x01, 0x6f, 0x62, 0x6a, 0x65, 0x63, 0x74, 0x73,
    0x00, 0x00, 0x00, 0x20, 0x02, 0x5a, 0x0b, 0x28, 0x0b, 0x3a, 0x20, 0x02, 0x74, 0x05, 0x64, 0x0b,
    0x4d, 0x04, 0x63, 0x6f, 0x6d, 0x70, 0x69, 0x6c, 0x65, 0x64, 0x00, 0x11, 0x02, 0x20, 0x00, 0x58,
    0x0b, 0x41, 0x0b, 0x5d, 0x01, 0x73, 0x74, 0x61, 0x63, 0x6b, 0x00, 0x11, 0x09, 0x20, 0x02, 0x74,
    0x0b, 0x54, 0x0b, 0x6f, 0x04, 0x63, 0x6f, 0x6d, 0x70, 0x61, 0x63, 0x74, 0x00, 0x11, 0x2f, 0x20,
    0x00, 0x58, 0x0b, 0x64, 0x0c, 0x01, 0x01, 0x6d, 0x65, 0x74, 0x68, 0x6f, 0x64, 0x73, 0x00, 0x11,
    0x40, 0x20, 0x02, 0x74, 0x0b, 0x76, 0x0c, 0x08, 0x20, 0x02, 0x5a, 0x0b, 0x76, 0x0c, 0x1c, 0x01,
    0x76, 0x61, 0x72, 0x69, 0x61, 0x62, 0x6c, 0x65, 0x73, 0x00, 0x11, 0x4e, 0x20, 0x02, 0x5a, 0x0c,
    0x0f, 0x0c, 0x23, 0x20, 0x02, 0x5a, 0x09, 0x6a, 0x0c, 0x32, 0x06, 0x73, 0x70, 0x69, 0x6e, 0x00,
    0x12, 0x1a, 0x21, 0x0c, 0x2a, 0x0c, 0x43, 0x01, 0x63, 0x6f, 0x6d, 0x70, 0x75, 0x74, 0x65, 0x72,
    0x00, 0x12, 0x55, 0x22, 0x00, 0x58, 0x0c, 0x37, 0x0a, 0x13, 0x00, 0x00, 0x01, 0x63, 0x6f, 0x67,
    0x00, 0x15, 0x2b, 0x21, 0x0c, 0x4c, 0x00, 0x00, 0x20, 0x00, 0x58, 0x04, 0x3e, 0x00, 0x00, 0x01,
    0x6c, 0x6f, 0x63, 0x6b, 0x00, 0x15, 0x4d, 0x21, 0x0c, 0x5f, 0x00, 0x00, 0x01, 0x63, 0x6f, 0x6e,
    0x6e, 0x65, 0x63, 0x74, 0x6f, 0x72, 0x73, 0x00, 0x00, 0x00, 0x20, 0x00, 0x4c, 0x0c, 0x6c, 0x0d,
    0x0a, 0x01, 0x70, 0x6c, 0x75, 0x67, 0x73, 0x00, 0x2c, 0x58, 0x20, 0x00, 0x4c, 0x0d, 0x01, 0x0d,
    0x18, 0x01, 0x70, 0x69, 0x6e, 0x00, 0x35, 0x5c, 0x21, 0x0d, 0x11, 0x00, 0x00, 0x20, 0x00, 0x58,
    0x04, 0x3e, 0x00, 0x00, 0x04, 0x73, 0x74, 0x72, 0x6f, 0x6e, 0x67, 0x00, 0x00, 0x00, 0x20, 0x00,
    0x58, 0x0d, 0x24, 0x00, 0x00, 0x01, 0x70, 0x38, 0x78, 0x33, 0x32, 0x61, 0x00, 0x10, 0x02, 0x21,
    0x0d, 0x35, 0x00, 0x00, 0x02, 0x67, 0x65, 0x6e, 0x65, 0x72, 0x61, 0x74, 0x65, 0x00, 0x10, 0x0e,
    0x21, 0x0d, 0x44, 0x00, 0x00, 0x04, 0x76, 0x69, 0x73, 0x69, 0x62, 0x6c, 0x65, 0x00, 0x21, 0x51,
    0x20, 0x00, 0x58, 0x0d, 0x55, 0x0d, 0x71, 0x01, 0x73, 0x69, 0x67, 0x6e, 0x61, 0x6c, 0x00, 0x00,
    0x00, 0x22, 0x00, 0x58, 0x08, 0x08, 0x0d, 0x67, 0x00, 0x00, 0x06, 0x73, 0x6f, 0x75, 0x6e, 0x64,
    0x00, 0x10, 0x13, 0x20, 0x00, 0x58, 0x0d, 0x7a, 0x00, 0x00, 0x04, 0x66, 0x6c, 0x65, 0x78, 0x69,
    0x62, 0x6c, 0x65, 0x00, 0x10, 0x2c, 0x20, 0x00, 0x58, 0x0e, 0x0a, 0x00, 0x00, 0x21, 0x02, 0x38,
    0x00, 0x00, 0x04, 0x73, 0x6f, 0x6d, 0x65, 0x74, 0x68, 0x69, 0x6e, 0x67, 0x00, 0x16, 0x5d, 0x20,
    0x00, 0x58, 0x0e, 0x22, 0x00, 0x00, 0x21, 0x02, 0x5a, 0x00, 0x00, 0x20, 0x00, 0x4c, 0x04, 0x3e,
    0x00, 0x00, 0x21, 0x02, 0x74, 0x00, 0x00, 0x01, 0x6d, 0x69, 0x6e, 0x65, 0x72, 0x61, 0x6c, 0x00,
    0x16, 0x71, 0x20, 0x00, 0x58, 0x0e, 0x47, 0x0e, 0x64, 0x01, 0x65, 0x6c, 0x65, 0x6d, 0x65, 0x6e,
    0x74, 0x00, 0x00, 0x00, 0x20, 0x00, 0x58, 0x0e, 0x59, 0x0e, 0x76, 0x06, 0x53, 0x69, 0x6c, 0x69,
    0x63, 0x6f, 0x6e, 0x00, 0x2c, 0x06, 0x21, 0x0e, 0x6b, 0x00, 0x00, 0x04, 0x65, 0x6c, 0x65, 0x63,
    0x74, 0x72, 0x69, 0x63, 0x00, 0x17, 0x42, 0x20, 0x00, 0x58, 0x0e, 0x7b, 0x00, 0x00, 0x07, 0x61,
    0x00, 0x00, 0x00, 0x03, 0x79, 0x6f, 0x75, 0x00, 0x0f, 0x24, 0x06, 0x4c, 0x65, 0x72, 0x6e, 0x65,
    0x72, 0x00, 0x0f, 0x30, 0x20, 0x00, 0x4c, 0x0f, 0x1a, 0x0f, 0x2b, 0x21, 0x0f, 0x1a, 0x00, 0x00,
    0x20, 0x00, 0x58, 0x00, 0x1b, 0x0f, 0x37, 0x21, 0x0f, 0x13, 0x0f, 0x3c, 0x20, 0x00, 0x58, 0x01,
    0x60, 0x0f, 0x43, 0x20, 0x00, 0x58, 0x02, 0x26, 0x0f, 0x4a, 0x20, 0x00, 0x58, 0x07, 0x0c, 0x0f,
    0x51, 0x22, 0x00, 0x58, 0x0c, 0x37, 0x00, 0x1b, 0x0f, 0x5a, 0x20, 0x00, 0x35, 0x01, 0x3d, 0x0f,
    0x79, 0x04, 0x61, 0x72, 0x74, 0x69, 0x66, 0x69, 0x63, 0x69, 0x61, 0x6c, 0x00, 0x00, 0x00, 0x01,
    0x65, 0x6e, 0x74, 0x69, 0x74, 0x79, 0x00, 0x00, 0x00, 0x22, 0x00, 0x58, 0x0f, 0x61, 0x0f, 0x6f,
    0x00, 0x00, 0x20, 0x00, 0x58, 0x02, 0x14, 0x10, 0x09, 0x21, 0x07, 0x0c, 0x00, 0x00, 0x21, 0x07,
    0x24, 0x00, 0x00, 0x20, 0x00, 0x58, 0x07, 0x41, 0x10, 0x25, 0x04, 0x61, 0x75, 0x64, 0x69, 0x62,
    0x6c, 0x65, 0x00, 0x2c, 0x76, 0x20, 0x00, 0x58, 0x10, 0x1a, 0x00, 0x00, 0x20, 0x00, 0x58, 0x07,
    0x51, 0x00, 0x00, 0x06, 0x73, 0x79, 0x6e, 0x74, 0x61, 0x78, 0x00, 0x16, 0x00, 0x20, 0x02, 0x5a,
    0x10, 0x33, 0x00, 0x00, 0x04, 0x69, 0x6e, 0x74, 0x65, 0x72, 0x70, 0x72, 0x65, 0x74, 0x65, 0x64,
    0x00, 0x00, 0x00, 0x20, 0x00, 0x4c, 0x10, 0x44, 0x10, 0x5a, 0x20, 0x00, 0x4c, 0x0b, 0x64, 0x10,
    0x69, 0x06, 0x63, 0x6f, 0x64, 0x65, 0x00, 0x16, 0x1b, 0x20, 0x00, 0x4c, 0x10, 0x61, 0x10, 0x79,
    0x01, 0x63, 0x6f, 0x64, 0x65, 0x73, 0x00, 0x00, 0x00, 0x22, 0x00, 0x4c, 0x0b, 0x64, 0x10, 0x70,
    0x00, 0x00, 0x20, 0x00, 0x58, 0x0a, 0x26, 0x00, 0x00, 0x20, 0x00, 0x58, 0x09, 0x47, 0x11, 0x10,
    0x20, 0x00, 0x58, 0x04, 0x3e, 0x11, 0x1f, 0x01, 0x46, 0x49, 0x46, 0x4f, 0x00, 0x2b, 0x71, 0x20,
    0x00, 0x58, 0x11, 0x17, 0x00, 0x00, 0x04, 0x73, 0x6d, 0x61, 0x6c, 0x6c, 0x00, 0x16, 0x2c, 0x20,
    0x00, 0x58, 0x11, 0x26, 0x00, 0x00, 0x01, 0x6d, 0x65, 0x74, 0x68, 0x6f, 0x64, 0x00, 0x16, 0x3e,
    0x21, 0x11, 0x36, 0x00, 0x00, 0x02, 0x73, 0x74, 0x6f, 0x72, 0x65, 0x00, 0x16, 0x4a, 0x20, 0x11,
    0x45, 0x09, 0x6a, 0x11, 0x6a, 0x04, 0x73, 0x79, 0x6d, 0x62, 0x6f, 0x6c, 0x69, 0x63, 0x00, 0x00,
    0x00, 0x01, 0x6e, 0x61, 0x6d, 0x65, 0x73, 0x00, 0x31, 0x11, 0x22, 0x00, 0x4c, 0x11, 0x55, 0x11,
    0x61, 0x11, 0x7f, 0x01, 0x76, 0x61, 0x72, 0x69, 0x61, 0x62, 0x6c, 0x65, 0x00, 0x35, 0x2a, 0x21,
    0x11, 0x73, 0x00, 0x00, 0x06, 0x69, 0x6e, 0x66, 0x6f, 0x72, 0x6d, 0x61, 0x74, 0x69, 0x6f, 0x6e,
    0x00, 0x2c, 0x12, 0x20, 0x00, 0x58, 0x12, 0x04, 0x00, 0x00, 0x21, 0x05, 0x5c, 0x00, 0x00, 0x02,
    0x65, 0x71, 0x75, 0x61, 0x6c, 0x73, 0x00, 0x12, 0x32, 0x02, 0x65, 0x71, 0x75, 0x61, 0x6c, 0x00,
    0x12, 0x37, 0x21, 0x12, 0x29, 0x00, 0x00, 0x21, 0x12, 0x1f, 0x00, 0x00, 0x08, 0x77, 0x68, 0x6f,
    0x00, 0x12, 0x4b, 0x08, 0x77, 0x68, 0x61, 0x74, 0x00, 0x12, 0x50, 0x21, 0x12, 0x43, 0x00, 0x00,
    0x21, 0x12, 0x3c, 0x00, 0x00, 0x20, 0x02, 0x38, 0x00, 0x1b, 0x12, 0x5c, 0x20, 0x00, 0x58, 0x02,
    0x47, 0x12, 0x6d, 0x01, 0x64, 0x65, 0x76, 0x69, 0x63, 0x65, 0x00, 0x00, 0x00, 0x22, 0x00, 0x58,
    0x08, 0x08, 0x12, 0x63, 0x12, 0x76, 0x20, 0x02, 0x38, 0x01, 0x76, 0x12, 0x7d, 0x20, 0x02, 0x38,
    0x01, 0x4d, 0x00, 0x00, 0x02, 0x64, 0x6f, 0x65, 0x73, 0x00, 0x13, 0x12, 0x02, 0x64, 0x6f, 0x00,
    0x13, 0x17, 0x21, 0x13, 0x0c, 0x00, 0x00, 0x21, 0x13, 0x04, 0x00, 0x00, 0x06, 0x50, 0x61, 0x72,
    0x61, 0x6c, 0x6c, 0x61, 0x78, 0x00, 0x13, 0x33, 0x01, 0x63, 0x6f, 0x6d, 0x70, 0x61, 0x6e, 0x79,
    0x00, 0x13, 0x66, 0x20, 0x00, 0x58, 0x13, 0x28, 0x13, 0x43, 0x02, 0x6d, 0x61, 0x6b, 0x65, 0x73,
    0x00, 0x13, 0x75, 0x20, 0x13, 0x3a, 0x02, 0x07, 0x13, 0x53, 0x02, 0x73, 0x65, 0x6c, 0x6c, 0x73,
    0x00, 0x14, 0x1f, 0x20, 0x13, 0x4a, 0x02, 0x26, 0x00, 0x00, 0x01, 0x62, 0x75, 0x73, 0x69, 0x6e,
    0x65, 0x73, 0x73, 0x00, 0x00, 0x00, 0x20, 0x00, 0x58, 0x13, 0x5a, 0x00, 0x00, 0x02, 0x6d, 0x61,
    0x6b, 0x65, 0x00, 0x14, 0x24, 0x21, 0x13, 0x6d, 0x14, 0x03, 0x02, 0x62, 0x75, 0x69, 0x6c, 0x64,
    0x00, 0x14, 0x33, 0x21, 0x13, 0x7a, 0x14, 0x12, 0x02, 0x62, 0x75, 0x69, 0x6c, 0x64, 0x73, 0x00,
    0x14, 0x42, 0x21, 0x14, 0x08, 0x00, 0x00, 0x02, 0x73, 0x65, 0x6c, 0x6c, 0x00, 0x14, 0x51, 0x21,
    0x14, 0x17, 0x00, 0x00, 0x21, 0x13, 0x3a, 0x14, 0x29, 0x21, 0x13, 0x7a, 0x14, 0x2e, 0x21, 0x14,
    0x08, 0x00, 0x00, 0x21, 0x14, 0x08, 0x14, 0x38, 0x21, 0x13, 0x6d, 0x14, 0x3d, 0x21, 0x13, 0x3a,
    0x00, 0x00, 0x21, 0x13, 0x7a, 0x14, 0x47, 0x21, 0x13, 0x6d, 0x14, 0x4c, 0x21, 0x13, 0x3a, 0x00,
    0x00, 0x21, 0x13, 0x4a, 0x00, 0x00, 0x20, 0x00, 0x58, 0x04, 0x3e, 0x00, 0x00, 0x21, 0x11, 0x45,
    0x00, 0x00, 0x20, 0x00, 0x4c, 0x0a, 0x6a, 0x14, 0x69, 0x21, 0x00, 0x1b, 0x00, 0x00, 0x20, 0x00,
    0x58, 0x0a, 0x13, 0x00, 0x00, 0x20, 0x00, 0x58, 0x0b, 0x41, 0x00, 0x00, 0x20, 0x00, 0x4c, 0x10,
    0x61, 0x15, 0x12, 0x01, 0x69, 0x6e, 0x73, 0x74, 0x72, 0x75, 0x63, 0x74, 0x69, 0x6f, 0x6e, 0x00,
    0x33, 0x1f, 0x21, 0x15, 0x03, 0x00, 0x00, 0x20, 0x00, 0x4c, 0x0a, 0x6a, 0x00, 0x00, 0x01, 0x70,
    0x72, 0x6f, 0x63, 0x65, 0x73, 0x73, 0x6f, 0x72, 0x00, 0x16, 0x56, 0x20, 0x00, 0x58, 0x15, 0x1e,
    0x15, 0x3a, 0x01, 0x63, 0x6f, 0x72, 0x65, 0x00, 0x00, 0x00, 0x20, 0x00, 0x58, 0x15, 0x32, 0x15,
    0x41, 0x20, 0x00, 0x35, 0x00, 0x3c, 0x15, 0x48, 0x21, 0x05, 0x3d, 0x00, 0x00, 0x21, 0x06, 0x2c,
    0x15, 0x5f, 0x01, 0x73, 0x65, 0x6d, 0x61, 0x70, 0x68, 0x6f, 0x72, 0x65, 0x00, 0x00, 0x00, 0x20,
    0x00, 0x58, 0x15, 0x52, 0x15, 0x70, 0x04, 0x61, 0x74, 0x6f, 0x6d, 0x69, 0x63, 0x00, 0x00, 0x00,
    0x20, 0x00, 0x58, 0x15, 0x66, 0x00, 0x00, 0x01, 0x72, 0x75, 0x6c, 0x65, 0x73, 0x00, 0x00, 0x00,
    0x20, 0x02, 0x5a, 0x15, 0x77, 0x16, 0x14, 0x04, 0x73, 0x74, 0x72, 0x75, 0x63, 0x74, 0x75, 0x72,
    0x65, 0x00, 0x17, 0x59, 0x20, 0x02, 0x5a, 0x16, 0x07, 0x00, 0x00, 0x20, 0x00, 0x58, 0x0a, 0x6a,
    0x00, 0x00, 0x04, 0x6c, 0x69, 0x74, 0x74, 0x6c, 0x65, 0x00, 0x2e, 0x0b, 0x20, 0x00, 0x58, 0x16,
    0x22, 0x00, 0x00, 0x01, 0x72, 0x6f, 0x75, 0x74, 0x69, 0x6e, 0x65, 0x00, 0x17, 0x60, 0x20, 0x00,
    0x58, 0x16, 0x33, 0x16, 0x45, 0x21, 0x0b, 0x76, 0x00, 0x00, 0x21, 0x09, 0x59, 0x00, 0x00, 0x01,
    0x43, 0x50, 0x55, 0x00, 0x1a, 0x77, 0x20, 0x00, 0x58, 0x16, 0x4f, 0x00, 0x00, 0x20, 0x00, 0x58,
    0x07, 0x78, 0x00, 0x00, 0x01, 0x73, 0x75, 0x62, 0x73, 0x74, 0x61, 0x6e, 0x63, 0x65, 0x00, 0x2c,
    0x34, 0x20, 0x00, 0x58, 0x16, 0x64, 0x17, 0x04, 0x01, 0x63, 0x68, 0x65, 0x6d, 0x69, 0x63, 0x61,
    0x6c, 0x00, 0x2c, 0x3b, 0x20, 0x00, 0x58, 0x16, 0x78, 0x17, 0x14, 0x04, 0x73, 0x6f, 0x6c, 0x69,
    0x64, 0x00, 0x2c, 0x42, 0x20, 0x00, 0x58, 0x17, 0x0b, 0x17, 0x1b, 0x22, 0x00, 0x58, 0x16, 0x78,
    0x16, 0x64, 0x17, 0x24, 0x22, 0x00, 0x58, 0x17, 0x0b, 0x16, 0x64, 0x17, 0x2d, 0x22, 0x00, 0x58,
    0x17, 0x0b, 0x16, 0x78, 0x00, 0x00, 0x04, 0x73, 0x68, 0x6f, 0x63, 0x6b, 0x69, 0x6e, 0x67, 0x00,
    0x2b, 0x6a, 0x20, 0x00, 0x58, 0x17, 0x36, 0x00, 0x00, 0x04, 0x6f, 0x72, 0x67, 0x61, 0x6e, 0x69,
    0x7a, 0x61, 0x74, 0x69, 0x6f, 0x6e, 0x00, 0x2c, 0x5f, 0x20, 0x00, 0x58, 0x17, 0x49, 0x00, 0x00,
    0x20, 0x00, 0x58, 0x00, 0x1b, 0x17, 0x67, 0x22, 0x00, 0x58, 0x11, 0x26, 0x00, 0x1b, 0x00, 0x00,
    0x06, 0x53, 0x75, 0x6e, 0x64, 0x61, 0x79, 0x00, 0x18, 0x01, 0x01, 0x64, 0x61, 0x79, 0x00, 0x18,
    0x20, 0x20, 0x00, 0x58, 0x17, 0x7a, 0x00, 0x00, 0x04, 0x74, 0x77, 0x65, 0x6e, 0x74, 0x79, 0x2d,
    0x66, 0x6f, 0x75, 0x72, 0x00, 0x32, 0x5f, 0x01, 0x68, 0x6f, 0x75, 0x72, 0x73, 0x00, 0x00, 0x00,
    0x22, 0x00, 0x58, 0x18, 0x08, 0x18, 0x17, 0x18, 0x31, 0x01, 0x64, 0x61, 0x79, 0x73, 0x00, 0x2d,
    0x3d, 0x21, 0x18, 0x29, 0x00, 0x00, 0x06, 0x4d, 0x6f, 0x6e, 0x64, 0x61, 0x79, 0x00, 0x18, 0x40,
    0x20, 0x00, 0x58, 0x17, 0x7a, 0x00, 0x00, 0x06, 0x54, 0x75, 0x65, 0x73, 0x64, 0x61, 0x79, 0x00,
    0x18, 0x52, 0x20, 0x00, 0x58, 0x17, 0x7a, 0x00, 0x00, 0x06, 0x57, 0x65, 0x64, 0x6e, 0x65, 0x73,
    0x64, 0x61, 0x79, 0x00, 0x18, 0x66, 0x20, 0x00, 0x58, 0x17, 0x7a, 0x00, 0x00, 0x06, 0x54, 0x68,
    0x75, 0x72, 0x73, 0x64, 0x61, 0x79, 0x00, 0x18, 0x79, 0x20, 0x00, 0x58, 0x17, 0x7a, 0x00, 0x00,
    0x06, 0x46, 0x72, 0x69, 0x64, 0x61, 0x79, 0x00, 0x19, 0x0a, 0x20, 0x00, 0x58, 0x17, 0x7a, 0x00,
    0x00, 0x06, 0x53, 0x61, 0x74, 0x75, 0x72, 0x64, 0x61, 0x79, 0x00, 0x19, 0x1d, 0x20, 0x00, 0x58,
    0x17, 0x7a, 0x00, 0x00, 0x06, 0x4a, 0x61, 0x6e, 0x75, 0x61, 0x72, 0x79, 0x00, 0x19, 0x38, 0x01,
    0x6d, 0x6f, 0x6e, 0x74, 0x68, 0x00, 0x19, 0x49, 0x20, 0x00, 0x58, 0x19, 0x2f, 0x00, 0x00, 0x01,
    0x6d, 0x6f, 0x6e, 0x74, 0x68, 0x73, 0x00, 0x2e, 0x4f, 0x21, 0x19, 0x3f, 0x00, 0x00, 0x06, 0x4d,
    0x61, 0x72, 0x63, 0x68, 0x00, 0x19, 0x57, 0x20, 0x00, 0x58, 0x19, 0x2f, 0x00, 0x00, 0x06, 0x41,
    0x70, 0x72, 0x69, 0x6c, 0x00, 0x19, 0x67, 0x20, 0x00, 0x58, 0x19, 0x2f, 0x00, 0x00, 0x06, 0x4d,
    0x61, 0x79, 0x00, 0x19, 0x75, 0x20, 0x00, 0x58, 0x19, 0x2f, 0x00, 0x00, 0x06, 0x4a, 0x75, 0x6e,
    0x65, 0x00, 0x1a, 0x04, 0x20, 0x00, 0x58, 0x19, 0x2f, 0x00, 0x00, 0x06, 0x4a, 0x75, 0x6c, 0x79,
    0x00, 0x1a, 0x13, 0x20, 0x00, 0x58, 0x19, 0x2f, 0x00, 0x00, 0x06, 0x41, 0x75, 0x67, 0x75, 0x73,
    0x74, 0x00, 0x1a, 0x24, 0x20, 0x00, 0x58, 0x19, 0x2f, 0x00, 0x00, 0x06, 0x53, 0x65, 0x70, 0x74,
    0x65, 0x6d, 0x62, 0x65, 0x72, 0x00, 0x1a, 0x38, 0x20, 0x00, 0x58, 0x19, 0x2f, 0x00, 0x00, 0x06,
    0x4f, 0x63, 0x74, 0x6f, 0x62, 0x65, 0x72, 0x00, 0x1a, 0x4a, 0x20, 0x00, 0x58, 0x19, 0x2f, 0x00,
    0x00, 0x06, 0x4e, 0x6f, 0x76, 0x65, 0x6d, 0x62, 0x65, 0x72, 0x00, 0x1a, 0x5d, 0x20, 0x00, 0x58,
    0x19, 0x2f, 0x00, 0x00, 0x06, 0x44, 0x65, 0x63, 0x65, 0x6d, 0x62, 0x65, 0x72, 0x00, 0x1a, 0x70,
    0x20, 0x00, 0x58, 0x19, 0x2f, 0x00, 0x00, 0x20, 0x00, 0x58, 0x15, 0x1e, 0x00, 0x00, 0x06, 0x46,
    0x65, 0x62, 0x72, 0x75, 0x61, 0x72, 0x79, 0x00, 0x1b, 0x0a, 0x20, 0x00, 0x58, 0x19, 0x2f, 0x00,
    0x00, 0x06, 0x57, 0x69, 0x6e, 0x74, 0x65, 0x72, 0x00, 0x1b, 0x25, 0x01, 0x73, 0x65, 0x61, 0x73,
    0x6f, 0x6e, 0x00, 0x1b, 0x56, 0x20, 0x00, 0x58, 0x1b, 0x1b, 0x1b, 0x34, 0x04, 0x63, 0x6f, 0x6c,
    0x64, 0x00, 0x1c, 0x0d, 0x20, 0x00, 0x58, 0x1b, 0x2c, 0x1b, 0x3b, 0x22, 0x00, 0x58, 0x1b, 0x2c,
    0x1b, 0x1b, 0x00, 0x00, 0x01, 0x74, 0x69, 0x6d, 0x65, 0x00, 0x00, 0x00, 0x01, 0x70, 0x65, 0x72,
    0x69, 0x6f, 0x64, 0x00, 0x00, 0x00, 0x22, 0x00, 0x58, 0x1b, 0x44, 0x1b, 0x4c, 0x1b, 0x75, 0x04,
    0x6f, 0x6e, 0x65, 0x2d, 0x66, 0x6f, 0x75, 0x72, 0x74, 0x68, 0x00, 0x33, 0x24, 0x01, 0x79, 0x65,
    0x61, 0x72, 0x00, 0x2e, 0x23, 0x22, 0x00, 0x58, 0x1b, 0x5f, 0x1b, 0x6d, 0x00, 0x00, 0x01, 0x74,
    0x65, 0x6d, 0x70, 0x65, 0x72, 0x61, 0x74, 0x75, 0x72, 0x65, 0x00, 0x00, 0x00, 0x20, 0x00, 0x58,
    0x1b, 0x7e, 0x00, 0x00, 0x06, 0x53, 0x75, 0x6d, 0x6d, 0x65, 0x72, 0x00, 0x1c, 0x1e, 0x20, 0x00,
    0x58, 0x1b, 0x1b, 0x1c, 0x2c, 0x04, 0x68, 0x6f, 0x74, 0x00, 0x1c, 0x3c, 0x20, 0x00, 0x58, 0x1c,
    0x25, 0x1c, 0x33, 0x22, 0x00, 0x58, 0x1c, 0x25, 0x1b, 0x1b, 0x00, 0x00, 0x20, 0x00, 0x58, 0x1b,
    0x7e, 0x00, 0x00, 0x06, 0x53, 0x70, 0x72, 0x69, 0x6e, 0x67, 0x00, 0x1c, 0x4d, 0x20, 0x00, 0x58,
    0x1b, 0x1b, 0x1c, 0x5c, 0x04, 0x6e, 0x69, 0x63, 0x65, 0x00, 0x1c, 0x63, 0x20, 0x00, 0x58, 0x1c,
    0x54, 0x00, 0x00, 0x20, 0x00, 0x58, 0x06, 0x59, 0x1c, 0x76, 0x04, 0x70, 0x6c, 0x65, 0x61, 0x73,
    0x69, 0x6e, 0x67, 0x00, 0x2c, 0x66, 0x20, 0x00, 0x58, 0x1c, 0x6a, 0x00, 0x00, 0x06, 0x41, 0x75,
    0x74, 0x75, 0x6d, 0x6e, 0x00, 0x1d, 0x07, 0x20, 0x00, 0x58, 0x1b, 0x1b, 0x1d, 0x0e, 0x20, 0x00,
    0x58, 0x1c, 0x54, 0x00, 0x00, 0x01, 0x63, 0x61, 0x74, 0x00, 0x1d, 0x1c, 0x20, 0x00, 0x58, 0x00,
    0x63, 0x1d, 0x23, 0x20, 0x00, 0x58, 0x03, 0x3a, 0x1d, 0x3a, 0x02, 0x73, 0x61, 0x79, 0x73, 0x00,
    0x00, 0x00, 0x04, 0x6d, 0x65, 0x6f, 0x77, 0x00, 0x1d, 0x5f, 0x20, 0x1d, 0x2a, 0x1d, 0x32, 0x1d,
    0x41, 0x20, 0x00, 0x35, 0x00, 0x3c, 0x1d, 0x48, 0x22, 0x00, 0x58, 0x11, 0x26, 0x00, 0x63, 0x1d,
    0x58, 0x01, 0x70, 0x65, 0x74, 0x00, 0x32, 0x26, 0x20, 0x00, 0x58, 0x1d, 0x51, 0x00, 0x00, 0x20,
    0x00, 0x58, 0x0d, 0x7a, 0x00, 0x00, 0x07, 0x61, 0x6e, 0x00, 0x00, 0x00, 0x01, 0x64, 0x6f, 0x67,
    0x00, 0x1d, 0x73, 0x20, 0x00, 0x58, 0x03, 0x3a, 0x1d, 0x7a, 0x20, 0x00, 0x58, 0x00, 0x63, 0x1e,
    0x09, 0x04, 0x77, 0x6f, 0x6f, 0x66, 0x00, 0x1e, 0x31, 0x20, 0x1d, 0x2a, 0x1e, 0x01, 0x1e, 0x10,
    0x20, 0x00, 0x35, 0x00, 0x3c, 0x1e, 0x17, 0x20, 0x00, 0x58, 0x1d, 0x51, 0x1e, 0x28, 0x01, 0x74,
    0x72, 0x69, 0x63, 0x6b, 0x73, 0x00, 0x00, 0x00, 0x22, 0x00, 0x35, 0x01, 0x3d, 0x1e, 0x1e, 0x00,
    0x00, 0x20, 0x00, 0x58, 0x0d, 0x7a, 0x00, 0x00, 0x01, 0x70, 0x6c, 0x61, 0x6e, 0x74, 0x00, 0x1e,
    0x41, 0x20, 0x00, 0x58, 0x00, 0x74, 0x1e, 0x51, 0x04, 0x67, 0x72, 0x65, 0x65, 0x6e, 0x00, 0x1e,
    0x61, 0x20, 0x00, 0x58, 0x1e, 0x48, 0x00, 0x00, 0x01, 0x63, 0x6f, 0x6c, 0x6f, 0x72, 0x00, 0x1e,
    0x71, 0x20, 0x00, 0x58, 0x1e, 0x58, 0x00, 0x00, 0x04, 0x6c, 0x69, 0x67, 0x68, 0x74, 0x00, 0x21,
    0x09, 0x20, 0x00, 0x58, 0x1e, 0x68, 0x1e, 0x78, 0x22, 0x00, 0x58, 0x0d, 0x55, 0x1e, 0x68, 0x00,
    0x00, 0x01, 0x74, 0x72, 0x65, 0x65, 0x00, 0x1f, 0x09, 0x20, 0x00, 0x58, 0x1e, 0x38, 0x1f, 0x1a,
    0x01, 0x6c, 0x65, 0x61, 0x76, 0x65, 0x73, 0x00, 0x00, 0x00, 0x20, 0x02, 0x5a, 0x1f, 0x10, 0x1f,
    0x2a, 0x01, 0x72, 0x6f, 0x6f, 0x74, 0x73, 0x00, 0x00, 0x00, 0x20, 0x02, 0x5a, 0x1f, 0x21, 0x00,
    0x00, 0x07, 0x74, 0x68, 0x65, 0x00, 0x00, 0x00, 0x01, 0x73, 0x6b, 0x79, 0x00, 0x1f, 0x47, 0x04,
    0x62, 0x6c, 0x75, 0x65, 0x00, 0x1f, 0x5f, 0x20, 0x00, 0x58, 0x1f, 0x3f, 0x1f, 0x58, 0x04, 0x63,
    0x6c, 0x6f, 0x75, 0x64, 0x73, 0x00, 0x1f, 0x6f, 0x20, 0x02, 0x5a, 0x1f, 0x4e, 0x00, 0x00, 0x20,
    0x00, 0x58, 0x1e, 0x58, 0x00, 0x00, 0x04, 0x77, 0x68, 0x69, 0x74, 0x65, 0x00, 0x20, 0x47, 0x20,
    0x00, 0x4c, 0x1f, 0x66, 0x1f, 0x7d, 0x04, 0x77, 0x65, 0x74, 0x00, 0x20, 0x7d, 0x20, 0x00, 0x4c,
    0x1f, 0x76, 0x20, 0x0d, 0x01, 0x63, 0x6c, 0x6f, 0x75, 0x64, 0x00, 0x21, 0x04, 0x21, 0x20, 0x04,
    0x20, 0x26, 0x02, 0x63, 0x6f, 0x6e, 0x74, 0x61, 0x69, 0x6e, 0x00, 0x00, 0x00, 0x04, 0x77, 0x61,
    0x74, 0x65, 0x72, 0x00, 0x25, 0x33, 0x20, 0x20, 0x12, 0x20, 0x1d, 0x20, 0x40, 0x02, 0x70, 0x72,
    0x6f, 0x64, 0x75, 0x63, 0x65, 0x00, 0x00, 0x00, 0x06, 0x72, 0x61, 0x69, 0x6e, 0x00, 0x30, 0x03,
    0x20, 0x20, 0x2d, 0x20, 0x38, 0x00, 0x00, 0x20, 0x00, 0x58, 0x1e, 0x58, 0x20, 0x6b, 0x02, 0x63,
    0x6f, 0x6e, 0x74, 0x61, 0x69, 0x6e, 0x73, 0x00, 0x00, 0x00, 0x04, 0x61, 0x6c, 0x6c, 0x00, 0x00,
    0x00, 0x01, 0x63, 0x6f, 0x6c, 0x6f, 0x72, 0x73, 0x00, 0x00, 0x00, 0x22, 0x20, 0x4e, 0x20, 0x5a,
    0x20, 0x61, 0x00, 0x00, 0x04, 0x6d, 0x6f, 0x69, 0x73, 0x74, 0x00, 0x2d, 0x04, 0x20, 0x00, 0x58,
    0x20, 0x74, 0x00, 0x00, 0x21, 0x1f, 0x4e, 0x00, 0x00, 0x20, 0x00, 0x58, 0x0d, 0x55, 0x21, 0x1d,
    0x06, 0x72, 0x61, 0x64, 0x69, 0x61, 0x74, 0x69, 0x6f, 0x6e, 0x00, 0x00, 0x00, 0x20, 0x00, 0x58,
    0x21, 0x10, 0x21, 0x24, 0x22, 0x00, 0x58, 0x0d, 0x55, 0x21, 0x10, 0x21, 0x40, 0x04, 0x65, 0x6c,
    0x65, 0x63, 0x74, 0x72, 0x6f, 0x6d, 0x61, 0x67, 0x6e, 0x65, 0x74, 0x69, 0x63, 0x00, 0x00, 0x00,
    0x22, 0x00, 0x58, 0x21, 0x2d, 0x21, 0x10, 0x00, 0x00, 0x04, 0x73, 0x65, 0x65, 0x6e, 0x00, 0x2b,
    0x2b, 0x20, 0x00, 0x58, 0x21, 0x49, 0x00, 0x00, 0x01, 0x63, 0x6f, 0x77, 0x00, 0x21, 0x5f, 0x20,
    0x00, 0x58, 0x00, 0x63, 0x21, 0x6d, 0x04, 0x6d, 0x6f, 0x6f, 0x00, 0x22, 0x41, 0x20, 0x1d, 0x2a,
    0x21, 0x66, 0x21, 0x74, 0x20, 0x00, 0x35, 0x00, 0x3c, 0x22, 0x0c, 0x02, 0x65, 0x61, 0x74, 0x73,
    0x00, 0x22, 0x4f, 0x04, 0x67, 0x72, 0x61, 0x73, 0x73, 0x00, 0x22, 0x54, 0x20, 0x21, 0x7b, 0x22,
    0x03, 0x22, 0x1a, 0x04, 0x62, 0x69, 0x67, 0x00, 0x27, 0x27, 0x22, 0x00, 0x58, 0x22, 0x13, 0x00,
    0x63, 0x22, 0x2b, 0x01, 0x66, 0x61, 0x72, 0x6d, 0x00, 0x32, 0x3f, 0x22, 0x00, 0x58, 0x22, 0x23,
    0x03, 0x3a, 0x22, 0x3c, 0x01, 0x63, 0x6f, 0x77, 0x73, 0x00, 0x33, 0x03, 0x21, 0x22, 0x34, 0x00,
    0x00, 0x20, 0x00, 0x58, 0x0d, 0x7a, 0x00, 0x00, 0x02, 0x65, 0x61, 0x74, 0x00, 0x22, 0x6b, 0x21,
    0x22, 0x48, 0x00, 0x00, 0x20, 0x00, 0x58, 0x1e, 0x48, 0x22, 0x5b, 0x20, 0x00, 0x58, 0x1e, 0x38,
    0x22, 0x62, 0x22, 0x00, 0x58, 0x1e, 0x48, 0x1e, 0x38, 0x00, 0x00, 0x21, 0x21, 0x7b, 0x00, 0x00,
    0x06, 0x4d, 0x65, 0x72, 0x63, 0x75, 0x72, 0x79, 0x00, 0x23, 0x05, 0x01, 0x70, 0x6c, 0x61, 0x6e,
    0x65, 0x74, 0x00, 0x23, 0x25, 0x20, 0x00, 0x58, 0x22, 0x7b, 0x23, 0x0c, 0x20, 0x00, 0x58, 0x1c,
    0x25, 0x23, 0x13, 0x22, 0x00, 0x58, 0x1c, 0x25, 0x22, 0x7b, 0x00, 0x00, 0x01, 0x77, 0x6f, 0x72,
    0x6c, 0x64, 0x00, 0x23, 0x7c, 0x20, 0x00, 0x58, 0x23, 0x1c, 0x23, 0x36, 0x01, 0x73, 0x70, 0x68,
    0x65, 0x72, 0x65, 0x00, 0x24, 0x0c, 0x20, 0x00, 0x58, 0x23, 0x2c, 0x23, 0x3d, 0x22, 0x00, 0x58,
    0x22, 0x13, 0x23, 0x2c, 0x23, 0x5b, 0x04, 0x63, 0x65, 0x6c, 0x65, 0x73, 0x74, 0x69, 0x61, 0x6c,
    0x00, 0x00, 0x00, 0x01, 0x62, 0x6f, 0x64, 0x79, 0x00, 0x00, 0x00, 0x22, 0x00, 0x58, 0x23, 0x46,
    0x23, 0x53, 0x23, 0x75, 0x02, 0x6f, 0x72, 0x62, 0x69, 0x74, 0x73, 0x00, 0x00, 0x00, 0x01, 0x73,
    0x75, 0x6e, 0x00, 0x35, 0x39, 0x20, 0x23, 0x64, 0x23, 0x6e, 0x00, 0x00, 0x20, 0x00, 0x58, 0x22,
    0x7b, 0x00, 0x00, 0x04, 0x72, 0x6f, 0x75, 0x6e, 0x64, 0x00, 0x24, 0x32, 0x20, 0x00, 0x58, 0x24,
    0x03, 0x24, 0x1d, 0x01, 0x6f, 0x62, 0x6a, 0x65, 0x63, 0x74, 0x00, 0x29, 0x4e, 0x22, 0x00, 0x58,
    0x24, 0x03, 0x24, 0x13, 0x00, 0x00, 0x04, 0x63, 0x69, 0x72, 0x63, 0x75, 0x6c, 0x61, 0x72, 0x00,
    0x2b, 0x24, 0x20, 0x00, 0x58, 0x24, 0x26, 0x00, 0x00, 0x06, 0x4d, 0x61, 0x72, 0x73, 0x00, 0x24,
    0x41, 0x20, 0x00, 0x58, 0x22, 0x7b, 0x24, 0x4f, 0x04, 0x72, 0x65, 0x64, 0x00, 0x24, 0x66, 0x20,
    0x00, 0x58, 0x24, 0x48, 0x24, 0x56, 0x20, 0x00, 0x58, 0x1b, 0x2c, 0x24, 0x5d, 0x22, 0x00, 0x58,
    0x24, 0x48, 0x22, 0x7b, 0x00, 0x00, 0x20, 0x00, 0x58, 0x1e, 0x58, 0x00, 0x00, 0x06, 0x45, 0x61,
    0x72, 0x74, 0x68, 0x00, 0x24, 0x76, 0x20, 0x00, 0x58, 0x22, 0x7b, 0x24, 0x7d, 0x20, 0x02, 0x5a,
    0x20, 0x1d, 0x25, 0x0e, 0x01, 0x6f, 0x63, 0x65, 0x61, 0x6e, 0x73, 0x00, 0x26, 0x04, 0x20, 0x02,
    0x5a, 0x25, 0x04, 0x25, 0x1d, 0x06, 0x6c, 0x61, 0x6e, 0x64, 0x00, 0x26, 0x21, 0x20, 0x02, 0x5a,
    0x25, 0x15, 0x25, 0x24, 0x20, 0x00, 0x58, 0x23, 0x1c, 0x00, 0x00, 0x04, 0x73, 0x6f, 0x66, 0x74,
    0x00, 0x00, 0x00, 0x20, 0x00, 0x58, 0x25, 0x2b, 0x25, 0x3a, 0x20, 0x00, 0x58, 0x1f, 0x76, 0x25,
    0x4b, 0x01, 0x6c, 0x69, 0x71, 0x75, 0x69, 0x64, 0x00, 0x2e, 0x04, 0x20, 0x00, 0x58, 0x25, 0x41,
    0x25, 0x59, 0x06, 0x48, 0x32, 0x4f, 0x00, 0x30, 0x35, 0x20, 0x00, 0x58, 0x25, 0x52, 0x25, 0x6c,
    0x06, 0x68, 0x79, 0x64, 0x72, 0x6f, 0x67, 0x65, 0x6e, 0x00, 0x30, 0x3c, 0x20, 0x20, 0x4e, 0x25,
    0x60, 0x25, 0x7d, 0x06, 0x6f, 0x78, 0x79, 0x67, 0x65, 0x6e, 0x00, 0x30, 0x43, 0x20, 0x20, 0x4e,
    0x25, 0x73, 0x00, 0x00, 0x20, 0x08, 0x2b, 0x20, 0x1d, 0x26, 0x14, 0x01, 0x6f, 0x63, 0x65, 0x61,
    0x6e, 0x00, 0x26, 0x36, 0x21, 0x26, 0x0b, 0x00, 0x00, 0x04, 0x68, 0x61, 0x72, 0x64, 0x00, 0x26,
    0x4b, 0x20, 0x00, 0x58, 0x26, 0x19, 0x26, 0x2f, 0x04, 0x64, 0x72, 0x79, 0x00, 0x00, 0x00, 0x20,
    0x00, 0x58, 0x26, 0x28, 0x00, 0x00, 0x20, 0x00, 0x58, 0x1f, 0x76, 0x26, 0x3d, 0x21, 0x25, 0x04,
    0x00, 0x00, 0x04, 0x72, 0x69, 0x67, 0x69, 0x64, 0x00, 0x26, 0x60, 0x20, 0x00, 0x58, 0x26, 0x42,
    0x00, 0x00, 0x04, 0x69, 0x6e, 0x66, 0x6c, 0x65, 0x78, 0x69, 0x62, 0x6c, 0x65, 0x00, 0x2b, 0x32,
    0x20, 0x00, 0x58, 0x26, 0x52, 0x00, 0x00, 0x06, 0x56, 0x65, 0x6e, 0x75, 0x73, 0x00, 0x26, 0x70,
    0x20, 0x00, 0x58, 0x22, 0x7b, 0x26, 0x77, 0x20, 0x00, 0x58, 0x1c, 0x25, 0x00, 0x00, 0x06, 0x4a,
    0x75, 0x70, 0x69, 0x74, 0x65, 0x72, 0x00, 0x27, 0x09, 0x20, 0x00, 0x58, 0x22, 0x7b, 0x27, 0x10,
    0x20, 0x00, 0x58, 0x1b, 0x2c, 0x27, 0x17, 0x20, 0x00, 0x58, 0x22, 0x13, 0x00, 0x00, 0x04, 0x6c,
    0x61, 0x72, 0x67, 0x65, 0x00, 0x27, 0x2e, 0x20, 0x00, 0x58, 0x27, 0x1e, 0x00, 0x00, 0x20, 0x00,
    0x58, 0x22, 0x13, 0x00, 0x00, 0x06, 0x53, 0x61, 0x74, 0x75, 0x72, 0x6e, 0x00, 0x27, 0x3f, 0x20,
    0x00, 0x58, 0x22, 0x7b, 0x27, 0x46, 0x20, 0x00, 0x58, 0x1b, 0x2c, 0x27, 0x63, 0x04, 0x62, 0x65,
    0x61, 0x75, 0x74, 0x69, 0x66, 0x75, 0x6c, 0x00, 0x00, 0x00, 0x01, 0x72, 0x69, 0x6e, 0x67, 0x73,
    0x00, 0x2f, 0x61, 0x22, 0x02, 0x5a, 0x27, 0x4d, 0x27, 0x5a, 0x00, 0x00, 0x06, 0x4e, 0x65, 0x70,
    0x74, 0x75, 0x6e, 0x65, 0x00, 0x27, 0x77, 0x20, 0x00, 0x58, 0x22, 0x7b, 0x27, 0x7e, 0x20, 0x00,
    0x58, 0x1b, 0x2c, 0x00, 0x00, 0x06, 0x55, 0x72, 0x61, 0x6e, 0x75, 0x73, 0x00, 0x28, 0x0f, 0x20,
    0x00, 0x58, 0x22, 0x7b, 0x28, 0x16, 0x20, 0x00, 0x58, 0x1b, 0x2c, 0x00, 0x00, 0x06, 0x50, 0x6c,
    0x75, 0x74, 0x6f, 0x00, 0x28, 0x2d, 0x02, 0x77, 0x61, 0x73, 0x00, 0x00, 0x00, 0x20, 0x28, 0x26,
    0x22, 0x7b, 0x28, 0x34, 0x20, 0x00, 0x58, 0x1b, 0x2c, 0x28, 0x44, 0x04, 0x64, 0x77, 0x61, 0x72,
    0x66, 0x00, 0x00, 0x00, 0x22, 0x00, 0x58, 0x28, 0x3b, 0x22, 0x7b, 0x00, 0x00, 0x01, 0x69, 0x6e,
    0x74, 0x65, 0x72, 0x72, 0x75, 0x70, 0x74, 0x00, 0x28, 0x61, 0x04, 0x62, 0x61, 0x64, 0x00, 0x29,
    0x47, 0x20, 0x00, 0x58, 0x28, 0x5a, 0x28, 0x77, 0x04, 0x69, 0x6e, 0x65, 0x66, 0x66, 0x69, 0x63,
    0x69, 0x65, 0x6e, 0x74, 0x00, 0x00, 0x00, 0x20, 0x00, 0x58, 0x28, 0x68, 0x29, 0x09, 0x02, 0x64,
    0x69, 0x73, 0x72, 0x75, 0x70, 0x74, 0x00, 0x00, 0x00, 0x20, 0x00, 0x35, 0x28, 0x7e, 0x29, 0x1e,
    0x04, 0x64, 0x69, 0x73, 0x72, 0x75, 0x70, 0x74, 0x69, 0x76, 0x65, 0x00, 0x00, 0x00, 0x20, 0x00,
    0x58, 0x29, 0x10, 0x29, 0x34, 0x04, 0x63, 0x6f, 0x6d, 0x70, 0x6c, 0x69, 0x63, 0x61, 0x74, 0x65,
    0x64, 0x00, 0x00, 0x00, 0x20, 0x00, 0x58, 0x29, 0x25, 0x00, 0x00, 0x04, 0x68, 0x6f, 0x72, 0x72,
    0x69, 0x62, 0x6c, 0x65, 0x00, 0x2d, 0x0b, 0x20, 0x00, 0x58, 0x29, 0x3b, 0x00, 0x00, 0x20, 0x02,
    0x5a, 0x0b, 0x76, 0x00, 0x00, 0x06, 0x45, 0x6e, 0x67, 0x6c, 0x69, 0x73, 0x68, 0x00, 0x29, 0x60,
    0x20, 0x00, 0x58, 0x0a, 0x13, 0x29, 0x67, 0x22, 0x00, 0x58, 0x00, 0x0b, 0x0a, 0x13, 0x00, 0x00,
    0x01, 0x6e, 0x6f, 0x75, 0x6e, 0x00, 0x00, 0x00, 0x02, 0x73, 0x61, 0x79, 0x00, 0x00, 0x00, 0x01,
    0x53, 0x75, 0x6e, 0x00, 0x2a, 0x0e, 0x01, 0x73, 0x74, 0x61, 0x72, 0x00, 0x2a, 0x30, 0x20, 0x00,
    0x58, 0x2a, 0x06, 0x2a, 0x15, 0x20, 0x00, 0x58, 0x1c, 0x25, 0x2a, 0x27, 0x04, 0x6e, 0x65, 0x61,
    0x72, 0x65, 0x73, 0x74, 0x00, 0x30, 0x72, 0x22, 0x00, 0x58, 0x2a, 0x1c, 0x2a, 0x06, 0x00, 0x00,
    0x20, 0x00, 0x58, 0x1c, 0x25, 0x2a, 0x41, 0x04, 0x62, 0x72, 0x69, 0x67, 0x68, 0x74, 0x00, 0x00,
    0x00, 0x20, 0x00, 0x58, 0x2a, 0x37, 0x2a, 0x51, 0x02, 0x65, 0x6d, 0x69, 0x74, 0x73, 0x00, 0x00,
    0x00, 0x20, 0x2a, 0x48, 0x1e, 0x68, 0x2a, 0x60, 0x04, 0x76, 0x65, 0x72, 0x79, 0x00, 0x00, 0x00,
    0x22, 0x00, 0x58, 0x2a, 0x58, 0x1c, 0x25, 0x2a, 0x69, 0x22, 0x00, 0x58, 0x23, 0x46, 0x23, 0x53,
    0x00, 0x00, 0x01, 0x73, 0x79, 0x6d, 0x62, 0x6f, 0x6c, 0x73, 0x00, 0x2b, 0x05, 0x01, 0x74, 0x61,
    0x67, 0x73, 0x00, 0x2b, 0x16, 0x20, 0x00, 0x4c, 0x2a, 0x7d, 0x00, 0x00, 0x01, 0x6c, 0x61, 0x62,
    0x65, 0x6c, 0x73, 0x00, 0x2b, 0x1d, 0x20, 0x00, 0x4c, 0x2b, 0x0c, 0x00, 0x00, 0x20, 0x00, 0x4c,
    0x2a, 0x7d, 0x00, 0x00, 0x20, 0x00, 0x58, 0x24, 0x03, 0x00, 0x00, 0x20, 0x00, 0x58, 0x0d, 0x55,
    0x00, 0x00, 0x20, 0x00, 0x58, 0x26, 0x19, 0x00, 0x00, 0x02, 0x6c, 0x69, 0x6b, 0x65, 0x00, 0x00,
    0x00, 0x02, 0x6d, 0x61, 0x64, 0x65, 0x00, 0x00, 0x00, 0x01, 0x61, 0x77, 0x61, 0x72, 0x65, 0x00,
    0x00, 0x00, 0x20, 0x00, 0x58, 0x01, 0x0c, 0x2b, 0x59, 0x21, 0x01, 0x0c, 0x00, 0x00, 0x20, 0x00,
    0x58, 0x00, 0x74, 0x00, 0x00, 0x21, 0x02, 0x61, 0x00, 0x00, 0x20, 0x00, 0x58, 0x08, 0x08, 0x00,
    0x00, 0x20, 0x00, 0x58, 0x09, 0x47, 0x2b, 0x78, 0x20, 0x00, 0x58, 0x04, 0x3e, 0x2b, 0x7f, 0x20,
    0x00, 0x58, 0x0b, 0x54, 0x00, 0x00, 0x21, 0x08, 0x68, 0x00, 0x00, 0x20, 0x00, 0x58, 0x1c, 0x54,
    0x00, 0x00, 0x20, 0x00, 0x58, 0x09, 0x6a, 0x2c, 0x26, 0x06, 0x6b, 0x6e, 0x6f, 0x77, 0x6c, 0x65,
    0x64, 0x67, 0x65, 0x00, 0x2c, 0x2d, 0x20, 0x00, 0x58, 0x2c, 0x19, 0x00, 0x00, 0x20, 0x00, 0x58,
    0x12, 0x04, 0x00, 0x00, 0x20, 0x00, 0x58, 0x07, 0x78, 0x00, 0x00, 0x20, 0x00, 0x58, 0x16, 0x64,
    0x00, 0x00, 0x20, 0x00, 0x58, 0x26, 0x42, 0x2c, 0x51, 0x04, 0x66, 0x69, 0x72, 0x6d, 0x00, 0x30,
    0x4a, 0x20, 0x00, 0x58, 0x2c, 0x49, 0x00, 0x00, 0x20, 0x00, 0x4c, 0x0c, 0x6c, 0x00, 0x00, 0x20,
    0x00, 0x58, 0x16, 0x07, 0x00, 0x00, 0x20, 0x00, 0x58, 0x1c, 0x54, 0x00, 0x00, 0x04, 0x68, 0x65,
    0x61, 0x72, 0x64, 0x00, 0x2c, 0x7d, 0x20, 0x00, 0x58, 0x2c, 0x6d, 0x00, 0x00, 0x20, 0x00, 0x58,
    0x10, 0x1a, 0x00, 0x00, 0x20, 0x00, 0x58, 0x1f, 0x76, 0x00, 0x00, 0x20, 0x00, 0x58, 0x28, 0x5a,
    0x00, 0x00, 0x04, 0x73, 0x65, 0x76, 0x65, 0x6e, 0x00, 0x2d, 0x25, 0x01, 0x6e, 0x75, 0x6d, 0x62,
    0x65, 0x72, 0x00, 0x00, 0x00, 0x20, 0x00, 0x58, 0x2d, 0x1b, 0x00, 0x00, 0x01, 0x77, 0x65, 0x65,
    0x6b, 0x00, 0x2d, 0x34, 0x22, 0x00, 0x58, 0x2d, 0x12, 0x18, 0x29, 0x00, 0x00, 0x21, 0x17, 0x7a,
    0x00, 0x00, 0x04, 0x6f, 0x6e, 0x65, 0x00, 0x2d, 0x49, 0x20, 0x00, 0x58, 0x2d, 0x1b, 0x2d, 0x62,
    0x04, 0x73, 0x69, 0x6e, 0x67, 0x6c, 0x65, 0x00, 0x00, 0x00, 0x01, 0x69, 0x74, 0x65, 0x6d, 0x00,
    0x2d, 0x74, 0x22, 0x00, 0x58, 0x2d, 0x50, 0x2d, 0x5a, 0x00, 0x00, 0x01, 0x74, 0x68, 0x69, 0x6e,
    0x67, 0x00, 0x00, 0x00, 0x20, 0x00, 0x58, 0x2d, 0x6b, 0x00, 0x00, 0x04, 0x66, 0x6c, 0x75, 0x69,
    0x64, 0x00, 0x00, 0x00, 0x20, 0x00, 0x58, 0x2d, 0x7b, 0x00, 0x00, 0x20, 0x00, 0x58, 0x11, 0x26,
    0x00, 0x00, 0x20, 0x00, 0x58, 0x2d, 0x1b, 0x00, 0x00, 0x04, 0x74, 0x77, 0x65, 0x6c, 0x76, 0x65,
    0x00, 0x2e, 0x48, 0x22, 0x02, 0x5a, 0x2e, 0x19, 0x19, 0x3f, 0x2e, 0x3f, 0x04, 0x66, 0x6f, 0x75,
    0x72, 0x00, 0x2e, 0x54, 0x01, 0x73, 0x65, 0x61, 0x73, 0x6f, 0x6e, 0x73, 0x00, 0x00, 0x00, 0x22,
    0x02, 0x5a, 0x2e, 0x2c, 0x2e, 0x34, 0x00, 0x00, 0x20, 0x00, 0x58, 0x2d, 0x1b, 0x00, 0x00, 0x21,
    0x19, 0x2f, 0x00, 0x00, 0x20, 0x00, 0x58, 0x2d, 0x1b, 0x00, 0x00, 0x04, 0x74, 0x77, 0x6f, 0x00,
    0x2e, 0x62, 0x20, 0x00, 0x58, 0x2d, 0x1b, 0x00, 0x00, 0x04, 0x74, 0x68, 0x72, 0x65, 0x65, 0x00,
    0x2e, 0x72, 0x20, 0x00, 0x58, 0x2d, 0x1b, 0x00, 0x00, 0x04, 0x66, 0x69, 0x76, 0x65, 0x00, 0x2f,
    0x01, 0x20, 0x00, 0x58, 0x2d, 0x1b, 0x00, 0x00, 0x04, 0x73, 0x69, 0x78, 0x00, 0x2f, 0x0f, 0x20,
    0x00, 0x58, 0x2d, 0x1b, 0x00, 0x00, 0x04, 0x6e, 0x69, 0x6e, 0x65, 0x00, 0x2f, 0x1e, 0x20, 0x00,
    0x58, 0x2d, 0x1b, 0x00, 0x00, 0x04, 0x74, 0x65, 0x6e, 0x00, 0x2f, 0x2c, 0x20, 0x00, 0x58, 0x2d,
    0x1b, 0x00, 0x00, 0x04, 0x65, 0x6c, 0x65, 0x76, 0x65, 0x6e, 0x00, 0x2f, 0x3d, 0x20, 0x00, 0x58,
    0x2d, 0x1b, 0x00, 0x00, 0x02, 0x6c, 0x65, 0x61, 0x72, 0x6e, 0x73, 0x00, 0x35, 0x34, 0x21, 0x2f,
    0x44, 0x00, 0x00, 0x01, 0x73, 0x74, 0x72, 0x75, 0x63, 0x74, 0x75, 0x72, 0x65, 0x73, 0x00, 0x00,
    0x00, 0x22, 0x00, 0x4c, 0x24, 0x26, 0x2f, 0x53, 0x00, 0x00, 0x06, 0x43, 0x00, 0x2f, 0x6f, 0x22,
    0x00, 0x58, 0x0c, 0x37, 0x0a, 0x13, 0x00, 0x00, 0x04, 0x66, 0x61, 0x6c, 0x6c, 0x69, 0x6e, 0x67,
    0x00, 0x30, 0x18, 0x22, 0x00, 0x58, 0x2f, 0x78, 0x20, 0x1d, 0x00, 0x00, 0x04, 0x64, 0x72, 0x6f,
    0x70, 0x70, 0x69, 0x6e, 0x67, 0x00, 0x30, 0x51, 0x20, 0x00, 0x58, 0x30, 0x0c, 0x00, 0x00, 0x04,
    0x72, 0x65, 0x73, 0x70, 0x69, 0x72, 0x61, 0x74, 0x69, 0x6f, 0x6e, 0x00, 0x31, 0x00, 0x20, 0x00,
    0x58, 0x30, 0x1f, 0x00, 0x00, 0x20, 0x00, 0x58, 0x20, 0x1d, 0x00, 0x00, 0x20, 0x00, 0x58, 0x0e,
    0x59, 0x00, 0x00, 0x20, 0x00, 0x58, 0x0e, 0x59, 0x00, 0x00, 0x20, 0x00, 0x58, 0x26, 0x42, 0x00,
    0x00, 0x20, 0x00, 0x58, 0x2f, 0x78, 0x00, 0x00, 0x01, 0x72, 0x6f, 0x63, 0x6b, 0x00, 0x30, 0x60,
    0x20, 0x00, 0x58, 0x0e, 0x47, 0x00, 0x00, 0x04, 0x63, 0x6c, 0x6f, 0x73, 0x65, 0x73, 0x74, 0x00,
    0x30, 0x79, 0x20, 0x00, 0x58, 0x30, 0x67, 0x00, 0x00, 0x20, 0x00, 0x58, 0x2a, 0x1c, 0x00, 0x00,
    0x20, 0x00, 0x58, 0x03, 0x69, 0x00, 0x00, 0x01, 0x74, 0x69, 0x74, 0x6c, 0x65, 0x73, 0x00, 0x36,
    0x04, 0x20, 0x00, 0x4c, 0x31, 0x07, 0x00, 0x00, 0x01, 0x68, 0x6f, 0x75, 0x72, 0x00, 0x31, 0x34,
    0x04, 0x73, 0x69, 0x78, 0x74, 0x79, 0x00, 0x31, 0x3d, 0x01, 0x6d, 0x69, 0x6e, 0x75, 0x74, 0x65,
    0x73, 0x00, 0x31, 0x4e, 0x22, 0x00, 0x58, 0x31, 0x20, 0x31, 0x29, 0x00, 0x00, 0x20, 0x00, 0x58,
    0x2d, 0x1b, 0x00, 0x00, 0x01, 0x6d, 0x69, 0x6e, 0x75, 0x74, 0x65, 0x00, 0x31, 0x7d, 0x21, 0x31,
    0x44, 0x00, 0x00, 0x22, 0x00, 0x58, 0x00, 0x74, 0x0f, 0x6f, 0x00, 0x00, 0x21, 0x04, 0x69, 0x00,
    0x00, 0x04, 0x63, 0x6f, 0x6d, 0x62, 0x69, 0x6e, 0x65, 0x64, 0x00, 0x35, 0x61, 0x20, 0x00, 0x58,
    0x31, 0x61, 0x00, 0x00, 0x22, 0x00, 0x58, 0x09, 0x15, 0x08, 0x7a, 0x00, 0x00, 0x21, 0x31, 0x29,
    0x32, 0x0d, 0x01, 0x73, 0x65, 0x63, 0x6f, 0x6e, 0x64, 0x73, 0x00, 0x00, 0x00, 0x22, 0x02, 0x5a,
    0x31, 0x20, 0x32, 0x02, 0x00, 0x00, 0x04, 0x64, 0x6f, 0x6d, 0x65, 0x73, 0x74, 0x69, 0x63, 0x61,
    0x74, 0x65, 0x64, 0x00, 0x33, 0x11, 0x22, 0x00, 0x58, 0x32, 0x16, 0x03, 0x3a, 0x00, 0x00, 0x04,
    0x61, 0x67, 0x72, 0x69, 0x63, 0x75, 0x6c, 0x74, 0x75, 0x72, 0x61, 0x6c, 0x00, 0x00, 0x00, 0x22,
    0x00, 0x58, 0x32, 0x2f, 0x13, 0x5a, 0x32, 0x51, 0x01, 0x63, 0x72, 0x6f, 0x70, 0x73, 0x00, 0x00,
    0x00, 0x20, 0x02, 0x5a, 0x32, 0x48, 0x32, 0x58, 0x20, 0x02, 0x5a, 0x22, 0x34, 0x00, 0x00, 0x20,
    0x00, 0x58, 0x2d, 0x1b, 0x00, 0x00, 0x01, 0x61, 0x64, 0x6a, 0x65, 0x63, 0x74, 0x69, 0x76, 0x65,
    0x00, 0x00, 0x00, 0x01, 0x6e, 0x61, 0x6d, 0x65, 0x00, 0x00, 0x00, 0x01, 0x76, 0x65, 0x72, 0x62,
    0x00, 0x00, 0x00, 0x21, 0x21, 0x58, 0x00, 0x00, 0x04, 0x74, 0x61, 0x6d, 0x65, 0x64, 0x00, 0x33,
    0x18, 0x20, 0x00, 0x58, 0x33, 0x08, 0x00, 0x00, 0x20, 0x00, 0x58, 0x32, 0x16, 0x00, 0x00, 0x21,
    0x0a, 0x6a, 0x00, 0x00, 0x20, 0x00, 0x58, 0x2d, 0x1b, 0x00, 0x00, 0x01, 0x62, 0x79, 0x74, 0x65,
    0x00, 0x33, 0x3b, 0x01, 0x62, 0x69, 0x74, 0x73, 0x00, 0x00, 0x00, 0x22, 0x00, 0x58, 0x05, 0x34,
    0x33, 0x33, 0x33, 0x4d, 0x01, 0x62, 0x79, 0x74, 0x65, 0x73, 0x00, 0x34, 0x5c, 0x21, 0x33, 0x44,
    0x00, 0x00, 0x01, 0x6e, 0x79, 0x62, 0x62, 0x6c, 0x65, 0x00, 0x33, 0x5c, 0x22, 0x00, 0x58, 0x2e,
    0x2c, 0x33, 0x33, 0x00, 0x00, 0x01, 0x77, 0x6f, 0x72, 0x64, 0x00, 0x33, 0x78, 0x04, 0x73, 0x69,
    0x78, 0x74, 0x65, 0x65, 0x6e, 0x00, 0x34, 0x18, 0x22, 0x00, 0x58, 0x33, 0x6d, 0x33, 0x33, 0x34,
    0x0a, 0x01, 0x77, 0x6f, 0x72, 0x64, 0x73, 0x00, 0x34, 0x57, 0x21, 0x34, 0x01, 0x34, 0x0f, 0x22,
    0x00, 0x58, 0x2e, 0x5b, 0x33, 0x44, 0x00, 0x00, 0x20, 0x00, 0x58, 0x2d, 0x1b, 0x00, 0x00, 0x01,
    0x6c, 0x6f, 0x6e, 0x67, 0x00, 0x34, 0x35, 0x04, 0x74, 0x68, 0x69, 0x72, 0x74, 0x79, 0x2d, 0x74,
    0x77, 0x6f, 0x00, 0x34, 0x50, 0x22, 0x00, 0x58, 0x34, 0x27, 0x33, 0x33, 0x34, 0x3e, 0x22, 0x00,
    0x58, 0x2e, 0x5b, 0x34, 0x01, 0x34, 0x47, 0x22, 0x00, 0x58, 0x2e, 0x2c, 0x33, 0x44, 0x00, 0x00,
    0x20, 0x00, 0x58, 0x2d, 0x1b, 0x00, 0x00, 0x21, 0x33, 0x65, 0x00, 0x00, 0x21, 0x33, 0x2b, 0x00,
    0x00, 0x01, 0x62, 0x69, 0x74, 0x00, 0x34, 0x7b, 0x04, 0x62, 0x69, 0x6e, 0x61, 0x72, 0x79, 0x00,
    0x00, 0x00, 0x01, 0x64, 0x69, 0x67, 0x69, 0x74, 0x00, 0x35, 0x12, 0x22, 0x00, 0x58, 0x34, 0x68,
    0x34, 0x72, 0x00, 0x00, 0x02, 0x72, 0x65, 0x70, 0x72, 0x65, 0x73, 0x65, 0x6e, 0x74, 0x73, 0x00,
    0x00, 0x00, 0x20, 0x35, 0x04, 0x2d, 0x1b, 0x35, 0x23, 0x01, 0x73, 0x79, 0x6d, 0x62, 0x6f, 0x6c,
    0x00, 0x00, 0x00, 0x20, 0x00, 0x58, 0x35, 0x19, 0x00, 0x00, 0x21, 0x0c, 0x0f, 0x00, 0x00, 0x21,
    0x00, 0x0b, 0x00, 0x00, 0x21, 0x01, 0x3d, 0x00, 0x00, 0x20, 0x00, 0x58, 0x2a, 0x06, 0x00, 0x00,
    0x01, 0x6d, 0x6f, 0x6f, 0x6e, 0x00, 0x35, 0x48, 0x20, 0x23, 0x64, 0x22, 0x7b, 0x00, 0x00, 0x04,
    0x69, 0x6e, 0x74, 0x65, 0x72, 0x70, 0x72, 0x65, 0x74, 0x00, 0x00, 0x00, 0x21, 0x06, 0x3c, 0x00,
    0x00, 0x20, 0x00, 0x58, 0x09, 0x15, 0x00, 0x00, 0x04, 0x73, 0x6c, 0x65, 0x65, 0x70, 0x69, 0x6e,
    0x67, 0x00, 0x35, 0x7b, 0x20, 0x00, 0x58, 0x35, 0x68, 0x00, 0x00, 0x22, 0x00, 0x58, 0x01, 0x1c,
    0x03, 0x12, 0x00, 0x00, 0x20, 0x00, 0x4c, 0x11, 0x61, 0x00, 0x00};

#ifndef CSPIN_FLAG
char dummy[] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
#endif

/* INLINE SPIN
  byte 0[3077]
*/

char *Memory[MEMORY_MASK + 1];

#ifdef CSPIN_FLAG
void start(int argc, char **argv)
{
    c_start();
    main(argc, argv);
    exit(0);
}

// Generate a random number
int rand(void)
{
    return(?seed) & 0x7fffffff;
}
#endif

// Check if a key has been pressed
int kbhit(void)
{
    return ser_kbhit();
}

// Return the system tick count
int GetTicks(void)
{
#ifdef CSPIN_FLAG
    return cnt;
#else
    return 0;
#endif
}

// Convert ASCII character to upper case
int toupper( int val)
{
    if (val >= 'a' &&val <= 'z')
    {
        val -= 'a' - 'A';
    }
    return val;
}

// Find the next character that does not match val
char *SkipChar(char *str, int val)
{
    while (*str)
    {
        if (*str != val) break;
        str++;
    }
    return str;
}

// Find the next character that matches val
char *FindChar(char *str, int val)
{
    while (*str)
    {
        if (*str == val) break;
        str++;
    }
    return str;
}

// Convert the space-delimited string to a list of tokens
int tokenize(char *str, char **tokens)
{
    char *ptr;
    int num, len, val;
    num = 0;
    len = strlen(str);
    val = str[len - 1];
    if (len == 0)
    {
        return 0;
    }
    if (val < 'A' || (val > 'Z' &&val < 'a') || val > 'z')
    {
        str[len + 1] = 0;
        str[len] = str[len - 1];
        str[len - 1] = ' ';
    }
    while (*str)
    {
        str = SkipChar(str, ' ');
        if (*str == 0)
        {
             break;
        }
        tokens[num ++] = str;
        ptr = FindChar(str, ' ');
        if (*ptr) *ptr++ = 0;
        str = ptr;
    }
    return num;
}

// Find the word "str" in the vocab table
char *FindWord(char *str)
{
    char *ptr;
    ptr = vocab;
    while (ptr < vocab + size)
    {
        if ((*ptr & TYPE_MASK) == TYPE_RELATED)
        {
            ptr += 7;
        }
        else if ((*ptr & TYPE_MASK) == TYPE_RELATE2)
        {
            ptr += 9;
        }
        else if ((*ptr & TYPE_MASK) == TYPE_EQUATE)
        {
            ptr += 5;
        }
        else
        {
            if ((*ptr & DELETE_MASK) == 0)
            {
                if (strcmp(str, ptr + 1) == 0)
                {
                    return ptr;
                }
            }
            ptr += strlen(ptr) + 3;
        }
    }
    return 0;
}

// Reset the vocab table
void ResetVocab(void)
{
    char buf[100];
    printf("< Ary you sure you want to forget everything?\n");
    printf("> ");
    gets(buf);
    if (toupper(*buf) == 'Y')
    {
        size = 0;
        changeflag = 1;
        printf("< Memory erased\n");
    }
    else
    {
        printf("< Reset aborted\n");
    }
}

// Reset to a basic vocab table
void ResetBasicVocab(void)
{
    char buf[100];
    printf("< Ary you sure you want to reset to a basic word list?\n");
    printf("> ");
    gets(buf);
    if (toupper(*buf) == 'Y')
    {
        size = basicsize;
        memcpy(vocab, basicvocab, size);
        changeflag = 1;
        printf("< Memory reset to basic words\n");
    }
    else
    {
        printf("< Reset aborted\n");
    }
}

// Dump out a Spin formatted version of the vocab table
void DumpHexWords(void)
{
    int i = 0;
    char *ptr = vocab;

    printf("  version long %d\n", version);
    printf("  size long %d\n", size);
    printf("  vocab byte\n");
    printf("  byte ");
    while (ptr < vocab + size)
    {
        printf("$%2.2x", *ptr++);
        if (ptr == vocab + size)
        {
            printf("\n");
        }
        else if ((i ++) % 16 == 15)
        {
            printf("\n  byte ");
        }
        else
        {
            printf(",");
        }
    }
    printf("  byte 0[%d]\n", BUF_SIZE - size);
}

// Dump an ASCII formatted version of the vocab table
void DumpAsciiWords(char *name, int num)
{
    char *ptr1;
    int index, index2, index3;
    char *ptr = vocab;

    if (num > 1)
    {
        ptr = FindWord(name);
        if (!ptr)
        {
            printf("$ %s not found\n", name);
            return;
        }
    }
    while (ptr < vocab + size)
    {
        if ((*ptr & TYPE_MASK) == TYPE_RELATED)
        {
            ptr += 7;
        }
        else if ((*ptr & TYPE_MASK) == TYPE_RELATE2)
        {
            ptr += 9;
        }
        else if ((*ptr & TYPE_MASK) == TYPE_EQUATE)
        {
            ptr += 5;
        }
        else
        {
            if ((*ptr & DELETE_MASK) == 0)
            {
                printf("$ %s - %s", ptr + 1, types[*ptr - 1]);
                ptr1 = ptr + strlen(ptr) + 1;
                while (ptr1 < vocab + size && (*ptr1 || ptr1[1]))
                {
                    index = (*ptr1 << 7) | ptr1[1];
                    ptr1 = vocab + index;
                    if ((*ptr1 &TYPE_MASK) == TYPE_RELATED)
                    {
                        if ((*ptr1 & DELETE_MASK) == 0)
                        {
                            index = (ptr1[1] << 7) | ptr1[2];
                            index2 = (ptr1[3] << 7) | ptr1[4];
                            printf(" (%s %s)", vocab + index + 1, vocab + index2 + 1);
                        }
                        ptr1 += 5;
                    }
                    else if ((*ptr1 &TYPE_MASK) == TYPE_RELATE2)
                    {
                        if ((*ptr1 & DELETE_MASK) == 0)
                        {
                            index = (ptr1[1] << 7) | ptr1[2];
                            index2 = (ptr1[3] << 7) | ptr1[4];
                            index3 = (ptr1[5] << 7) | ptr1[6];
                            printf(" (%s %s %s)", vocab + index + 1, vocab + index2 + 1, vocab + index3 + 1);
                        }
                        ptr1 += 7;
                    }
                    else if ((*ptr1 &TYPE_MASK) == TYPE_EQUATE)
                    {
                        if ((*ptr1 & DELETE_MASK) == 0)
                        {
                            index = (ptr1[1] << 7) | ptr1[2];
                            printf(" (=%s)", vocab + index + 1);
                        }
                        ptr1 += 3;
                    }
                    else
                    {
                        printf("$ Invalid type = %2.2x\n", *ptr1);
                    }
                }
                printf("\n");
                if (num > 1)
                {
                    break;
                }
            }
            ptr += strlen(ptr) + 3;
        }
    }
}

// Add a word to the vocab table
void AddWord( int attrib, char *str)
{
    char *ptr = vocab + size;

    if (size + strlen(str) + 4 > BUF_SIZE)
    {
        printf("$ Out of memory\n");
        return;
    }
    if (debugmode)
    {
        printf("AddWord %s - %s\n", str, types[attrib - 1]);
    }
    *ptr++ = attrib;
    strcpy(ptr, str);
    ptr += strlen(ptr) + 1;
    *ptr++ = 0;
    *ptr = 0;
    size += strlen(str) + 4;
    changeflag = 1;
}

// Determine the attribute of an unknown word, and add it to the vocab table
int UnknownWord(char **tokens, int num, int i)
{
    char *str = tokens[i];
    char buf[100];
    char *ptr;
    int caseflag = 0;

    if (i == 0)
    {
        if (*str >= 'A' && *str <= 'Z')
        {
            caseflag = 1;
            *str += 'a' - 'A';
            if (FindWord(str))
            {
                return 0;
            }
            *str -= 'a' - 'A';
            AddWord(TYPE_NAME, str);
            return 0;
        }
        else
        {
            if (num > 2)
            {
                if (strcmp(tokens[1], "equal") == 0 || strcmp(tokens[1], "equals") == 0)
                {
                    ptr = FindWord(tokens[2]);
                    if (ptr)
                    {
                        AddWord(*ptr, tokens[0]);
                        return 0;
                    }
                }
                else
                {
                    AddWord(TYPE_ADJECT, tokens[0]);
                    return 0;
                }
            }
        }
    }
    else
    {
        ptr = FindWord(tokens[i - 1]);
        if (ptr == 0)
        {
            SomethingWrong(tokens[i - 1]);
        }
        else
        {
            if (*ptr == TYPE_ARTICLE && (i == 1 || i == num - 1))
            {
                AddWord(TYPE_NOUN, str);
                return 0;
            }
            else if (i < 3 && (*ptr == TYPE_NOUN || *ptr == TYPE_NAME))
            {
                AddWord(TYPE_VERB, str);
                return 0;
            }
            else if ((i > 1 && *ptr == TYPE_VERB) || i > 2)
            {
                if (*str >= 'A' && *str <= 'Z')
                {
                    AddWord(TYPE_NAME, str);
                    return 0;
                }
                else if (i < num - 1 || *ptr == TYPE_VERB)
                {
                    AddWord(TYPE_ADJECT, str);
                    return 0;
                }
            }
            if (i == num - 1 && num >= 4)
            {
                ptr = FindWord(tokens[i - 2]);
                if (ptr &&*ptr == TYPE_ARTICLE)
                {
                    AddWord(TYPE_NOUN, str);
                    return 0;
                }
                else
                {
                    AddWord(TYPE_ADJECT, str);
                    return 0;
                }
            }
        }
    }
    printf("< I don't know %s\n", str);
    while (1)
    {
        printf("< Is it a noun, verb, pronoun, adjective, Name, Article, query or ignore?\n");
        printf("> ");
        gets(buf);
        if (caseflag && *buf != 'N' && strcmp(str, "I"))
        {
            *str += 'a' - 'A';
        }
        if (*buf == 'n')
        {
            AddWord(TYPE_NOUN, str);
            break;
        }
        else if (*buf == 'v')
        {
            AddWord(TYPE_VERB, str);
            break;
        }
        else if (*buf == 'a')
        {
            AddWord(TYPE_ADJECT, str);
            break;
        }
        else if (*buf == 'p')
        {
            AddWord(TYPE_PRONOUN, str);
            break;
        }
        else if (*buf == 'N')
        {
            AddWord(TYPE_NAME, str);
            break;
        }
        else if (*buf == 'A')
        {
            AddWord(TYPE_ARTICLE, str);
            break;
        }
        else if (*buf == 'q')
        {
            AddWord(TYPE_QUERY, str);
            break;
        }
        else if (*buf == 'i')
        {
            return 1;
        }
        printf("I don't understand\n");
    }
    return 0;
}

// Delete a word and all references to it from the vocab table
void DeleteWord( char **tokens, int num)
{
    char *ptr;
    char buf[100];
    int index, index2;

    if (num < 2)
    {
        printf("< I don't know what to forget\n");
        return;
    }
    ptr = FindWord(tokens[1]);
    if (ptr == 0)
    {
        printf("< I don't know %s\n", tokens[1]);
        return;
    }
    while (1)
    {
        printf("< Do you want me to forget %s?\n", tokens[1]);
        printf("> ");
        gets(buf);
        if (*buf == 'y')
        {
            index = ptr - vocab;
            // Delete the word
            *ptr |= DELETE_MASK;
            // Delete the chain
            while ((ptr = GetNextItem(ptr)))
            {
                *ptr |= DELETE_MASK;
            }
            // Delete any reference to this word
            ptr = vocab;
            while (ptr < vocab + size)
            {
                if ((*ptr &TYPE_MASK) == TYPE_EQUATE)
                {
                    if ((*ptr & DELETE_MASK) == 0)
                    {
                        index2 = (ptr[1] << 7) | ptr[2];
                        if (index == index2)
                        {
                            *ptr |= DELETE_MASK;
                        }
                    }
                    ptr += 5;
                }
                else if ((*ptr &TYPE_MASK) == TYPE_RELATED)
                {
                    if ((*ptr & DELETE_MASK) == 0)
                    {
                        index2 = (ptr[1] << 7) | ptr[2];
                        if (index == index2)
                        {
                            *ptr |= DELETE_MASK;
                        }
                        index2 = (ptr[3] << 7) | ptr[4];
                        if (index == index2)
                        {
                            *ptr |= DELETE_MASK;
                        }
                    }
                    ptr += 7;
                }
                else if ((*ptr &TYPE_MASK) == TYPE_RELATE2)
                {
                    if ((*ptr & DELETE_MASK) == 0)
                    {
                        index2 = (ptr[1] << 7) | ptr[2];
                        if (index == index2)
                        {
                            *ptr |= DELETE_MASK;
                        }
                        index2 = (ptr[3] << 7) | ptr[4];
                        if (index == index2)
                        {
                            *ptr |= DELETE_MASK;
                        }
                        index2 = (ptr[5] << 7) | ptr[6];
                        if (index == index2)
                        {
                            *ptr |= DELETE_MASK;
                        }
                    }
                    ptr += 9;
                }
                else
                {
                    ptr += strlen(ptr) + 3;
                }
            }
            changeflag = 1;
            printf("< %s forgotten\n", tokens[1]);
            return;
        }
        else if (*buf == 'n')
        {
            printf("< %s is not forgotten\n", tokens[1]);
            return;
        }
        else
        {
            printf("< I don't understand\n");
        }
    }
}

// Could find a word that was expected.  Print a message.
void SomethingWrong(char *str)
{
    printf("Something's wrong\n");
    printf("I couldn't find %s\n", str);
}

// Equate the first word to the second word
void AddEquate(char *ptr1, char *ptr2)
{
    int index;
    char *ptr3 = ptr1 + strlen(ptr1) + 1;
    if (size + 5 > BUF_SIZE)
    {
        printf("< Out of memory\n");
        return;
    }
    while (ptr3 < vocab + size && (*ptr3 || ptr3[1]))
    {
        index = (*ptr3 << 7) | ptr3[1];
        ptr3 = vocab + index;
        if ((*ptr3 & TYPE_MASK) == TYPE_EQUATE)
        {
            if ((*ptr3 & DELETE_MASK) == 0)
            {
                index = (ptr3[1] << 7) | ptr3[2];
                if (ptr2 == vocab + index)
                {
                    return;
                }
            }
            ptr3 += 3;
        }
        else if ((*ptr3 & TYPE_MASK) == TYPE_RELATED)
        {
            ptr3 += 5;
        }
        else if ((*ptr3 & TYPE_MASK) == TYPE_RELATE2)
        {
            ptr3 += 7;
        }
        else
        {
            printf("Unknown type = %2.2x\n", *ptr3);
        }
    }
    *ptr3 = size >> 7;
    ptr3[1] = size &127;
    ptr3 = vocab + size;
    *ptr3++ = TYPE_EQUATE;
    index = ptr2 - vocab;
    *ptr3 = index >> 7;
    ptr3[1] = index &127;
    ptr3[2] = 0;
    ptr3[3] = 0;
    size += 5;
    changeflag = 1;
}

// Get the next item in a chain
char *GetNextItem(char *ptr)
{
    int index;
    if (ptr == 0)
    {
        return 0;
    }
    while (ptr < vocab + size)
    {
        if ((*ptr &TYPE_MASK) == TYPE_EQUATE)
        {
            ptr += 3;
        }
        else if ((*ptr &TYPE_MASK) == TYPE_RELATED)
        {
            ptr += 5;
        }
        else if ((*ptr &TYPE_MASK) == TYPE_RELATE2)
        {
            ptr += 7;
        }
        else
        {
            ptr += strlen(ptr) + 1;
        }
        index = (*ptr << 7) | ptr[1];
        if (index == 0)
        {
            break;
        }
        ptr = vocab + index;
        if ((*ptr & DELETE_MASK) == 0)
        {
            return ptr;
        }
    }
    return 0;
}

// Get the next equate item in a chain
char *GetNextEquate(char *ptr)
{
    while ((ptr = GetNextItem(ptr)))
    {
        if (*ptr == TYPE_EQUATE)
        {
            return ptr;
        }
    }
    return 0;
}

// Match two word entries and their equates in the vocab table
int Match(char *ptr1, char *ptr2)
{
    int index;
    char *ptr3;
    char *name1;
    char *name2;
    name1 = ptr1 + 1;
    while (1)
    {
        ptr3 = ptr2;
        name2 = ptr2 + 1;
        while (1)
        {
            if (strcmp(name1, name2) == 0)
            {
                return 1;
            }
            ptr3 = GetNextEquate(ptr3);
            if (!ptr3)
            {
                break;
            }
            index = (ptr3[1] << 7) | ptr3[2];
            name2 = vocab + index + 1;
        }
        ptr1 = GetNextEquate(ptr1);
        if (!ptr1)
        {
            break;
        }
        index = (ptr1[1] << 7) | ptr1[2];
        name1 = vocab + index + 1;
    }
    return 0;
}

// Remove articles from a token list
int RemoveArticles(char **tokens, int num)
{
    int i, j;
    char *ptr;
    i = 0;
    while (i < num)
    {
        ptr = FindWord(tokens[i]);
        if (*ptr == 0)
        {
            printf("RemoveArticles: Can't find %s\n", tokens[i]);
            i++;
            continue;
        }
        if (*ptr == TYPE_ARTICLE)
        {
            num--;
            j = i;
            while (j < num)
            {
                tokens[j] = tokens[j + 1];
                j++;
            }
            continue;
        }
        i++;
    }
    return num;
}

// Locate exact matches of a word and a verb in a related field
char *FindWords(char *word1, char *word2)
{
    char *ptr1;
    char *ptr2;
    int index;
    ptr1 = FindWord(word1);
    while (ptr1)
    {
        if (*ptr1 == TYPE_RELATED)
        {
            index = (ptr1[1] << 7) | ptr1[2];
            ptr2 = vocab + index;
            if (strcmp(ptr2 + 1, word2) == 0)
            {
                return ptr1;
            }
        }
        ptr1 = GetNextItem(ptr1);
    }
    return 0;
}

// Return true if val is a vowel
int isvowel(int val)
{
    val = toupper(val);
    if (val == 'A' || val == 'E' || val == 'I' || val == 'O' || val == 'U')
    {
        return 1;
    }
    return 0;
}

void AddToMemory(char *name, char *ptr1)
{
    char *ptr0 = FindWord(name);
    if (ptr0 == 0 && *name >= 'A' && *name <= 'Z')
    {
        *name += 'a' - 'A';
        ptr0 = FindWord(name);
        *name -= 'a' - 'A';
        if (ptr0 == 0) return;
    }
    Memory[MemoryIndex] = ptr0;
    Memory[MemoryIndex + 1] = ptr1;
    MemoryIndex = (MemoryIndex + 2) & MEMORY_MASK;
}

void PrintMemory(void)
{
    char *ptr0;
    char *ptr1;
    int index = MemoryIndex;

    while (1)
    {
        ptr0 = Memory[index];
        ptr1 = Memory[index + 1];
        if (ptr0 && ptr1)
        {
            if ((*ptr0 & DELETE_MASK) == 0 && (*ptr1 & DELETE_MASK) == 0)
            {
                PrintStatement(ptr0 + 1, ptr1, 0);
            }
        }
        index = (index + 2) & MEMORY_MASK;
        if (index == MemoryIndex) break;
    }
    printf("$\n");
}

int CheckMemory(char *ptr0, char *ptr1)
{
    char *ptr0a;
    char *ptr1a;
    int index = MemoryIndex;

    while (1)
    {
        ptr0a = Memory[index];
        ptr1a = Memory[index + 1];
        if (ptr0 == ptr0a && ptr1 == ptr1a) return 1;
        index = (index + 2) & MEMORY_MASK;
        if (index == MemoryIndex) break;
    }

    return 0;
}

// Determine whether the word is plural, singular or something else
int GetPlural(char *name)
{
    int len = strlen(name);
    char *ptr1 = FindWord(name);
    //printf("GetPlural %s\n", name);
    if (strcmp(name, "I") == 0)
    {
        return -1;
    }
    if (len < 2)
    {
        return 0;
    }
    if (strcmp(name, "you") == 0 || strcmp(name, "You") == 0)
    {
        return 1;
    }
    if (ptr1 && *ptr1 != TYPE_NOUN && *ptr1 != TYPE_ADJECT)
    {
        return 0;
    }
    if (name[len - 1] == 's' &&name[len - 2] != 's')
    {
        return 1;
    }
    return 0;
}

char *GetVerb(int plural, char *ptr1)
{
    char *verbptr;
    char *isptr = FindWord("is");
    if (isptr && Match(isptr, ptr1))
    {
        if (plural == 1)
        {
            verbptr = "are";
        }
        else if (plural == -1)
        {
            verbptr = "am";
        }
        else
        {
            verbptr = "is";
        }
    }
    else
    {
        verbptr = ptr1 + 1;
    }
    //printf("GetVerb %d %s %s\n", plural, ptr1+1, verbptr);
    return verbptr;
}

// Print a statement formed from the name and the field pointed to by ptr1
// Substitue I for you, and You for I in the name
void PrintStatement(char *name, char *ptr1, int remember)
{
    char *ptr;
    int nullstr;
    char *Astr;
    int index, index2, index3;
    char namestr[100];
    char *astr_0;
    int plural;
    char *verbptr;
    nullstr = 0;
    Astr = (char *)&nullstr;
    astr_0 = (char *)&nullstr;
    if (remember)
    {
        if (strcmp(name, "you") == 0)
        {
            name = "I";
        }
        else if (strcmp(name, "I") == 0)
        {
            name = "you";
        }
    }
    strcpy(namestr, name);
    plural = GetPlural(namestr);
    ptr = FindWord(namestr);
    if (ptr)
    {
        if (*ptr == TYPE_NOUN && plural == 0 && (*ptr1 == TYPE_RELATED || *ptr1 == TYPE_RELATE2))
        {
            if (isvowel(*namestr))
            {
                Astr = "An ";
            }
            else
            {
                Astr = "A ";
            }
        }
    }
    if (!Astr[0] && *namestr >= 'a' && *namestr <= 'z')
    {
        *namestr -= 'a' - 'A';
    }
    if (remember)
    {
        printf("< ");
    }
    else
    {
        printf("$ ");
    }
    if (*ptr1 == TYPE_RELATED)
    {
        index = (ptr1[1] << 7) | ptr1[2];
        index2 = (ptr1[3] << 7) | ptr1[4];
        if (vocab[index2] == TYPE_NOUN &&GetPlural(vocab + index2 + 1) == 0)
        {
            if (isvowel(vocab[index2 + 1]))
            {
                astr_0 = "an ";
            }
            else
            {
                astr_0 = "a ";
            }
        }
        verbptr = GetVerb(plural, vocab + index);
        printf("%s%s %s %s%s\n", Astr, namestr, verbptr, astr_0, vocab + index2 + 1);
        if (speak)
        {
            fprintf(speak, "%s%s %s %s%s\n", Astr, namestr, verbptr, astr_0, vocab + index2 + 1);
        }
    }
    else if (*ptr1 == TYPE_RELATE2)
    {
        index = (ptr1[1] << 7) | ptr1[2];
        index2 = (ptr1[3] << 7) | ptr1[4];
        index3 = (ptr1[5] << 7) | ptr1[6];
        if (vocab[index3] == TYPE_NOUN &&GetPlural(vocab + index3 + 1) == 0)
        {
            if (isvowel(vocab[index2 + 1]))
            {
                astr_0 = "an ";
            }
            else
            {
                astr_0 = "a ";
            }
        }
        verbptr = GetVerb(plural, vocab + index);
        printf("%s%s %s %s%s %s\n", Astr, namestr, verbptr, astr_0, vocab + index2 + 1, vocab + index3 + 1);
        if (speak)
        {
            fprintf(speak, "%s%s %s %s%s %s\n", Astr, namestr, verbptr, astr_0, vocab + index2 + 1, vocab + index3 + 1);
        }
    }
    else if (*ptr1 == TYPE_EQUATE)
    {
        index = (ptr1[1] << 7) | ptr1[2];
        printf("%s is %s\n", namestr, vocab + index + 1);
        if (speak)
        {
            fprintf(speak, "%s is %s\n", namestr, vocab + index + 1);
        }
    }
    else
    {
        printf("%s is a %s\n", namestr, types[*ptr1 - 1]);
        if (speak)
        {
            fprintf(speak, "%s is a %s\n", namestr, types[*ptr1 - 1]);
        }
    }
    if (remember)
    {
        AddToMemory(namestr, ptr1);
    }
}

void PrintQuestion(char *ptr1)
{
    int nullstr;
    char *Astr;
    char *isptr;
    int plural;
    char *verbptr;
    nullstr = 0;
    Astr = (char *)&nullstr;
    isptr = FindWord("is");
    plural = GetPlural(ptr1 + 1);
    verbptr = GetVerb(plural, isptr);
    if (*ptr1 == TYPE_NOUN && plural != 1)
    {
        if (isvowel(ptr1[1]))
        {
            Astr = "an ";
        }
        else
        {
            Astr = "a ";
        }
    }
    printf("< What %s %s%s?\n", verbptr, Astr, ptr1 + 1);
    if (speak)
    {
        fprintf(speak, "What %s %s%s\n", verbptr, Astr, ptr1 + 1);
    }
}

#if 1
// Process a list of tokens received from the user
void Process(char **tokens, int num, int question)
{
    int i, index, index2, index3;
    char *ptr[4];
    char *ptr1;
    char *ptr0;
    char *ptrx; 
    int count, counta;
    char *ptry;
    char *ptrya;
    int match_0, equalflag, tempflag;

#if 1
    match_0 = 0;
    equalflag = 0;
    tempflag = 0;
    // Equate the two words if it's an equal statement
    if (num == 3)
    {
        if (!question && (strcmp(tokens[1], "equal") == 0 || strcmp(tokens[1], "equals") == 0))
        {
            if (debugmode)
            {
                printf("equals mode\n");
            }
            equalflag = 1;
            ptr0 = FindWord(tokens[0]);
            ptr1 = FindWord(tokens[2]);
            if (!ptr0)
            {
                SomethingWrong(tokens[0]);
                return;
            }
            if (!ptr1)
            {
                SomethingWrong(tokens[2]);
                return;
            }
            AddEquate(ptr0, ptr1);
            AddEquate(ptr1, ptr0);
            return;
        }
    }
    // Check if the first word is do/does/did.
    // If so, remove the first word and make it a question.
    if ((ptr0 = FindWord("do")))
    {
        if (debugmode)
        {
            printf("does query\n");
        }
        ptr1 = FindWord(tokens[0]);
        if (!ptr1)
        {
            SomethingWrong(tokens[0]);
            return;
        }
        if (Match(ptr0, ptr1))
        {
            question = 1;
            num--;
            i = 0;
            while (i < num)
            {
                tokens[i] = tokens[i + 1];
                i++;
            }
        }
    }
    // Remove the articles
    num = RemoveArticles(tokens, num);
    // There must be three items in the list or it is too short or too long
    if (num < 3)
    {
        printf("I don't understand\n");
        printf("Use more words\n");
        return;
    }
    if (num > 4)
    {
        printf("I don't understand\n");
        printf("Use less words\n");
        return;
    }
    // Locate the tokens in the vocab table
    i = 0;
    while (i < num)
    {
        ptr[i] = FindWord(tokens[i]);
        if (ptr[i] == 0)
        {
            SomethingWrong(tokens[i]);
            return;
        }
        i++;
    }
    // Change the word type if commanded to do so
    if (num == 3 && strcmp(tokens[1], "is") == 0)
    {
        if (strcmp(tokens[2], "noun") == 0)
        {
            *ptr[0] = TYPE_NOUN;
            return;
        }
        else if (strcmp(tokens[2], "name") == 0)
        {
            *ptr[0] = TYPE_NAME;
            return;
        }
        else if (strcmp(tokens[2], "adjective") == 0)
        {
            *ptr[0] = TYPE_ADJECT;
            return;
        }
        else if (strcmp(tokens[2], "verb") == 0)
        {
            *ptr[0] = TYPE_VERB;
            return;
        }
    }
#endif
    // If the first word is a query, the second word must be is/am/are
    if (*ptr[0] == TYPE_QUERY)
    {
#if 1
        if (debugmode)
        {
            printf("query mode\n");
        }
        question = 1;
        ptr1 = FindWord("is");
        tempflag = 0;
        //if (num != 3 || !ptr1 || !Match(ptr1, ptr[1]))
        if (num == 3)
        {
            if (ptr1)
            {
                if (Match(ptr1, ptr[1]))
                {
                    tempflag = 1;
                }
            }
        }
        if (!tempflag)
        {
            printf("I don't understand\n");
            return;
        }
        // Scan related entries and equate's related entries
        ptr1 = ptr[2];
        ptr0 = ptr1;
        ptrx = ptr1;
        ptry = ptr1;
        count = 0;
        // variables for items not in short term memory
        ptrya = 0;
        counta = 0;
        if (debugmode)
        {
            PrintStatement(tokens[2], ptr1, 0);
        }
        while (1)
        {
            while (1)
            {
                ptr1 = GetNextItem(ptr1);
                if (!ptr1)
                {
                    break;
                }
                if (*ptr1 != TYPE_RELATED &&*ptr1 != TYPE_RELATE2)
                {
                    continue;
                }
                //index = (ptr1[1] << 7) | ptr1[2];
                //index2 = (ptr1[3] << 7) | ptr1[4];
                if (debugmode)
                {
                    PrintStatement(tokens[2], ptr1, 0);
                }
                if (CheckMemory(ptr[2], ptr1) == 0)
                {
                    counta++;
                    if ((rand()% counta) == 0)
                    {
                        ptrya = ptr1;
                    }
                }
                if (ptrya == 0)
                {
                    count++;
                    if ((rand()% count) == 0)
                    {
                        ptry = ptr1;
                    }
                }
            }
            if (ptrya == 0)
            {
                ptrya = ptry;
            }
            ptr0 = GetNextEquate(ptr0);
            if (!ptr0)
            {
                break;
            }
            index = (ptr0[1] << 7) | ptr0[2];
            ptrx = vocab + index;
            ptr1 = ptrx;
        }
        if (ptrya)
        {
            if (*ptrya == TYPE_RELATED || *ptrya == TYPE_RELATE2 || ptrya[1] == TYPE_EQUATE)
            {
                PrintStatement(tokens[2], ptrya, 1);
            }
            else
            {
                printf("< I don't know\n");
            }
        }
        return;
#endif
    }
#if 1
    // If not an equal statement or a question check if first word is a verb
    // Is so, swap the first and second tokens
    if (!equalflag && !question && *ptr[0] == TYPE_VERB)
    {
        if (debugmode)
        {
            printf("Lead verb query\n");
        }
        question = 1;
        ptr1 = ptr[0];
        ptr[0] = ptr[1];
        ptr[1] = ptr1;
        ptr1 = tokens[0];
        tokens[0] = tokens[1];
        tokens[1] = ptr1;
    }
    // Look for a match to the three words.  If it is a question and
    // a match is found print "Yes"
    ptr1 = ptr[0];
    ptr0 = ptr1;
    ptrx = ptr1;
    while (1)
    {
        while (1)
        {
            ptr1 = GetNextItem(ptr1);
            if (!ptr1)
            {
                break;
            }
            if (*ptr1 != TYPE_RELATED &&*ptr1 != TYPE_RELATE2)
            {
                continue;
            }
            index = (ptr1[1] << 7) | ptr1[2];
            index2 = (ptr1[3] << 7) | ptr1[4];
            tempflag = Match(ptr[1], vocab + index) &&Match(ptr[2], vocab + index2);
            if (num == 4)
            {
                index3 = (ptr1[5] << 7) | ptr1[6];
                tempflag = tempflag &&Match(ptr[3], vocab + index3);
                if (debugmode)
                {
                    printf("%s (%s,%s) (%s,%s) (%s,%s)\n", ptrx + 1, ptr[1] + 1, vocab + index + 1, ptr[2] + 1, vocab + index2 + 1, ptr[3], vocab + index3 + 1);
                }
            }
            else
            {
                if (debugmode)
                {
                    printf("%s (%s,%s) (%s,%s)\n", ptrx + 1, ptr[1] + 1, vocab + index + 1, ptr[2] + 1, vocab + index2 + 1);
                }
            }
            if (tempflag)
            {
                if (!question)
                {
                    printf("< I know that\n");
                }
                else
                {
                    printf("< Yes\n");
                }
                return;
            }
        }
        ptr0 = GetNextEquate(ptr0);
        if (!ptr0)
        {
            break;
        }
        index = (ptr0[1] << 7) | ptr0[2];
        ptrx = vocab + index;
        ptr1 = ptrx;
    }
#endif
    // Print "No" if it was a question and no match found
    if (question)
    {
        printf("< No\n");
        return;
    }
#if 1
    // Add relationship if there was no match and it's not a question
    if (!match_0)
    {
        if (size + 1 + (num << 1) > BUF_SIZE)
        {
            printf("< Out of memory\n");
            return;
        }
        ptr1 = ptr[0];
        while ((ptr0 = GetNextItem(ptr1)))
        {
            ptr1 = ptr0;
        }
        if (*ptr1 == TYPE_RELATED)
        {
            ptr1 += 5;
        }
        else if (*ptr1 == TYPE_RELATE2)
        {
            ptr1 += 7;
        }
        else if (*ptr1 == TYPE_EQUATE)
        {
            ptr1 += 3;
        }
        else
        {
            ptr1 += strlen(ptr1) + 1;
        }
        *ptr1 = size >> 7;
        ptr1[1] = size & 127;
        ptr1 = vocab + size;
        if (num == 3)
        {
            *ptr1 = TYPE_RELATED;
        }
        else
        {
            *ptr1 = TYPE_RELATE2;
        }
        index = ptr[1] - vocab;
        ptr1[1] = index >> 7;
        ptr1[2] = index & 127;
        index = ptr[2] - vocab;
        ptr1[3] = index >> 7;
        ptr1[4] = index & 127;
        if (num == 3)
        {
            ptr1[5] = 0;
            ptr1[6] = 0;
            size += 7;
        }
        else
        {
            index = ptr[3] - vocab;
            ptr1[5] = index >> 7;
            ptr1[6] = index & 127;
            ptr1[7] = 0;
            ptr1[8] = 0;
            size += 9;
        }
        changeflag = 1;
        // Check if predicate is undefined
        index = 0;
        ptr1 = ptr[2];
        if (*ptr1 < 32)
        {
            ptr0 = ptr1 + strlen(ptr1) + 1;
            index = (*ptr0 << 7) | ptr0[1];
        }
        if (index == 0)
        {
            PrintQuestion(ptr1);
        }
        else
        {
            printf("< I will remember that\n");
        }
    }
#endif
}
#endif

// Wait from 3 to 10 seconds for input.  Print a random fact from
// the vocab table if no input received.
void WaitForInput(void)
{
    int waittime, starttime;
    char *ptr1;
    char *ptr2;
    char *ptr1a;
    char *ptr2a;
    int count, FirstTime;
    waittime = ((rand() % 8) + 3) * clkfreq;
    starttime = GetTicks();
    printf("> ");
    while (!kbhit())
    {
        // Scan all the entries in the vocab table
        if (GetTicks()- starttime > waittime)
        {
            count = 0;
            ptr1 = vocab;
            ptr1a = ptr2a = 0;
            while (ptr1 < vocab + size)
            {
                if ((*ptr1 & TYPE_MASK) == TYPE_RELATED)
                {
                    ptr1 += 7;
                }
                else if ((*ptr1 & TYPE_MASK) == TYPE_RELATE2)
                {
                    ptr1 += 9;
                }
                else if ((*ptr1 & TYPE_MASK) == TYPE_EQUATE)
                {
                    ptr1 += 5;
                }
                else if (*ptr1 & DELETE_MASK)
                {
                    ptr1 += strlen(ptr1) + 3;
                }
                else
                {
                    ptr2 = ptr1;
                    FirstTime = 1;
                    while ((ptr2 = GetNextItem(ptr2)) || FirstTime)
                    {
                        FirstTime = 0;
                        if (ptr2 == 0 && *ptr1 != TYPE_NAME && *ptr1 != TYPE_NOUN && *ptr1 != TYPE_ADJECT)
                        {
                            break;
                        }
                        if (ptr2 == 0 || *ptr2 == TYPE_RELATED)
                        {
                            count++;
                            if ((rand()% count) == 0)
                            {
                                if (count == 1 || CheckMemory(ptr1, ptr2) == 0)
                                {
                                    ptr1a = ptr1;
                                    ptr2a = ptr2;
                                }
                            }
                        }
                    }
                    ptr1 += strlen(ptr1) + 3;
                }
            }
            // Print the entry that was randomly selected
            if (ptr1a && ptr2a)
            {
                if (*ptr2a == TYPE_RELATED)
                {
                    putchar(8);
                    putchar(8);
                    PrintStatement(ptr1a + 1, ptr2a, 1);
                    printf("> ");
                }
            }
            else if (ptr1a)
            {
                putchar(8);
                putchar(8);
                PrintQuestion(ptr1a);
                printf("> ");
            }
            waittime = ((rand()% 10) + 5) * clkfreq;
            starttime = GetTicks();
        }
    }
}

// Save the vocab table
void SaveVocab(void)
{
    //int addr, i, val, val1;
    printf("$ Save disabled\n");
}

/*
  addr := @size
  i := (size + 7) >> 2
  repeat while (i > 0)
    val := long[addr]
    if (val <> i2c.ReadLong(28, $a0, addr))
      i2c.WriteLong(28, $a0, addr, val)
      repeat while (i2c.WriteWait(28, $a0, addr))
    addr += 4
    i--
*/
int CheckCommand(char **tokens, int num, char *buffer)
{
    char *ptr1;
    int retval, caseshift;
    retval = 1;
    caseshift = (tokens[0][0] >= 'A' && tokens[0][0] <= 'Z');
    if (num > 2)
    {
        return 0;
    }
    if (caseshift)
    {
        tokens[0][0] += 'a' - 'A';
    }
    if (strcmp(tokens[0], "help") == 0)
    {
        printf("$ Type a sentence, or enter one of these commands\n");
        printf("$ vocab [word] - Print the words I know\n");
        printf("$ dump         - Print the words as hex codes\n");
        printf("$ forget       - Remove a word from memroy\n");
        printf("$ save         - Save the words to EEPROM\n");
        printf("$ reset        - Reset the word table\n");
        printf("$ basic        - Reset the word table to a basic list\n");
        printf("$ history      - Print the short term memory\n");
        printf("$ goodbye      - Exit the program\n");
    }
    else if (strcmp(tokens[0], "goodbye") == 0)
    {
        if ((changeflag = 0))
        {
            while (1)
            {
                printf("< Should I save?\n");
                printf("> ");
                gets(buffer);
                ptr1 = SkipChar(buffer, ' ');
                if (toupper(*ptr1) == 'Y')
                {
                    SaveVocab();
                    break;
                }
                else if (toupper(*ptr1) == 'N')
                {
                    break;
                }
                printf("< I don't undertand\n");
            }
        }
        printf("< Goodbye\n");
/* INLINE SPIN
waitcnt(cnt+clkfreq/10)
*/
        //abort(0)
        exit(0);
    }
    else if (strcmp(tokens[0], "vocab") == 0)
    {
        DumpAsciiWords(tokens[1], num);
    }
    else if (strcmp(tokens[0], "dump") == 0)
    {
        DumpHexWords();
    }
    else if (strcmp(tokens[0], "forget") == 0)
    {
        DeleteWord(tokens, num);
    }
    else if (strcmp(tokens[0], "save") == 0)
    {
        SaveVocab();
    }
    else if (strcmp(tokens[0], "reset") == 0)
    {
        ResetVocab();
    }
    else if (strcmp(tokens[0], "basic") == 0)
    {
        ResetBasicVocab();
    }
    else if (strcmp(tokens[0], "history") == 0)
    {
        PrintMemory();
    }
    else if (strcmp(tokens[0], "debug") == 0)
    {
        debugmode = !debugmode;
    }
    else
    {
        retval = 0;
    }
    // Restore the case if needed
    if (caseshift)
    {
        tokens[0][0] -= 'a' - 'A';
    }
    return retval;
}

// This routine set the linefeed enable flag based on the config word
// The config word is the first long after the last argument
void config_linefeed(int argc, char **argv)
{
    char *ptr = argv[argc - 1];
    ptr += strlen(ptr) + 1;
    ptr = (char *)((((int)ptr + 3) >> 2) << 2);
#ifdef CSPIN_FLAG
    set_linefeed(*ptr & 1);
#endif
}

int main( int argc, char **argv)
{
    int i, num;
    char *ptr1;
    int ignore, question;
    char buffer[100];
    char *tokens[50];
    //speak := c.getstdout
    //speak := c.openserial(16, 16, 0, 9600, 64, 64)
    config_linefeed(argc, argv);
    printf("Lerner 0.08 -- Spinix version\n");
    printf("Save command disabled in this version\n\n");
    //i2c.Initialize(28)
    i = 0;
    while (i <= MEMORY_MASK)
    {
        Memory[i] = 0;
        i++;
    }
    //printf("< Hello\n");
    ptr1 = FindWords("you", "are");
    if (ptr1)
    {
        PrintStatement("you", ptr1, 1);
    }
    // Loop on each user input
    while (1)
    {
        WaitForInput();
        gets(buffer);
        num = tokenize(buffer, tokens);
        if (num == 0)
        {
            continue;
        }
        if (CheckCommand(tokens, num, buffer))
        {
            continue;
        }
        // Check for a question mark at the end
        if (strcmp(tokens[num - 1], "?") == 0)
        {
            question = 1;
            num--;
        }
        else
        {
            question = 0;
        }
        // Check if all the words are in the vocab table
        ignore = 0;
        i = 0;
        while (i < num)
        {
            if (FindWord(tokens[i]) == 0)
            {
                if ((ignore = UnknownWord(tokens, num, i)))
                {
                    break;
                }
            }
            i++;
        }
        if (ignore)
        {
            continue;
        }
        // Process the user input
        Process(tokens, num, question);
    }
    return 0;
}
/*
+-----------------------------------------------------------------------------+
|                          TERMS OF USE: MIT License                          |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
*/
