//******************************************************************************
// Copyright (c) 2014 Dave Hein
// See end of file for terms of use.
//******************************************************************************
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* INLINE SPIN
CON
  _clkmode = xinput
  _xinfreq = $4e495053 ' SPIN
*/
// SPIN OBJECT c : "clibos"
void enter(void);
// SPIN OBJECT

#ifdef CSPIN_FLAG
void start(int argc, char **argv)
{
    enter();
    main(argc, argv);
    exit(0);
}
#endif

typedef struct sobheaderS {
    int id;
    int ts;
    int hash;
    unsigned short numexports;
    unsigned short exportsize;
    unsigned short numimports;
    unsigned short importsize;
    unsigned short objectsize;
    unsigned char checksum;
    unsigned char reserved;
    unsigned short varsize;
} sobheaderT;

void DumpExport(FILE *infile);
void DumpImport(FILE *infile);
void GetString(char *ptr, int size, FILE *infile);

unsigned char buffer[1000];

int main(int argc, char **argv)
{
    int i, num;
    FILE *infile;
    int checksum = 0;
    int hash = 0;
    sobheaderT header;

    if (argc != 2)
    {
        printf("usage: sobdump file\n");
        exit(1);
    }

    if (!(infile = fopen(argv[1], "r")))
    {
        printf("Could not open %s\n", argv[1]);
        exit(1);
    }

    fread(&header, 1, 26, infile);

    printf("id         = %8.8x\n", header.id);
    printf("ts         = %8.8x\n", header.ts);
    printf("hash       = %8.8x\n", header.hash);
    printf("numexports = %8.8x\n", header.numexports);
    printf("exportsize = %8.8x\n", header.exportsize);
    printf("numimports = %8.8x\n", header.numimports);
    printf("importsize = %8.8x\n", header.importsize);
    printf("objectsize = %8.8x\n", header.objectsize);
    printf("checksum   = %8.8x\n", header.checksum);
    printf("reserved   = %8.8x\n", header.reserved);
    printf("varsize    = %8.8x\n", header.varsize);

    for (i = 0; i < header.numexports; i++) DumpExport(infile);
    for (i = 0; i < header.numimports; i++) DumpImport(infile);

    while ((num = fread(buffer, 1, 1000, infile)) > 0)
    {
        for (i = 0; i < num; i++)
        {
            checksum += buffer[i];
            hash = ((hash << 1) | ((hash >> 31) & 1)) ^ buffer[i];
        }
    }

    printf("Checksum = %2.2x, hash = %8.8x\n", checksum & 255, hash);

    return 0;
}

void GetString(char *ptr, int size, FILE *infile)
{
    int temp = 0;
    while (size > 1)
    {
        fread(&temp, 1, 1, infile);
        if (!temp) break;
        *ptr++ = temp;
        size--;
    }
    *ptr = 0;
}
    
void DumpExport(FILE *infile)
{
    char name[100];
    int type = 0;
    int parm = 0;
    GetString(name, 100, infile);
    fread(&type, 1, 1, infile);
    if (type <= 1)
        fread(&parm, 1, 4, infile);
    else
        fread(&parm, 1, 2, infile);
    printf("Export %s %x %x\n", name, type, parm);
}

void DumpImport(FILE *infile)
{
    char name[100];
    int count = 0;
    int parm = 0;
    GetString(name, 100, infile);
    fread(&count, 1, 2, infile);
    fread(&parm, 1, 1, infile);
    printf("Import %s %x %x\n", name, count, parm);
}
/*
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
*/
