//******************************************************************************
// Copyright (c) 2014 Dave Hein
// See end of file for terms of use.
//******************************************************************************
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

/* INLINE SPIN
CON
  _clkmode = xinput
  _xinfreq = $4e495053 ' SPIN
*/

// SPIN OBJECT c : "clibos"
void CheckStackSpace(void);

// SPIN OBJECT strsub : "strsubs"
#include "strsubs.h"

// SPIN OBJECT sys : "sysdefs"

// SPIN OBJECT
#include "spinit.h"
void AddObject(int num, char **tokens);

// *********************************************************************
// Defines
// *********************************************************************
// Symbol Types
#define TYPE_PUB    1
#define TYPE_DAT    2
#define TYPE_LOCAL  3
#define TYPE_CON    4
#define TYPE_KEY    5
#define TYPE_OP     0x100
#define TYPE_NUM    7
#define TYPE_ACCESS 8
#define TYPE_VAR    9
#define TYPE_REG    10
#define TYPE_POSTOP 0x200
#define TYPE_OBJECT 11

// Expected type for next symbol - i.e., value or operator
#define EXPECT_VAL  0
#define EXPECT_OP   1

// Actions that are performed on a value
#define ACTION_NONE  0
#define ACTION_LOAD  1
#define ACTION_STORE 2
#define ACTION_ADDR  3
#define ACTION_EXEC  4

// Types of Spin blocks
#define LOOP_PUB    1
#define LOOP_REPEAT 2
#define LOOP_IF     3
#define LOOP_ELSE   4
#define LOOP_ELSEIF 5
#define LOOP_CASE   6
#define LOOP_CASE1  7

// Size of the symbol table and string pool
#define MAX_LEVELS  5
#define MAX_SYMBOLS 250
#define STRING_POOL_SIZE MAX_SYMBOLS * 8

#ifdef CSPIN_FLAG
#define SEEK_SET 0
#endif

// Debug print flag
//#define PRINT_PROCESS

// *********************************************************************
// Typedefs
// *********************************************************************
// Function pointer used by a symbol
#ifdef NO_FUNCTION_PTR
typedef int FuncPtrT;
#else
typedef int (* FuncPtrT)(int, int, char **, int);
#endif

// Symbol struct
//NOSPIN_START
typedef struct SymbolS {
    char *name;
    int type;
    FuncPtrT func;
    int parm;
    int size;
} SymbolT;
//NOSPIN_END
/* INLINE SPIN
name = 0
type = 2
func = 4
parm = 6
size = 8
*/

// *********************************************************************
// Global variables
// *********************************************************************
// Input line buffer and token buffers
char buffer[100];
char buf2[200];
char *tokens[50];
int commentflag;
int linenum;

// Variables used for each block level
// PUBs and PRIs start at level zero
int maxstackaddr = 0;
int stacksize = 0;
int indent2 = 0;
int level = 0;
int indent[MAX_LEVELS];
int labelstart[MAX_LEVELS];
int labelend[MAX_LEVELS];
int looplevel[MAX_LEVELS];
int looptype[MAX_LEVELS];

// This variable is used to generate unique labels
int labelnum = 1;

// Symbol table and string storage buffer
int symnum = 0;
int pubnum = 0;
int symnum_save;
int strindex = 0;
SymbolT symtable[MAX_SYMBOLS];
char strpool[STRING_POOL_SIZE];

// Keyword lists
char AccessChar[] = {'b', 'w', 'l'};
char *ret_list[] = {"coginit", "cognew", "lockset", "lockclr", 0};
char *section_keyword[] = {"pub", "pri", "dat", "con", "var", "obj", 0};
char *elselist[] = {"else", "elseif", "elseifnot", 0};
char *type_keyword[] = {"byte", "word", "long", 0};
char nullstr[1] = {0};

// Input and output files
#ifdef CSPIN_FLAG
int infile = 0;
int outfile = 0;
#else
FILE *infile = 0;
FILE *outfile = 0;
#endif

// Debug flag
int debugflag = 0;

// Spin startup code
/* INLINE SPIN
PUB start(argc, argv)
  maxstackaddr := @result + 48
  stacksize := (sys#rendezvous - maxstackaddr - 2000) / 4
  c.enter1(stacksize)
  'c.start1(31, 30, %10000, 115_200, result)
  'c.printf1(string("Starting stack address = %4.4x\n"), maxstackaddr)
  'c.printf1(string("Selecting stack size of %d longs\n"), result)
  strsub.start
  main(argc, argv)
  c.exit(0)
*/

// *********************************************************************
// This program compiles a Spin file to a Spasm file.  It appends
// a ".spin" extension to the specified file name and opens it for
// input.  It appends a ".spasm" extension for the output file name.
// Note: Currently writing the Spasm code to the standard output
// The compiler uses two passes though the source code.  The first
// pass is used to gather all the global symbols, which include the
// method names, constants, DAT variables and VAR variables.
// During the second pass, the local variables are extracted for each
// method, and the Spasm code is generated.
// *********************************************************************
int main(int argc, char **argv)
{
    char *ptr;
    char *fname;

    if (argc == 2)
    {
        fname = argv[1];
    }
    else if (argc == 3 && !strcmp(argv[1], "-d"))
    {
        debugflag = 1;
        fname = argv[2];
    }
    else
    {
        usage();
    }

    if (debugflag)
        printf("Selecting stack size of %d longs\n", stacksize);

    // Initialize the symbol table with the Spin keywords
    AddKeywords();

    // Open the input file
    strcpy(buffer, fname);
    ptr = FindChar(buffer, '.');
    if (!(*ptr))
        strcpy(ptr, ".spn");
    infile = fopen(buffer, "r");
    if (!infile) usage();

    //c.printf2(string("infile = %4.4x, outfile = %4.4x\n"), infile, outfile)
    //c.printf1(string("symnum = %d after AddKeywords\n"), symnum)

    // Perform the first pass and get all the global symbols
    GetGlobalSymbols();
    //c.printf1(string("symnum = %d after GetGlobalSymbols\n"), symnum)

    // Rewind the file and perform the second pass
    fseek(infile, 0, SEEK_SET);

    // Open the output file
    strcpy(buffer, fname);
    ptr = FindChar(buffer, '.');
    strcpy(ptr, ".spa");
    outfile = fopen(buffer, "w");
    if (!outfile)
        usage();

    GenerateSpasm();

    //c.printf1(string("symnum = %d after GenerateSpasm\n"), symnum)
    if (debugflag)
    {
        printf("maxstackaddr = %4.4x\n", maxstackaddr);
        CheckStackSpace();
    }

    fclose(infile);
    fclose(outfile);

    return 0;
}

// *********************************************************************
// Print the usage information and exit
// *********************************************************************
void usage(void)
{
    printf("usage: spinit [-d] file\n");
    printf("  -d - Enable debug prints\n");
    exit(1);
}

// *********************************************************************
// This routine allocates a string buffer from the buffer pool
// *********************************************************************
char *StrAlloc(int size)
{
    char *ptr = &strpool[strindex];
    if (strindex + size > STRING_POOL_SIZE)
    {
	printf("String pool is out of space\n");
        ErrorExit();
    }
    strindex += size;
    return ptr;
}

// *********************************************************************
// This routine adds a symbol to the symbol table
// *********************************************************************
int AddSymbol(char *str, int type, FuncPtrT func, int parm)
{
    char *symptr = StrAlloc(strlen(str) + 1);
    if (symnum >= MAX_SYMBOLS)
    {
	printf("Symbol table is full\n");
        ErrorExit();
    }
    //c.printf2(string("AddSymbol: %s, %d\n"), str, type)
    symtable[symnum].name = symptr;
    symtable[symnum].type = type;
    symtable[symnum].func = func;
    symtable[symnum].parm = parm;
    symtable[symnum].size = 4;
    strcpy(symptr, str);
    symnum++;
    return symnum - 1;
}

// *********************************************************************
// This routine finds a symbol in the table and returns
// it's index, or -1 if not found.
// *********************************************************************
int FindSymbol(char *str)
{
    int i;

    for (i = 0; i < symnum; i++)
    {
	//if (strcmp(str, symtable[i].name) == 0)
	if (StrCompNoCase(str, symtable[i].name))
	{
	    //printf("FindSymbol: found %s\n", str);
	    break;
	}
    }
    if (i == symnum) i = -1;

    return i;
}

// *********************************************************************
// This routine adds a symbol to the symbol table if it is not
// already in the table.  If a duplicate exists, and error message
// is printed and the program exits.
// *********************************************************************
int AddSymbolCheck(char *str, int type, FuncPtrT func, int parm)
{
    if (FindSymbol(str) >= 0)
    {
	printf("Duplicate symbol - %s\n", str);
        ErrorExit();
    }
    return AddSymbol(str, type, func, parm);
}

// *********************************************************************
// This routine determines if a token is a number or
// a symbol in the symbol table.  Numbers use symbol
// table entry zero.  If the token is undefined an error
// message is printed, and the program exits.
// *********************************************************************
int FindSymbolType(char *str)
{
    int j = FindSymbol(str);
    int val;

    if (j < 0)
    {
        val = str[0];
	if (val == '$' || val == DOUBLE_QUOTE || (val >= '0' && val <= '9')) j = 0;
	else
	{
	    printf("Undefined symbol %s\n", str);
            ErrorExit();
	}
    }
    return j;
}

// *********************************************************************
// This routine returns the precedence of a symbol if its
// an operator, or zero if not.
// *********************************************************************
int GetSymbolPrecedence(char *str)
{
    int precedence = 0;
    int j = FindSymbolType(str);
    if (symtable[j].type & TYPE_OP)
	precedence = symtable[j].type & 15;

    return precedence;
}

// *********************************************************************
// This routine locates a token in the symbol table and executes
// it's function.  It combines the expected type, precedence and
// action value into a single state variable.
// *********************************************************************
int ExecuteSymbol(int i, int num, char **tokens, int expect, int prec, int action)
{
    int state = (expect << 8) | (prec << 4) | action;
    int j = FindSymbolType(tokens[i]);
    //printf("ExecuteSymbol %s %d %d %d %8.8x\n", tokens[i], expect, prec, action, state);
#ifdef NO_FUNCTION_PTR
    return ExecuteFunction(symtable[j].func, i, num, tokens, state);
#else
    return (symtable[j].func)(i, num, tokens, state);
#endif
}

// *********************************************************************
// This routine checks if the expected type passed in the state
// variable matches the current token type.
// *********************************************************************
void CheckExpect(int state, int current, char *str)
{
    int expect = state >> 8;

    if (expect != current)
    {
	printf("expect = %d, current = %d, %s\n", expect, current, str);
        ErrorExit();
    }
}

// *********************************************************************
// This is a default keyword function that should never be called.
// *********************************************************************
int ProcessKey(int i, int num, char **tokens, int state)
{
    printf("KEY\n");
    return i + 1;
}

// *********************************************************************
// This routine handles the repeat statement
// *********************************************************************
int ProcessRepeat(int i, int num, char **tokens, int state)
{
    labelstart[++level] = labelnum;
    fprintf(outfile, "label%4.4d\n", labelnum++);
    labelend[level] = labelnum;
    //printf("ProcessRepeat: level = %d, labelend = %d\n", level, labelnum);
    indent[level] = indent2;
    looplevel[level] = level;
    looptype[level] = LOOP_REPEAT;
    i++;
    if (i >= num || tokens[i][0] == SINGLE_QUOTE)
    {
        labelnum++;
        return i;
    }

    if (!StrCompNoCase(tokens[i], "while"))
    {
        printf("Expected while or EOL - %s\n", tokens[i]);
        ErrorExit();
    }
    if (++i >= num)
    {
	printf("Encountered EOL\n");
	ErrorExit();
    }
    //i = ExecuteSymbol(i, num, tokens, EXPECT_VAL, 12, ACTION_LOAD);
    i = ProcessExpression(i, num, tokens, 12, ACTION_LOAD);
    fprintf(outfile, "        jz label%4.4d\n", labelnum++);
    return i;
}

// *********************************************************************
// This routine handles the string function
// *********************************************************************
int ProcessString(int i, int num, char **tokens, int state)
{
    if (++i >= num)
    {
	printf("Encountered EOL\n");
	ErrorExit();
    }
    if (strcmp(tokens[i], "("))
    {
	printf("Expected '('\n");
	ErrorExit();
    }
    if (++i >= num)
    {
	printf("Encountered EOL\n");
	ErrorExit();
    }
    fprintf(outfile, "        jmp label%4.4d\n", labelnum+1);
    fprintf(outfile, "label%4.4d\n", labelnum++);
    fprintf(outfile, "        byte ");
    while (i < num && strcmp(tokens[i], ")"))
    {
        fprintf(outfile, "%s", tokens[i++]);
    }
    fprintf(outfile, ",0\n");
    if (i++ >= num)
    {
	printf("Encountered EOL\n");
	ErrorExit();
    }
    fprintf(outfile, "label%4.4d\n", labelnum++);
    fprintf(outfile, "        labo label%4.4d\n", labelnum-2);
    return i;
}

// *********************************************************************
// This routine handles case statements
// *********************************************************************
int ProcessCase(int i, int num, char **tokens, int state)
{
    fprintf(outfile, "        ldl label%4.4d\n", labelnum);
    labelstart[++level] = 0;
    labelend[level] = labelnum++;
    labelstart[level] = 0;
    indent[level] = indent2;
    looplevel[level] = looplevel[level-1];
    looptype[level] = LOOP_CASE;
    i++;
    if (i >= num)
    {
	printf("Expected expression - %s\n", tokens[i]);
	ErrorExit();
    }
    i = ProcessExpression(i, num, tokens, 12, ACTION_LOAD);
    return i;
}

// *********************************************************************
// This routine handles if and ifnot statements
// *********************************************************************
int ProcessIf(int i, int num, char **tokens, int state)
{
    int ifnotflag = StrCompNoCase(tokens[i], "ifnot");
    labelstart[++level] = 0;
    labelend[level] = labelnum;
    labelstart[level] = 0;
    indent[level] = indent2;
    looplevel[level] = looplevel[level-1];
    looptype[level] = LOOP_IF;
    i++;
    if (i >= num || tokens[i][0] == SINGLE_QUOTE)
    {
	printf("Expected expression - %s\n", tokens[i]);
	ErrorExit();
    }
    i = ProcessExpression(i, num, tokens, 12, ACTION_LOAD);
    if (ifnotflag)
        fprintf(outfile, "        jnz label%4.4d\n", labelnum++);
    else
        fprintf(outfile, "        jz label%4.4d\n", labelnum++);
    return i;
}

// *********************************************************************
// This routine handles elseif and elseifnot statements
// *********************************************************************
int ProcessElseIf(int i, int num, char **tokens, int state)
{
    int ifnotflag = StrCompNoCase(tokens[i], "elseifnot");

    if ((looptype[level] != LOOP_IF && looptype[level] != LOOP_ELSEIF)
        || indent[level] != indent2)
    {
	printf("%s statement without an if\n", tokens[0]);
	ErrorExit();
    }

    if (looptype[level] == LOOP_IF)
    {
	labelstart[level] = labelnum++;
    }
    fprintf(outfile, "        jmp label%4.4d\n", labelstart[level]);
    fprintf(outfile, "label%4.4d\n", labelend[level]);

    looptype[level] = LOOP_ELSEIF;
    labelend[level] = labelnum;

    if (++i >= num || tokens[i][0] == SINGLE_QUOTE)
    {
	printf("Expected expression - %s\n", tokens[i]);
	ErrorExit();
    }
    i = ProcessExpression(i, num, tokens, 12, ACTION_LOAD);
    if (ifnotflag)
        fprintf(outfile, "        jnz label%4.4d\n", labelnum++);
    else
        fprintf(outfile, "        jz label%4.4d\n", labelnum++);

    return i;
}

// *********************************************************************
// This routine handles else statements
// *********************************************************************
int ProcessElse(int i, int num, char **tokens, int state)
{
    if ((looptype[level] != LOOP_IF && looptype[level] != LOOP_ELSEIF)
        || indent[level] != indent2)
    {
	printf("%s statement without an if\n", tokens[0]);
	ErrorExit();
    }
    if (looptype[level] == LOOP_IF)
    {
        fprintf(outfile, "        jmp label%4.4d\n", labelnum);
        fprintf(outfile, "label%4.4d\n", labelend[level]);
        labelend[level] = labelnum++;
    }
    else
    {
        fprintf(outfile, "        jmp label%4.4d\n", labelstart[level]);
        fprintf(outfile, "label%4.4d\n", labelend[level]);
	labelend[level] = labelstart[level];
        labelstart[level] = 0;
    }
    looptype[level] = LOOP_ELSE;
    return i + 1;
}

// *********************************************************************
// This routine handles the quit statement
// *********************************************************************
int ProcessQuit(int i, int num, char **tokens, int state)
{
    int j = looplevel[level];
    //printf("ProcessQuit: level = %d, labelend = %d\n", j, labelend[j]);
    if (labelend[j])
        fprintf(outfile, "        jmp label%4.4d\n", labelend[j]);
    return i + 1;
}

// *********************************************************************
// This routine handles the next statement
// *********************************************************************
int ProcessNext(int i, int num, char **tokens, int state)
{
    int j = looplevel[level];
    if (labelstart[j])
        fprintf(outfile, "        jmp label%4.4d\n", labelstart[j]);
    return i + 1;
}

// *********************************************************************
// This routine handles return statements
// *********************************************************************
int ProcessReturn(int i, int num, char **tokens, int state)
{
    i++;
    if (i >= num || tokens[i][0] == SINGLE_QUOTE)
    {
	fprintf(outfile, "        ret\n");
	return i;
    }
    i = ProcessExpression(i, num, tokens, 12, ACTION_LOAD);
    fprintf(outfile, "        retval\n");
    return i;
}

// *********************************************************************
// This routine handles the byte[], word[] and long[] operator
// *********************************************************************
int ProcessAccess(int i, int num, char **tokens, int state)
{
    int index;
    int action = state & 15;
    char *xstr = nullstr;
    int postop = -1;

    // Determine the size
    if (StrCompNoCase(tokens[i], "byte")) index = 0;
    else if (StrCompNoCase(tokens[i], "word")) index = 1;
    else index = 2;

    CheckExpect(state, EXPECT_VAL, tokens[i]);

    if (++i >= num)
    {
	printf("Encountered EOL\n");
	ErrorExit();
    }

    if (strcmp(tokens[i], "["))
    {
	printf("Expected '[' - %s\n", tokens[i]);
	ErrorExit();
    }
    if (++i >= num)
    {
	printf("Encountered EOL\n");
	ErrorExit();
    }

    // Get the value between the brackets
    i = ProcessExpression(i, num, tokens, 12, ACTION_LOAD);

    if (strcmp(tokens[i], "]"))
    {
	printf("Expected ']' - %s\n", tokens[i]);
	ErrorExit();
    }

    // Check for index
    if (++i < num && tokens[i][0] == '[')
    {
	i = ProcessExpression(i+1, num, tokens, 12, ACTION_LOAD);
        if (strcmp(tokens[i], "]"))
        {
	    printf("Expected ']' - %s\n", tokens[i]);
	    ErrorExit();
        }
	i++;
	xstr = "x";
    }

    // Check for post operator
    if (i < num)
    {
	postop = FindSymbolType(tokens[i]);
	if (symtable[postop].type & TYPE_POSTOP)
	    action = ACTION_EXEC;
	else
	    postop = -1;
    }

    // Peform the memory operation
    if (action == ACTION_LOAD)
        fprintf(outfile, "        ld%ca%s\n", AccessChar[index], xstr);
    else if (action == ACTION_STORE)
        fprintf(outfile, "        st%ca%s\n", AccessChar[index], xstr);
    else if (action == ACTION_ADDR)
        fprintf(outfile, "        la%ca%s\n", AccessChar[index], xstr);
    else if (action == ACTION_EXEC)
    {
        fprintf(outfile, "        ex%ca%s", AccessChar[index], xstr);
	if (postop != -1)
        {
#ifdef NO_FUNCTION_PTR
	    i = ExecuteFunction(symtable[postop].func, i, num, tokens, (EXPECT_OP << 8) | (state & 255));
#else
	    i = (symtable[postop].func)(i, num, tokens, (EXPECT_OP << 8) | (state & 255));
#endif
        }
    }

    return i;
}

// *********************************************************************
// This routine should never be called
// *********************************************************************
int ProcessOp(int i, int num, char **tokens, int state)
{
    printf("OP\n");
    return i + 1;
}

// *********************************************************************
// This routine handles the binary operators.  It will return if
// the current operator has greater than or equal precedence than
// the previous opertor.  Otherwise, is will call ProcessExpression
// to recursively process the next part of the expression.
// *********************************************************************
int ProcessBinaryOp(int i, int num, char **tokens, int state)
{
    int j = FindSymbolType(tokens[i]);
    int prec1 = (state >> 4) & 15;
    int prec2 = symtable[j].type & 15;

    CheckExpect(state, EXPECT_OP, tokens[i]);

#ifdef PRINT_PROCESS
    printf("ProcessBinaryOp tokens[%d] = %s, %d %d\n", i, tokens[i], prec1, prec2);
#endif

    // This test is not needed, and it should never be true.
    // However, I'm keeping it for now for debug purposes.
    if (prec1 <= prec2)
    {
	printf("XXXXXXXXXX ProcessBinaryOp XXXXXXXXXX\n");
	return i;
    }

    if (++i >= num)
    {
	printf("Expected value, found EOL\n");
	ErrorExit();
    }
    i = ProcessExpression(i, num, tokens, prec2, ACTION_LOAD);
    fprintf(outfile, "        %s\n", (char *)symtable[j].parm);

    return i;
}

// *********************************************************************
// This routine handles the binary assignment operators
// *********************************************************************
int ProcessBinaryEqOp(int i, int num, char **tokens, int state)
{
    int j = FindSymbolType(tokens[i]);
    //int prec1 = (state >> 4) & 15;
    //int prec2 = symtable[j].type & 15;

    CheckExpect(state, EXPECT_OP, tokens[i]);

    //if (prec1 <= prec2) return i;

    if (i + 1 >= num)
    {
	printf("Expected value, found EOL\n");
	ErrorExit();
    }

    //i = ExecuteSymbol(i, num, tokens, EXPECT_VAL, prec2, ACTION_LOAD);
    fprintf(outfile, " %s\n", (char *)symtable[j].parm);

    return i;
}

// *********************************************************************
// This routine handles the non-assignment unary operators
// *********************************************************************
int ProcessUnaryOp(int i, int num, char **tokens, int state)
{
    int j = FindSymbolType(tokens[i]);
    int prec2 = symtable[j].type & 15;

    CheckExpect(state, EXPECT_VAL, tokens[i]);

    if (++i >= num)
    {
	printf("Expected value, found EOL\n");
	ErrorExit();
    }

    i = ExecuteSymbol(i, num, tokens, EXPECT_VAL, prec2, ACTION_LOAD);
    fprintf(outfile, "        %s\n", (char *)symtable[j].parm);

    return i;
}

// *********************************************************************
// This routine handles the extra unary assignment operators
// *********************************************************************
int ProcessExtraOp(int i, int num, char **tokens, int state)
{
    int j = FindSymbolType(tokens[i]);
    int prec = (state >> 4) & 15;
    int action = state & 15;
    int expect = state >> 8;
    char opstr[20];
    char *ptr1 = (char *)(int)symtable[j].parm;
    char *ptr2 = FindChar(ptr1, '-');
    char *lstr = nullstr;
    int len;

    if (action == ACTION_LOAD) lstr = " load";

    // Check if it is a pre-op
    if (expect == EXPECT_VAL)
    {
	len = ptr2 - ptr1;
	memcpy(opstr, ptr1, len);
	opstr[len] = 0;
    }

    // Otherwise, it's a post-op
    else
    {
        fprintf(outfile, " %s%s\n", ptr2+1, lstr);
        return i + 1;
    }

    if (++i >= num)
    {
	printf("Expected value, found EOL\n");
	ErrorExit();
    }

    i = ExecuteSymbol(i, num, tokens, EXPECT_VAL, prec, ACTION_EXEC);

    fprintf(outfile, " %s%s\n", opstr, lstr);

    return i;
}

// *********************************************************************
// This routine handles the "@" address operator
// *********************************************************************
int ProcessAt(int i, int num, char **tokens, int state)
{
    int j;
    int action = state & 15;

    //printf("ProcessAt %d %8.8x\n", action, state);

    if (action != ACTION_LOAD)
    {
	printf("Invalid action for address operator - %d\n", action);
	ErrorExit();
    }

    if (++i >= num)
    {
	printf("Expected variable - EOL\n");
	ErrorExit();
    }

    j = FindSymbolType(tokens[i]);
    state = (state & 0xffffff00) | ACTION_ADDR;
#ifdef NO_FUNCTION_PTR
    i = ExecuteFunction(symtable[j].func, i, num, tokens, state);
#else
    i = (symtable[j].func)(i, num, tokens, state);
#endif

    return i;
}

// *********************************************************************
// This routine handles intrinsic functions and calling methods
// *********************************************************************
int ProcessFunction(int i, int num, char **tokens, int state, int intrinsic)
{
    int first = 1;
    char *pubname = tokens[i];
    int action = state & 15;
    int prec = (state >> 4) & 15;

    if (intrinsic != 1)
    {
        if (action == ACTION_LOAD)
            fprintf(outfile, "        ldfrmr\n");
        else
            fprintf(outfile, "        ldfrm\n");
    }
    else if (StrCompNoCase(pubname, "cognew"))
    {
        pubname = "coginit";
        fprintf(outfile, "        ldl 8\n");
    }

    if (++i < num && tokens[i][0] == '(')
    {
        i++;
        while (i < num && tokens[i][0] != ')')
        {
	    if (first) first = 0;
	    else
	    {
	        if (strcmp(tokens[i], ","))
		{
	            printf("Expected ',' or ')' - %s\n", tokens[i]);
	            ErrorExit();
		}
	        i++;
	    }
	    i = ProcessExpression(i, num, tokens, 12, ACTION_LOAD);
        }
    }

    if (intrinsic == 1)
    {
        if (action == ACTION_LOAD && FindListNoCase(pubname, ret_list) >= 0)
            fprintf(outfile, "        %sret\n", pubname);
        else
            fprintf(outfile, "        %s\n", pubname);
    }
    else if (intrinsic == 0)
        fprintf(outfile, "        call %s\n", pubname);
    else
        fprintf(outfile, "        callobj %s\n", pubname);

    if (++i < num)
    {
        i = ExecuteSymbol(i, num, tokens, EXPECT_OP, prec, action);
    }

    return i;
}

// *********************************************************************
// This is used to call methods in external objects
// *********************************************************************
int ProcessCallObj(int i, int num, char **tokens, int state)
{
    return ProcessFunction(i, num, tokens, state, 3);
}

// *********************************************************************
// This is used to call local methods
// *********************************************************************
int ProcessPub(int i, int num, char **tokens, int state)
{
    return ProcessFunction(i, num, tokens, state, 0);
}

// *********************************************************************
// This handles intrinsic functions
// *********************************************************************
int ProcessIntrinsic(int i, int num, char **tokens, int state)
{
    return ProcessFunction(i, num, tokens, state, 1);
}

// *********************************************************************
// This handles tokens that are variables, numbers or constants
// *********************************************************************
int ProcessValue(int i, int num, char **tokens, int state)
{
    int j;
    char *xstr = nullstr;
    char *vstr = tokens[i];
    char *sstr;
    int action = state & 15;
    int action1 = action;
    int prec = (state >> 4) & 15;
    int postop = -1;

    // Determine the variable size
    j = FindSymbolType(tokens[i]);
    if (symtable[j].type == TYPE_REG)
        sstr = "reg";
    else if (symtable[j].size == 1)
        sstr = "b";
    else if (symtable[j].size == 2)
        sstr = "w";
    else
        sstr = "l";

    //printf("ProcessValue %s\n", tokens[i]);
#ifdef PRINT_PROCESS
    printf("ProcessValue tokens[%d] = %s\n", i, tokens[i]);
#endif

    CheckExpect(state, EXPECT_VAL, tokens[i]);

    // Check for index
    if (++i < num && tokens[i][0] == '[')
    {
	i = ProcessExpression(i+1, num, tokens, 12, ACTION_LOAD);
        if (strcmp(tokens[i], "]"))
        {
	    printf("Expected ']' - %s\n", tokens[i]);
	    ErrorExit();
        }
	i++;
	xstr = "x";
    }

    // Check for post operator
    if (i < num)
    {
	postop = FindSymbolType(tokens[i]);
	if (symtable[postop].type & TYPE_POSTOP)
	    action = ACTION_EXEC;
	else
	    postop = -1;
    }

    // Perform action
    if (action == ACTION_LOAD)
        fprintf(outfile, "        ld%s%s %s\n", sstr, xstr, vstr);
    else if (action == ACTION_STORE)
        fprintf(outfile, "        st%s%s %s\n", sstr, xstr, vstr);
    else if (action == ACTION_ADDR)
        fprintf(outfile, "        la%s%s %s\n", sstr, xstr, vstr);
    else if (action == ACTION_EXEC)
    {
        fprintf(outfile, "        ex%s%s %s", sstr, xstr, vstr);
        if (postop != -1)
        {
#ifdef NO_FUNCTION_PTR
	    i = ExecuteFunction(symtable[postop].func, i, num, tokens, (EXPECT_OP << 8) | (prec << 4) | action1);
#else
	    i = (symtable[postop].func)(i, num, tokens, (EXPECT_OP << 8) | (prec << 4) | action1);
#endif
        }
    }

    return i;
}

// *********************************************************************
// This handles the right parenthesis and bracket and the comma
// *********************************************************************
int ProcessRightBracket(int i, int num, char **tokens, int state)
{
    CheckExpect(state, EXPECT_OP, tokens[i]);
    return i;
}

// *********************************************************************
// This handles the clkfreq variable
// *********************************************************************
int ProcessClkfreq(int i, int num, char **tokens, int state)
{
    int action = state & 15;
    fprintf(outfile, "        ldli0\n");
    if (action == ACTION_LOAD)
        fprintf(outfile, "        ldla\n");
    else if (action == ACTION_STORE)
        fprintf(outfile, "        stla\n");
    return i + 1;
}

// *********************************************************************
// This routine initializes the symbol table with
// the standard Spin tokens.
// *********************************************************************
#if 0
void AddKeywords(void)
{
    // Entry zero is used for numbers
    AddSymbol("0", TYPE_NUM, fProcessValue, 0);

    // Spin keywords
    AddSymbol("result",  TYPE_LOCAL, fProcessValue, 0);
    AddSymbol("clkfreq", 0, fProcessClkfreq, 0);
    AddSymbol("byte",    TYPE_ACCESS, fProcessAccess, 0);
    AddSymbol("word",    TYPE_ACCESS, fProcessAccess, 0);
    AddSymbol("long",    TYPE_ACCESS, fProcessAccess, 0);
    AddSymbol("repeat",  TYPE_KEY, fProcessRepeat, 0);
    AddSymbol("while",   TYPE_KEY, fProcessKey, 0);
    AddSymbol("case",    TYPE_KEY, fProcessCase, 0);
    AddSymbol("if",      TYPE_KEY, fProcessIf, 0);
    AddSymbol("ifnot",   TYPE_KEY, fProcessIf, 0);
    AddSymbol("elseif",  TYPE_KEY, fProcessElseIf, 0);
    AddSymbol("elseifnot", TYPE_KEY, fProcessElseIf, 0);
    AddSymbol("else",    TYPE_KEY, fProcessElse, 0);
    AddSymbol("quit",    TYPE_KEY, fProcessQuit, 0);
    AddSymbol("next",    TYPE_KEY, fProcessNext, 0);
    AddSymbol("return",  TYPE_KEY, fProcessReturn, 0);
    AddSymbol("string",  TYPE_PUB, fProcessString, 0);

    // Binary Operators - The 4 LSBs of the type is the precedence
    AddSymbol("-",   TYPE_OP +  6, fProcessBinaryOp, (int)"sub");
    AddSymbol("+",   TYPE_OP +  6, fProcessBinaryOp, (int)"add");
    AddSymbol("/",   TYPE_OP +  5, fProcessBinaryOp, (int)"div");
    AddSymbol("*",   TYPE_OP +  5, fProcessBinaryOp, (int)"mul");
    AddSymbol("**",  TYPE_OP +  5, fProcessBinaryOp, (int)"mulh");
    AddSymbol("//",  TYPE_OP +  5, fProcessBinaryOp, (int)"mod");
    AddSymbol("#>",  TYPE_OP +  7, fProcessBinaryOp, (int)"minlim");
    AddSymbol("<#",  TYPE_OP +  7, fProcessBinaryOp, (int)"maxlim");
    AddSymbol("~>",  TYPE_OP +  2, fProcessBinaryOp, (int)"sar");
    AddSymbol("<<",  TYPE_OP +  2, fProcessBinaryOp, (int)"shl");
    AddSymbol(">>",  TYPE_OP +  2, fProcessBinaryOp, (int)"shr");
    AddSymbol("->",  TYPE_OP +  2, fProcessBinaryOp, (int)"ror");
    AddSymbol("<-",  TYPE_OP +  2, fProcessBinaryOp, (int)"rol");
    AddSymbol("><",  TYPE_OP +  2, fProcessBinaryOp, (int)"rev");
    AddSymbol("&",   TYPE_OP +  3, fProcessBinaryOp, (int)"and");
    AddSymbol("|",   TYPE_OP +  4, fProcessBinaryOp, (int)"or");
    AddSymbol("^",   TYPE_OP +  4, fProcessBinaryOp, (int)"xor");
    AddSymbol("and", TYPE_OP + 10, fProcessBinaryOp, (int)"andl");
    AddSymbol("or",  TYPE_OP + 11, fProcessBinaryOp, (int)"orl");
    AddSymbol("==",  TYPE_OP +  8, fProcessBinaryOp, (int)"cmpeq");
    AddSymbol("<>",  TYPE_OP +  8, fProcessBinaryOp, (int)"cmpne");
    AddSymbol("<",   TYPE_OP +  8, fProcessBinaryOp, (int)"cmplt");
    AddSymbol(">",   TYPE_OP +  8, fProcessBinaryOp, (int)"cmpgt");
    AddSymbol("=<",  TYPE_OP +  8, fProcessBinaryOp, (int)"cmple");
    AddSymbol("=>",  TYPE_OP +  8, fProcessBinaryOp, (int)"cmpge");

    // BinaryEqual Operators
    AddSymbol("-=",   TYPE_POSTOP +  6, fProcessBinaryEqOp, (int)"sub");
    AddSymbol("+=",   TYPE_POSTOP +  6, fProcessBinaryEqOp, (int)"add");
    AddSymbol("/=",   TYPE_POSTOP +  5, fProcessBinaryEqOp, (int)"div");
    AddSymbol("*=",   TYPE_POSTOP +  5, fProcessBinaryEqOp, (int)"mul");
    AddSymbol("**=",  TYPE_POSTOP +  5, fProcessBinaryEqOp, (int)"mulh");
    AddSymbol("//=",  TYPE_POSTOP +  5, fProcessBinaryEqOp, (int)"mod");
    AddSymbol("#>=",  TYPE_POSTOP +  7, fProcessBinaryEqOp, (int)"minlim");
    AddSymbol("<#=",  TYPE_POSTOP +  7, fProcessBinaryEqOp, (int)"maxlim");
    AddSymbol("~>=",  TYPE_POSTOP +  2, fProcessBinaryEqOp, (int)"sar");
    AddSymbol("<<=",  TYPE_POSTOP +  2, fProcessBinaryEqOp, (int)"shl");
    AddSymbol(">>=",  TYPE_POSTOP +  2, fProcessBinaryEqOp, (int)"shr");
    AddSymbol("->=",  TYPE_POSTOP +  2, fProcessBinaryEqOp, (int)"ror");
    AddSymbol("<-=",  TYPE_POSTOP +  2, fProcessBinaryEqOp, (int)"rol");
    AddSymbol("><=",  TYPE_POSTOP +  2, fProcessBinaryEqOp, (int)"rev");
    AddSymbol("&=",   TYPE_POSTOP +  3, fProcessBinaryEqOp, (int)"and");
    AddSymbol("|=",   TYPE_POSTOP +  4, fProcessBinaryEqOp, (int)"or");
    AddSymbol("^=",   TYPE_POSTOP +  4, fProcessBinaryEqOp, (int)"xor");
    AddSymbol("and=", TYPE_POSTOP + 10, fProcessBinaryEqOp, (int)"andl");
    AddSymbol("or=",  TYPE_POSTOP + 11, fProcessBinaryEqOp, (int)"orl");
    AddSymbol("===",  TYPE_POSTOP +  8, fProcessBinaryEqOp, (int)"cmpeq");
    AddSymbol("<>=",  TYPE_POSTOP +  8, fProcessBinaryEqOp, (int)"cmpne");
    AddSymbol("<=",   TYPE_POSTOP +  8, fProcessBinaryEqOp, (int)"cmplt");
    AddSymbol(">=",   TYPE_POSTOP +  8, fProcessBinaryEqOp, (int)"cmpgt");
    AddSymbol("=<=",  TYPE_POSTOP +  8, fProcessBinaryEqOp, (int)"cmple");
    AddSymbol("=>=",  TYPE_POSTOP +  8, fProcessBinaryEqOp, (int)"cmpge");
    AddSymbol(":=",   TYPE_OP + 12,     fProcessOp,         (int)nullstr);

    // Unary Operators
    AddSymbol("^^",  TYPE_OP +  1, fProcessUnaryOp, (int)"sqrt");
    AddSymbol("||",  TYPE_OP +  1, fProcessUnaryOp, (int)"abs");
    AddSymbol("|<",  TYPE_OP +  1, fProcessUnaryOp, (int)"encode");
    AddSymbol(">|",  TYPE_OP +  1, fProcessUnaryOp, (int)"decode");
    AddSymbol("not", TYPE_OP +  9, fProcessUnaryOp, (int)"notl");
    AddSymbol("!",   TYPE_OP +  1, fProcessUnaryOp, (int)"comp");
    AddSymbol("@",   TYPE_OP +  0, fProcessAt,      (int)nullstr);

    // Extra Operators
    AddSymbol("++",   TYPE_POSTOP +  1, fProcessExtraOp, (int)"preinc-postinc");
    AddSymbol("--",   TYPE_POSTOP +  1, fProcessExtraOp, (int)"predec-postdec");
    AddSymbol("~",    TYPE_POSTOP +  1, fProcessExtraOp, (int)"sexb-postclr");
    AddSymbol("~~",   TYPE_POSTOP +  1, fProcessExtraOp, (int)"sexw-postset");
    AddSymbol("?",    TYPE_POSTOP +  1, fProcessExtraOp, (int)"randf-randr");

    // Seperators
    AddSymbol("(",  0, fProcessLeftParen, 0);
    AddSymbol(")",  0, fProcessRightBracket, 0);
    AddSymbol("]",  0, fProcessRightBracket, 0);
    AddSymbol(",",  0, fProcessRightBracket, 0);
    AddSymbol(":",  0, fProcessRightBracket, 0);

    // Intrinsic functions
    AddSymbol("strsize",  TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("strcomp",  TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("bytemove", TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("wordmove", TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("longmove", TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("bytefill", TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("wordfill", TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("longfill", TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("cogid",    TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("cognew",   TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("coginit",  TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("cogstop",  TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("reboot",   TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("locknew",  TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("lockret",  TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("lockclr",  TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("lockset",  TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("waitcnt",  TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("waitpeq",  TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("waitpne",  TYPE_PUB, fProcessIntrinsic, 0);
    AddSymbol("waitvid",  TYPE_PUB, fProcessIntrinsic, 0);

    // Registers
    AddSymbol("par",  TYPE_REG, fProcessValue, 0x1f0);
    AddSymbol("cnt",  TYPE_REG, fProcessValue, 0x1f1);
    AddSymbol("ina",  TYPE_REG, fProcessValue, 0x1f2);
    AddSymbol("inb",  TYPE_REG, fProcessValue, 0x1f3);
    AddSymbol("outa", TYPE_REG, fProcessValue, 0x1f4);
    AddSymbol("outb", TYPE_REG, fProcessValue, 0x1f5);
    AddSymbol("dira", TYPE_REG, fProcessValue, 0x1f6);
    AddSymbol("dirb", TYPE_REG, fProcessValue, 0x1f7);
    AddSymbol("ctra", TYPE_REG, fProcessValue, 0x1f8);
    AddSymbol("ctrb", TYPE_REG, fProcessValue, 0x1f9);
    AddSymbol("frqa", TYPE_REG, fProcessValue, 0x1fa);
    AddSymbol("frqb", TYPE_REG, fProcessValue, 0x1fb);
    AddSymbol("phsa", TYPE_REG, fProcessValue, 0x1fc);
    AddSymbol("phsb", TYPE_REG, fProcessValue, 0x1fd);
    AddSymbol("vcfg", TYPE_REG, fProcessValue, 0x1fe);
    AddSymbol("vscl", TYPE_REG, fProcessValue, 0x1ff);

    // Constants
    AddSymbol("false",  TYPE_CON, fProcessValue, 0);
    AddSymbol("true",   TYPE_CON, fProcessValue, 0xffffffff);
    AddSymbol("negx",   TYPE_CON, fProcessValue, 0x80000000);
    AddSymbol("posx",   TYPE_CON, fProcessValue, 0x7fffffff);
    AddSymbol("pi",     TYPE_CON, fProcessValue, 0x40490fdb);
    AddSymbol("rcfast", TYPE_CON, fProcessValue, 0x001);
    AddSymbol("rcslow", TYPE_CON, fProcessValue, 0x002);
    AddSymbol("xinput", TYPE_CON, fProcessValue, 0x004);
    AddSymbol("xtal1",  TYPE_CON, fProcessValue, 0x008);
    AddSymbol("xtal2",  TYPE_CON, fProcessValue, 0x010);
    AddSymbol("xtal3",  TYPE_CON, fProcessValue, 0x020);
    AddSymbol("pll1x",  TYPE_CON, fProcessValue, 0x040);
    AddSymbol("pll2x",  TYPE_CON, fProcessValue, 0x080);
    AddSymbol("pll4x",  TYPE_CON, fProcessValue, 0x100);
    AddSymbol("pll8x",  TYPE_CON, fProcessValue, 0x200);
    AddSymbol("pll16x", TYPE_CON, fProcessValue, 0x400);

    DumpSymbols();
}
#endif

#if 0
void DumpSymbols(void)
{
    int i;
    SymbolT *ptr;

    for (i = 0; i < symnum; i++)
    {
        ptr = &symtable[i];
        printf("%d %d %s ", ptr->type, ptr->func, ptr->name);
        if (ptr->type >= 256)
            printf("%s\n", ptr->parm);
        else
            printf("%d\n", ptr->parm & 0xffff);
    }
}
#endif

#if 1
void AddKeywords(void)
{
    int i, j, k;
    char *ptr;

    if (debugflag)
        printf("Adding Spin keywords to the symbol table\n");
#ifdef CSPIN_FLAG
    infile = fopen("/bin/spinsymb.txt", "r");
#else
    infile = fopen("spinsymb.txt", "r");
#endif
    if (!infile)
    {
#ifdef CSPIN_FLAG
        printf("Could not open /bin/spinsymb.txt\n");
#else
        printf("Could not open spinsymb.txt\n");
#endif
        exit(1);
    }

    while (fgets(buffer, 100, infile))
    {
        RemoveCRLF(buffer);
        tokens[0] = ptr = buffer;
        for (i = 1; i < 4; i++)
        {
            ptr = FindChar(ptr, ' ');
            *ptr++ = 0;
            tokens[i] = ptr;
        }
        i = atol(tokens[0]);
        j = atol(tokens[1]);
        if (isdigit((int)tokens[3][0]))
        {
            k = atol(tokens[3]);
        }
        else
        {
            k = (int)StrAlloc(strlen(tokens[3]) + 1);
            if (symnum >= MAX_SYMBOLS)
            {
                printf("Symbol table is full\n");
                ErrorExit();
            }
            strcpy((char *)k, tokens[3]);
        }
        AddSymbol(tokens[2], i, j, k);
    }
    fclose(infile);
    infile = 0;
    //DumpSymbols();
}

#ifdef CSPIN_FLAG
int atol(char *ptr)
{
    int val = 0;
    while(*ptr)
      val = (val * 10) + *ptr++ - '0';
    return val;
}  
#endif
#endif

// *********************************************************************
// This routine handles lines that use the { or {{ comment delimeter.
// Currently, the entire line is treated as a comment.
// *********************************************************************
int CommentLine(char *str)
{
    int retval = commentflag;

    if (commentflag == 1) // Search for a '}' character
    {
        str = FindChar(str, '}');
	if (*str == '}')
	    commentflag = 0;
    }
    else if (commentflag == 2) // Search for a "}}" string
    {
        str = FindStr(str, "}}");
        if (str[0]  == '}')
            commentflag = 0;
    }
    else // Check first nonblank character for a '{'
    {
	str = SkipChars(str, " \t");
	if (str[0] == '{')
        {
	    if (str[1] == '{')
		commentflag = 2;
	    else
		commentflag = 1;
	}
	retval = commentflag;
    }
    return retval;
}

// *********************************************************************
// This routine executes the first pass through the input
// file.  It extracts the global methods, variables and
// constants and adds them to the symbol table.  Every
// symbol that is used by the Spin prgram should be in
// the symbol table after this pass, except for the local
// variables, which are added at the beginning of each
// method.
// *********************************************************************
void GetGlobalSymbols(void)
{
    int i, num;
    char *ptr;
    int mode = TYPE_CON;

    commentflag = 0;
    linenum = 0;

    printf("Pass 1: Compiling Global Symbols\n");

    while (fgets(buffer, 100, infile))
    {
        linenum++;
        //c.printf1(string("1: %s\n"), @buffer)
	RemoveCRLF(buffer);
	if (CommentLine(buffer)) continue;
	num = GetTokens(buffer, buf2, tokens);
	if (num == 0) continue;
	if (tokens[0][0] == SINGLE_QUOTE) continue;
	if (StrCompNoCase(tokens[0], "pub"))
	{
	    mode = TYPE_PUB;
            AddSymbolCheck(tokens[1], mode, fProcessPub, 0);
	}
	else if (StrCompNoCase(tokens[0], "con"))
	{
	    mode = TYPE_CON;
	}
	else if (StrCompNoCase(tokens[0], "dat"))
	{
	    mode = TYPE_DAT;
	}
	else if (StrCompNoCase(tokens[0], "var"))
	{
	    mode = TYPE_VAR;
	}
	else if (StrCompNoCase(tokens[0], "obj"))
	{
	    mode = TYPE_OBJECT;
	}
	else if (mode == TYPE_CON)
	{
	    if (num >= 3 && strcmp(tokens[1], "=") == 0)
	        AddSymbolCheck(tokens[0], mode, fProcessValue, 0);
	}
	else if (mode == TYPE_DAT)
	{
	    if (!StrCompNoCase(tokens[0], "byte") &&
	        !StrCompNoCase(tokens[0], "word") && !StrCompNoCase(tokens[0], "long"))
	    {
	        AddSymbolCheck(tokens[0], mode, fProcessValue, 0);
		if (num > 1)
		{
		    if (StrCompNoCase(tokens[1], "byte"))
		        symtable[symnum-1].size = 1;
		    else if (StrCompNoCase(tokens[1], "word"))
		        symtable[symnum-1].size = 2;
		}
	    }
	}
	else if (mode == TYPE_VAR)
	{
	    if (num >= 2 && (StrCompNoCase(tokens[0], "byte") ||
	        StrCompNoCase(tokens[0], "word") || StrCompNoCase(tokens[0], "long")))
	        AddSymbolCheck(tokens[1], mode, fProcessValue, 0);
	    if (StrCompNoCase(tokens[0], "byte"))
	        symtable[symnum-1].size = 1;
	    else if (StrCompNoCase(tokens[0], "word"))
	        symtable[symnum-1].size = 2;
	}
	else if (mode == TYPE_OBJECT)
	{
	    AddObject(num, tokens);
	}
        else if (mode == TYPE_PUB)
        {
            for (i = 0; i < num; i++)
            {
                if (tokens[i][0] == SINGLE_QUOTE) continue;
                if (tokens[i][0] == DOUBLE_QUOTE) continue;
                ptr = FindChar(tokens[i], '.');
                if (*ptr)
                {
                    if (debugflag)
                        printf("Found external method %s\n", tokens[i]);
                    if (FindSymbol(tokens[i]) < 0)
                        AddSymbolCheck(tokens[i], TYPE_PUB, fProcessCallObj, 0);
                    continue;
                }
                ptr = FindChar(tokens[i], '#');
                if (*ptr)
                {
                    printf("Found external constant %s\n", tokens[i]);
                    if (FindSymbol(tokens[i]) < 0)
                        AddSymbolCheck(tokens[i], TYPE_CON, fProcessValue, 0);
                    continue;
                }
            }
        }
    }

    // Save the index for the start of the local symbols
    symnum_save = symnum;
}

// *********************************************************************
// This routine processes the PUB statement.  It sets the level
// to zero and initializes the level variables.  It also adds
// the local variables to the symbol table.
// *********************************************************************
int ProcessPubStatement(int i, int num, char **tokens)
{
    char *ptr;
    int skipflag = 0;

    symnum = symnum_save;

    // Initialize level zero
    level = 0;
    indent[level] = 0;
    labelstart[level] = 0;
    labelend[level] = 0;
    looplevel[level] = 0;
    looptype[level] = LOOP_PUB;

    // Add the local variables to the symbol table
    for (;i < num; i++)
    {
	ptr = FindChar("(),|:", tokens[i][0]);
	if (*ptr) continue;
	if (skipflag)
	{
	    if (tokens[i][0] == ']') skipflag = 0;
	}
	else if (tokens[i][0] == '[')
	    skipflag = 1;
	else
	    AddSymbolCheck(tokens[i], TYPE_LOCAL, fProcessValue, 0);
    }
    return i;
}

// *********************************************************************
// This routine should never be called
// *********************************************************************
int CatchObject(int i, int num, char **tokens, int state)
{
    return i + 1;
}

// *********************************************************************
// This routine is used during the first pass to add the object prefixes
// to the symbol table.  To save space, the objects themselves are not
// examined, and their globals are not added to the symbol table.
// *********************************************************************
void AddObject(int num, char **tokens)
{
    //char *ptr;

    if (num < 1 || tokens[0][0] == SINGLE_QUOTE) return;
    if (num != 3 || strcmp(tokens[1], ":"))
    {
        printf("Error in object line\n");
        ErrorExit();
    }
    //ptr = StrAlloc(strlen(tokens[2]) + 1);
    //strcpy(ptr, tokens[2]);
    //printf("ProcessObject: AddSymbol %s\n", tokens[0]);
    //AddSymbolCheck(tokens[0], TYPE_OBJECT, fCatchObject, (int)ptr);
    AddSymbolCheck(tokens[0], TYPE_OBJECT, fCatchObject, 0);
}
#if 0
// *********************************************************************
// This routine is used during the first passed to extract
// the constants and public methods from an external object
// and add them to the symbol table.
// *********************************************************************
void ProcessObject(int num, char **tokens)
{
    int len;
    int objnum;
    FILE *objfile;
    char objname[20];
    int mode = TYPE_CON;

    if (num < 1 || tokens[0][0] == SINGLE_QUOTE) return;
    if (num != 3 || strcmp(tokens[1], ":"))
    {
	printf("Error in object line\n");
        ErrorExit();
    }
    //printf("ProcessObject: AddSymbol %s\n", tokens[0]);
    objnum = AddSymbolCheck(tokens[0], TYPE_OBJECT, fCatchObject, 0);
    strcpy(objname, tokens[2] + 1);
    len = strlen(objname);
    if (len >= 1) objname[len-1] = 0;
    strcat(objname, ".spn");
    //printf("ProcessObject %s %s %s\n", tokens[0], tokens[1], tokens[2]);
    objfile = fopen(objname, "r");
    if (!objfile)
    {
	printf("Could not open object file %s\n", tokens[2]);
        ErrorExit();
    }

    while (fgets(buffer, 100, objfile))
    {
	RemoveCRLF(buffer);
	num = GetTokens(buffer, buf2, tokens);
	if (StrCompNoCase(tokens[0], "pub"))
	{
	    mode = TYPE_PUB;
	    if (num >= 2)
	    {
		strcpy(objname, symtable[objnum].name);
		strcat(objname, ".");
		strcat(objname, tokens[1]);
		//printf("ProcessObject: AddSymbol %s\n", objname);
                AddSymbolCheck(objname, mode, fProcessCallObj, 0);
	    }
	}
	else if (StrCompNoCase(tokens[0], "pri"))
	    mode = TYPE_PUB;
	else if (StrCompNoCase(tokens[0], "obj"))
	    mode = TYPE_OBJECT;
	else if (StrCompNoCase(tokens[0], "dat"))
	    mode = TYPE_DAT;
	else if (StrCompNoCase(tokens[0], "var"))
	    mode = TYPE_VAR;
	else if (StrCompNoCase(tokens[0], "con"))
	    mode = TYPE_CON;
	else if (mode == TYPE_CON)
	{
	    if (num >= 2 && strcmp(tokens[1], "=") == 0)
	    {
		strcpy(objname, symtable[objnum].name);
		strcat(objname, "#");
		strcat(objname, tokens[0]);
		//printf("ProcessObject: AddSymbol %s\n", objname);
                AddSymbolCheck(objname, mode, fProcessValue, 0);
	    }
	}
    }
    fclose(objfile);
}
#endif

// *********************************************************************
// This routine handles the left parenthesis by calling
// ProcessExpression with the highest operator precedence
// and an action of "load".  ProcessExpression must return
// pointing at a right parenthesis.
// *********************************************************************
int ProcessLeftParen(int i, int num, char **tokens, int state)
{
    i = ProcessExpression(i+1, num, tokens, 12, ACTION_LOAD);
    if (strcmp(tokens[i], ")"))
    {
        printf("Expected ')' - %s\n", tokens[i]);
        ErrorExit();
    }
    return i + 1;
}

// *********************************************************************
// This routine processes an expression by first processing a value,
// and then processing operators in a loop.  The operator functions
// will recursively call this routine to execute higher precedent
// operations.  The loop terminates when we hit the end of the line,
// a right parenthesis or bracket, a comma, or an operator with greater
// or equal precedence.
// *********************************************************************
int ProcessExpression(int i, int num, char **tokens, int prec, int action)
{
#ifdef PRINT_PROCESS
    printf("ProcessExpression0 tokens[%d] = %s, action = %d\n", i, tokens[i], action);
#endif
    i = ExecuteSymbol(i, num, tokens, EXPECT_VAL, prec, action);
    while (i < num)
    {
        if (prec <= GetSymbolPrecedence(tokens[i]))
            break;
        if (tokens[i][0] == ':')
            break;
        if (tokens[i][0] == ')')
            break;
        if (tokens[i][0] == ']')
            break;
        if (tokens[i][0] == ',')
            break;
#ifdef PRINT_PROCESS
        printf("ProcessExpression1 tokens[%d] = %s\n", i, tokens[i]);
#endif
        i = ExecuteSymbol(i, num, tokens, EXPECT_OP, prec, ACTION_LOAD);
    }
    return i;
}

// *********************************************************************
// This routine process a case item statement
// *********************************************************************
int ProcessCaseItem(int i, int num, char **tokens)
{
    int j;

    // Process items to the left of the colon
    while (i < num)
    {
        i = ProcessExpression(i, num, tokens, 12, ACTION_LOAD);
	if (i < num)
	{
	    if (tokens[i][0] == ',')
	        i++;
	    else
	        break;
	}
    }

    // Loop must have exited at a colon
    if (i >= num || tokens[i][0] != ':')
    {
        printf("Expected a colon, but encountered EOL\n");
        ErrorExit();
    }
    i++;

    // Increment the level
    labelstart[++level] = 0;
    labelend[level] = labelnum;
    labelstart[level] = 0;
    indent[level] = indent2;
    looplevel[level] = looplevel[level-1];
    looptype[level] = 0;

    // Process inline statement
    if (i < num)
    {
	j = LocateAssignment(num, tokens);
	if (j < num)
	    ProcessStore(0, num, tokens, j);
	else
	    ProcessNonStore(0, num, tokens);
    }

    return i;
}

// *********************************************************************
// This routine process a statement that contains an assignment
// operator.
// *********************************************************************
int ProcessStore(int i, int num, char **tokens, int assignpos)
{
    int l;

    if (assignpos + 1 >= num)
    {
	printf("Expected value, found EOL\n");
        ErrorExit();
    }

    // Process the right side of the statement
    l = ProcessExpression(assignpos+1, num, tokens, 12, ACTION_LOAD);

    // Process the left side
    i = ExecuteSymbol(i, num, tokens, EXPECT_VAL, 12, ACTION_STORE);

    if (i != assignpos)
    {
	printf("Expected assignment op, found %s\n", tokens[i]);
        ErrorExit();
    }

    return l;
}

// *********************************************************************
// This routine processes a statement that does not contain an
// assignment operator.
// *********************************************************************
int ProcessNonStore(int i, int num, char **tokens)
{
    i = ProcessExpression(i, num, tokens, 12, ACTION_NONE);
    if (i < num)
    {
	printf("Expected EOL, found %s\n", tokens[i]);
        ErrorExit();
    }
    return i;
}

// *********************************************************************
// This routine determines the indention of a line.
// *********************************************************************
void GetIndent(char *str)
{
    for (indent2 = 0; *str; str++)
    {
	if (*str == ' ') indent2++;
	else if (*str == TAB_CHAR) indent2 = (indent2 + 8) & ~7;
	else break;
    }
}

// *********************************************************************
// This routine drops down to lower levels until the block
// indention is less than the current statement's indention.
// It prints the labels and jmps for each level depding on
// the label values that were set.  The ELSEIF block must be
// handled differently.
// *********************************************************************
void DropLevels(int indent1)
{
    while (level >= 0 && indent[level] >= indent1)
    {
	if (looptype[level] == LOOP_ELSEIF)
	{
            fprintf(outfile, "label%4.4d\n", labelend[level]);
            fprintf(outfile, "label%4.4d\n", labelstart[level]);
	}
	else if (looptype[level] == LOOP_CASE)
	{
            fprintf(outfile, "label%4.4d\n", labelend[level]);
	    fprintf(outfile, "        casedone\n");
	}
	else
	{
            if (labelstart[level])
                fprintf(outfile, "        jmp label%4.4d\n", labelstart[level]);
            if (labelend[level])
                fprintf(outfile, "label%4.4d\n", labelend[level]);
	}
        level--;
    }
}

// *********************************************************************
// This routine handles the end of method by dropping to
// level zero and printing a return at the end.
// *********************************************************************
void PrintEndOfPub(void)
{
    DropLevels(-1);
    fprintf(outfile, "        ret\n\n");
}

// *********************************************************************
// This routine searches for an assignment operator, and
// returns it's token index if found.
// *********************************************************************
int LocateAssignment(int num, char **tokens)
{
    int i, len;
    for (i = 0; i < num; i++)
    {
	len = strlen(tokens[i]);
	//printf("%d, %d, %s\n", i, len, tokens[i]);
	if (len > 0 && tokens[i][len-1] == '=' && strcmp(tokens[i], "==")) break;
    }
    //if (i < num) printf("Found %s\n", tokens[i]);
    return i;
}

// *********************************************************************
// This routine performs the second pass through the input file.
// It converts the statements within the Spin methods to Spasm code.
// *********************************************************************
void GenerateSpasm(void)
{
    int index;
    int i, j, num;
    int sectiontype;

    printf("Pass 2: Generating Spasm Code\n");

    sectiontype = 3;
    commentflag = 0;
    linenum = 0;

    while (fgets(buffer, 100, infile))
    {
        linenum++;
	RemoveCRLF(buffer);
	if (CommentLine(buffer))
	{
	    fprintf(outfile, "%s\n", buffer);
	    continue;
	}
	num = GetTokens(buffer, buf2, tokens);

	if (num == 0)
	{
            if (sectiontype != 0 && sectiontype != 1)
	        fprintf(outfile, "\n");
	    continue;
	}
	else if (tokens[0][0] == SINGLE_QUOTE)
	{
	    fprintf(outfile, "%s\n", buffer);
	    continue;
	}

	GetIndent(buffer);

        // Search for a start of a Spin section
        index = FindListNoCase(tokens[0], section_keyword);

        // Check if in a current pub
        if (sectiontype < 2)
	{
	    if (index >= 0)
	    {
	        PrintEndOfPub();
                sectiontype = index;
                fprintf(outfile, "%s\n", buffer);
                // pub or pri
		if (index < 2)
		    ProcessPubStatement(2, num, tokens);
	    }
	    else
	    {
		if (FindListNoCase(tokens[0], elselist) >= 0)
		    DropLevels(indent2+1);
		else
		    DropLevels(indent2);
	        fprintf(outfile, "\n'%s\n", buffer);
	        if (looptype[level] == LOOP_CASE && indent2 > indent[level])
	        {
		    ProcessCaseItem(0, num, tokens);
	        }
		else
		{
		    j = LocateAssignment(num, tokens);
		    if (j < num)
		        ProcessStore(0, num, tokens, j);
		    else
		        ProcessNonStore(0, num, tokens);
		}
	    }
	}
        // Otherwise, just print the line and check for a pub
	else
	{
            i = 0;
            // Ensure that labels start in the first colunm in a DAT section
            if (sectiontype == 2 && FindListNoCase(tokens[0], type_keyword) < 0)
            {
                while (buffer[i] == ' ' || buffer[i] == TAB_CHAR)
                    i++;
            }
            fprintf(outfile, "%s\n", &buffer[i]);
            if (index >= 0)
            {
                // Check if pub or pri statement
                if (index < 2)
                    ProcessPubStatement(2, num, tokens);
                sectiontype = index;
            }
	}
    }

    // Print the end of the pub
    if (sectiontype < 2) PrintEndOfPub();
}

void CloseFiles(void)
{
    if (infile)
        fclose(infile);
    if (outfile)
        fclose(outfile);
}

void ErrorExit(void)
{
    printf("%d: %s\n", linenum, buffer);
    CloseFiles();
    exit(1);
}

#ifndef CSPIN_FLAG
void CheckStackSpace(void)
{
}
#endif

#ifdef NO_FUNCTION_PTR
int ExecuteFunction(int fFunction, int i, int num, char **tokens, int state)
{
    int retval;
    //c.printf6(string("ExecuteFunction(%d, %d, %d, %s, %d, %4.4x)\n"), fFunction, i, num, long[tokens][i], state, @retval)
#ifdef CSPIN_FLAG
    if (&retval > maxstackaddr)
    {
        maxstackaddr = &retval;
        if (maxstackaddr > 0x767a)
        {
            printf("Out of stack space\n");
            ErrorExit();
        }
    }
#endif
  
    switch(fFunction)
    {
        case fProcessValue :
            retval = ProcessValue(i, num, tokens, state);
            break;
        case fProcessClkfreq :
            retval = ProcessClkfreq(i, num, tokens, state);
            break;
        case fProcessAccess :
            retval = ProcessAccess(i, num, tokens, state);
            break;
        case fProcessRepeat :
            retval = ProcessRepeat(i, num, tokens, state);
            break;
        case fProcessKey :
            retval = ProcessKey(i, num, tokens, state);
            break;
        case fProcessCase :
            retval = ProcessCase(i, num, tokens, state);
            break;
        case fProcessIf :
            retval = ProcessIf(i, num, tokens, state);
            break;
        case fProcessElseIf :
            retval = ProcessElseIf(i, num, tokens, state);
            break;
        case fProcessElse :
            retval = ProcessElse(i, num, tokens, state);
            break;
        case fProcessQuit :
            retval = ProcessQuit(i, num, tokens, state);
            break;
        case fProcessNext :
            retval = ProcessNext(i, num, tokens, state);
            break;
        case fProcessReturn :
            retval = ProcessReturn(i, num, tokens, state);
            break;
        case fProcessBinaryOp :
            retval = ProcessBinaryOp(i, num, tokens, state);
            break;
        case fProcessBinaryEqOp :
            retval = ProcessBinaryEqOp(i, num, tokens, state);
            break;
        case fProcessOp :
            retval = ProcessOp(i, num, tokens, state);
            break;
        case fProcessUnaryOp :
            retval = ProcessUnaryOp(i, num, tokens, state);
            break;
        case fProcessAt :
            retval = ProcessAt(i, num, tokens, state);
            break;
        case fProcessExtraOp :
            retval = ProcessExtraOp(i, num, tokens, state);
            break;
        case fProcessLeftParen :
            retval = ProcessLeftParen(i, num, tokens, state);
            break;
        case fProcessRightBracket :
            retval = ProcessRightBracket(i, num, tokens, state);
            break;
        case fProcessPub :
            retval = ProcessPub(i, num, tokens, state);
            break;
        case fProcessIntrinsic :
            retval = ProcessIntrinsic(i, num, tokens, state);
            break;
        case fProcessCallObj :
            retval = ProcessCallObj(i, num, tokens, state);
            break;
        case fCatchObject :
            retval = CatchObject(i, num, tokens, state);
            break;
        case fProcessString :
            retval = ProcessString(i, num, tokens, state);
            break;
        default :
            printf("ProcessFunction: Invalid function number - %d\n",
                fFunction);
            ErrorExit();
    }
    return retval;
}
#endif
/*
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
*/
